module debug;

{============================================================================}
{====================   Debugging Information for Testing   =================}
{====================   =====                               =================}
{============================================================================}
{                                                                            }
{    File    :  DEBUG.P                                                      }
{                                                                            }
{    Version :  1.0                                                          }
{                                                                            }
{    Date    :  22/08/91  ( first implementation )                           }
{               28/11/91  ( last change )                                    }
{                                                                            }
{    Language:  Pascal-XSC, Vers. 2.00                                       }
{                                                                            }
{    Author  :  R. Hammer                                                    }
{                                                                            }
{    Site    :  Institute for Applied Mathematics,                           }
{               University of Karlsruhe,                                     }
{               Kaiserstr. 12, 7500 Karlsruhe, Germany.                      }
{                                                                            }
{----------------------------------------------------------------------------}
{                                                                            }
{    Short Description:                                                      }
{                                                                            }
{       Some global flags are declared which may be used in different        }
{       modules to control debug information.                                }
{                                                                            }
{                                                                            }
{    History :                                                               }
{                                                                            }
{       Date     | Author     | Kind of action                               }
{       =========|============|============================================= }
{       22/08/91 | Hammer     | First implementation                         }
{       ---------|------------|--------------------------------------------- }
{       12/09/91 | Hammer     | _READ(), _WRITE() for echoed I/O-operations. }
{       ---------|------------|--------------------------------------------- }
{       30/10/91 | Hammer     | Procedure trace added. It may be used for    }
{                |            | debugging.                                   }
{       ---------|------------|--------------------------------------------- }
{       30/11/91 | Hammer     | 1. I/O for chars added.                      }
{                |            | 2. Procedure FILL added.                     }
{       ---------|------------|--------------------------------------------- }
{                                                                            }
{============================================================================}



{============================================================================}
{==========   Global Variables   ============================================}
{============================================================================}


var
  dbg_flag   : boolean;               { Control output }
  inf_flag   : boolean;               { Information output }
  prt_flag   : boolean;               { Output echo }

global var
  prt        : text;                  { Protocol file }


{============================================================================}
{==========   Set Status   ==================================================}
{============================================================================}


global procedure debug_on;
  begin  dbg_flag := true  end;

{----------------------------------------------------------------------------}

global procedure debug_off;
  begin  dbg_flag := false  end;

{----------------------------------------------------------------------------}

global procedure info_on;
  begin  inf_flag := true  end;

{----------------------------------------------------------------------------}

global procedure info_off;
  begin  inf_flag := false  end;

{----------------------------------------------------------------------------}

global procedure proto_on;

var
  name : string;

begin
  prt_flag := true;
  write('Name of protocol file: ');
  if eoln then readln;
  read(name);
  rewrite(prt,name);
  writeln(prt,'Name of protocol file: ',name);
  writeln; writeln(prt);
end;

{----------------------------------------------------------------------------}

global procedure proto_off;
  begin  prt_flag := false  end;



{============================================================================}
{==========   Get Status   ==================================================}
{============================================================================}


global function debug : boolean;
  begin  debug := dbg_flag  end;

{----------------------------------------------------------------------------}

global function info : boolean;
  begin  info := inf_flag  end;

{----------------------------------------------------------------------------}

global function proto : boolean;
  begin  proto := prt_flag  end;



{============================================================================}
{==========   Echoed I/O Procedures   =======================================}
{============================================================================}


global procedure fill ( n : integer );
var
  i : integer;
begin
  for i := 1 to n do write(' ');
  if (proto) then
    for i := 1 to n do write(prt,' ');
end;

{----------------------------------------------------------------------------}

global procedure _writeln;
  begin  writeln; if (proto) then writeln(prt) end;

{----------------------------------------------------------------------------}

global procedure _write ( ch : char );
  begin  write(ch); if (proto) then write(prt,ch)  end;

{----------------------------------------------------------------------------}

global procedure _writeln ( ch : char );
  begin  writeln(ch); if (proto) then writeln(prt,ch)  end;

{----------------------------------------------------------------------------}

global procedure _read ( var ch : char );
begin
  if (eoln) then readln;
  read(ch);
  if (proto) then writeln(prt,ch)
end;


{----------------------------------------------------------------------------}


global procedure _write ( str : string; n : integer );
  begin  write(str:n); if (proto) then write(prt,str:n)  end;

{----------------------------------------------------------------------------}

global procedure _write ( str : string );
  begin  _write(str,length(str))  end;

{----------------------------------------------------------------------------}

global procedure _writeln ( str : string; n : integer );
  begin  _write(str,n); _writeln  end;

{----------------------------------------------------------------------------}

global procedure _writeln ( str : string );
  begin  _write(str); _writeln  end;

{----------------------------------------------------------------------------}

global procedure _read ( var str : string );
begin
  if (eoln) then readln;
  read(str);
  if (proto) then writeln(prt,str)
end;


{----------------------------------------------------------------------------}


global procedure _write ( n : integer );
  begin  write(n); if (proto) then write(prt,n)  end;

{----------------------------------------------------------------------------}

global procedure _writeln ( n : integer);
  begin  _write(n); _writeln  end;

{----------------------------------------------------------------------------}

global procedure _read ( var n : integer );
  begin  read(n); if (proto) then writeln(prt,n)  end;


{----------------------------------------------------------------------------}


global procedure _write ( r : real );
  begin  write(r); if (proto) then write(prt,r)  end;

{----------------------------------------------------------------------------}

global procedure _writeln ( r : real );
  begin  _write(r); _writeln  end;

{----------------------------------------------------------------------------}

global procedure _read ( var r : real );
  begin  read(r); if (proto) then writeln(prt,r)  end;



{============================================================================}
{==========   Initialization Part   =========================================}
{============================================================================}


procedure trace ( n : integer );
begin
  writeln('Trace point ',n);

  { Idividual output statments may be supplemented }
  {------------------------------------------------}

end;
{============================================================================}
{==========   Initialization Part   =========================================}
{============================================================================}


begin
  debug_off;
  info_off;
  proto_off;
end.
