module mp_ari;

{============================================================================}
{=======================   Multi Precision Arithmetic   =====================}
{=======================   =     =         ===          =====================}
{============================================================================}
{                                                                            }
{    File    :  MP_ARI.P                                                     }
{                                                                            }
{    Version :  1.0                                                          }
{                                                                            }
{    Date    :  29/05/91  ( first implementation )                           }
{               20/12/91  ( last change )                                    }
{                                                                            }
{    Language:  Pascal-XSC, Vers. 2.00                                       }
{                                                                            }
{    Author  :  D. Cordes   ( Design and Implementation )                    }
{               R. Hammer   ( Operators MPREAL <--> REAL )                   }
{                                                                            }
{    Site    :  Institute for Applied Mathematics,                           }
{               University of Karlsruhe,                                     }
{               Kaiserstr. 12, 7500 Karlsruhe, Germany.                      }
{                                                                            }
{----------------------------------------------------------------------------}
{                                                                            }
{    Short Description:                                                      }
{                                                                            }
{       Module for multi precision real arithmetic. This arithmetic is       }
{       also used internally by Pascal-XSC.                                  }
{                                                                            }
{                                                                            }
{    Status  :  Development.                                                 }
{                                                                            }
{               1. If the Pascal-XSC compiler will support multi-precision   }
{                  arithmetic, lines marked by the following symbols are to  }
{                  be updated:                                               }
{                                                                            }
{                     !!!  : to be deleted                                   }
{                                                                            }
{               2. Arguments of not externally declared arithmetic operators }
{                  are defined as reference arguments (--> modified reference}
{                  call).                                                    }
{                                                                            }
{                                                                            }
{                                                                            }
{    History :                                                               }
{                                                                            }
{       Date     | Author     |                                              }
{       =========|============|============================================= }
{       29/05/91 | Cordes     | First implementation                         }
{       ---------|------------|--------------------------------------------- }
{       04/06/91 | Hammer     | Operators for MPREAL <--> REAL added         }
{       ---------|------------|--------------------------------------------- }
{       11/06/91 | Hammer     | Coercion functions added, so the number      }
{                |            | of operators to be implemented is reduced    }
{                |            | to a minumum.                                }
{       ---------|------------|--------------------------------------------- }
{       12/07/91 | Hammer     | Some procedures added: L_RVAL(), MPVAL()     }
{       ---------|------------|--------------------------------------------- }
{       22/10/91 | Hammer     | Global INTEGER- and REAL-Coercion substituted}
{                |            | by explicitly defined operators MPREAL <-->  }
{                |            | REAL.                                        }
{       ---------|------------|--------------------------------------------- }
{       23/11/91 | Hammer     | L_RVAL() now only declared locally.          }
{       ---------|------------|--------------------------------------------- }
{       10/12/91 | Hammer     | 1. MPTTST() added to test the temporary flag }
{                |            |    (-> modified reference call).             }
{                |            | 2. Arguments of SQR(), LOG2(), LOG10(),      }
{                |            |    EXP2(), and EXP10() changed to value      }
{                |            |    parameters.                               }
{       ---------|------------|--------------------------------------------- }
{       20/12/91 | Hammer     | 1. New procedure LTOR() for B_LTOR() which   }
{                |            |    is declared locally, now.                 }
{                |            | 2. Procedure B_RTOL() renamed to RTOL()      }
{       ---------|------------|--------------------------------------------- }
{                                                                            }
{                                                                            }
{============================================================================}



{============================================================================}
{==========   Datatypes, Global Constants and Variables   ===================}
{============================================================================}


global type
  mpreal = ^integer;



{============================================================================}
{==========   Temporary Utilities   =========================================}
{============================================================================}


global procedure mpfree ( var x : mpreal );                      {!!!}
  external l_free;

{----------------------------------------------------------------------------}

global procedure mpinit ( var x : mpreal );                      {!!!}
  external l_init;

{----------------------------------------------------------------------------}

global procedure mptemp ( var x : mpreal );                      {!!!}
  external l_temp;

{----------------------------------------------------------------------------}

global 