(* RCS $Id: x_intg.p,v 1.1 1994/02/25 13:11:15 ae59 Exp $ *)
(*-----------------------------------------------------------*)
(* DESCRIPTION : module with integer operations              *)
(*-----------------------------------------------------------*)
{---------------------------------------------------------------
 |      x_intg : module with integer operations                |
 |      ======                                                 |
 |                                                             |
 |      Operators :   a  AND   b   ==> INTEGER    Priotity: *  |
 |                    a  BCLR  m   ==> INTEGER    Priotity: *  |
 |                    a  BSET  m   ==> INTEGER    Priotity: *  |
 |                    a BTEST  m   ==> BOOLEAN    Priotity: *  |
 |                    a  EQV   b   ==> INTEGER    Priotity: +  |
 |                       MSB   a   ==> INTEGER    Priotity: ^  |
 |                       NOT   a   ==> INTEGER    Priotity: ^  |
 |                       ONES  a   ==> INTEGER    Priotity: ^  |
 |                    a   OR   b   ==> INTEGER    Priotity: +  |
 |                    a ROTATE m   ==> INTEGER    Priotity: *  |
 |                    a SHIFT  m   ==> INTEGER    Priotity: *  |
 |                    a  XOR   b   ==> INTEGER    Priotity: +  |
 |                                                             |
 |      The following operators replace the default INTEGER    |
 |      operations +,-,* ! Range checking is done.             |
 |      !                 -    a   ==> INTEGER    Priority: ^  |
 |      !             a   +    b   ==> INTEGER    Priority: +  |
 |      !             a   -    b   ==> INTEGER    Priority: +  |
 |      !             a   *    b   ==> INTEGER    Priority: *  |
 |                                                             |
 |      Procedure :     write(f,i:mode:m:n)                    |
 |                      mode = 'b','B','x','X'                 |
 |                                                             |
 |      An integer is assumed to consist of 32 bits which are  |
 |      numbered with increasing significance from 0 to 31.    |
 |      Bit 0 is the least significant bit of an integer.      |
 |      Bit 31 is the most significant bit of an integer.      |
 |      A bit is called 'set' if it represents the value '1'.  |
 |      A bit is called 'cleared' if it represents the         |
 |      value '0'.                                             |
 |      A 'shift'/'rotate' to the left is done for a           |
 |      positive shift value.                                  |
 |      A 'shift'/'rotate' to the right is done for a          |
 |      negative shift value.                                  |
 |                                                             |
 |      Date : 1991-11-06                                      |
 ---------------------------------------------------------------}
MODULE x_intg;

{---------------------------------------------------------------
 |      priority setting of operator names                     |
 ---------------------------------------------------------------}
PRIORITY BCLR   = *;
	 BSET   = *;
	 BTEST  = *;
	 EQV    = +;
	 MSB    = ^;
	 ONES   = ^;
	 ROTATE = *;
	 SHIFT  = *;
	 XOR    = +;

{---------------------------------------------------------------
 |      AND : operator performs bit-wise 'and' operation for   |
 |      ===   integers a,b                                     |
 ---------------------------------------------------------------}
GLOBAL OPERATOR AND ( a,b : INTEGER ) bit_and : INTEGER;
   EXTERNAL a_band;

{---------------------------------------------------------------
 |      BCLR : operator returns integer 'a' with bit 'm'       |
 |      ====   cleared                                         |
 ---------------------------------------------------------------}
GLOBAL OPERATOR BCLR ( a,m : INTEGER ) bit_clr : INTEGER;
   EXTERNAL a_bclr;

{---------------------------------------------------------------
 |      BSET : operator returns integer 'a' with bit 'm' set   |
 |      ====                                                   |
 ---------------------------------------------------------------}
GLOBAL OPERATOR BSET ( a,m : INTEGER ) bit_set : INTEGER;
   EXTERNAL a_bset;

{---------------------------------------------------------------
 |      BTEST : operator tests bit 'm' of integer 'a'          |
 |      =====   to be set                                      |
 ---------------------------------------------------------------}
GLOBAL OPERATOR BTEST ( a,m : INTEGER ) bit_set : BOOLEAN;
   EXTERNAL a_btst;

{---------------------------------------------------------------
 |      EQV : operator performs bit-wise 'equivalence'         |
 |      ===   operation for integers a,b                       |
 ---------------------------------------------------------------}
GLOBAL OPERATOR EQV ( a,b : INTEGER ) bit_eqv : INTEGER;
   EXTERNAL a_beqv;

{---------------------------------------------------------------
 |      MSB : operator returns the bit index of the most       |
 |      ===   significant bit of integer 'a' which is set      |
 ---------------------------------------------------------------}
GLOBAL OPERATOR MSB ( a : INTEGER ) bit_msb : INTEGER;
   EXTERNAL a_bmsb;

{---------------------------------------------------------------
 |      NOT : operator returns the one's complement of         |
 |      ===   integer 'a'                                      |
 ---------------------------------------------------------------}
GLOBAL OPERATOR NOT ( a : INTEGER ) bit_not : INTEGER;
   EXTERNAL a_bnot;

{---------------------------------------------------------------
 |      ONES : operator returns the number of bits which are   |
 |      ====   set in integer 'a'                              |
 ---------------------------------------------------------------}
GLOBAL OPERATOR ONES ( a : INTEGER ) bit_ones : INTEGER;
   EXTERNAL a_bons;

{---------------------------------------------------------------
 |      OR : operator performs bit-wise 'or' operation for     |
 |      ==   integers a,b                                      |
 ---------------------------------------------------------------}
GLOBAL OPERATOR OR  ( a,b : INTEGER ) bit_or  : INTEGER;
   EXTERNAL a_bor_;

{---------------------------------------------------------------
 |      ROTATE : operator returns integer 'a' with bits        |
 |      ======   rotated by 'm' positions to the left if m>0   |
 |               (to the right if m<0)                         |
 ---------------------------------------------------------------}
GLOBAL OPERATOR ROTATE ( a,m : INTEGER ) bit_rotate : INTEGER;
   EXTERNAL a_brtt;

{---------------------------------------------------------------
 |      SHIFT : operator returns integer 'a' with bits shifted |
 |      =====   by 'm' positions to the left if m>0 (to the    |
 |              right if m<0)                                  |
 ---------------------------------------------------------------}
GLOBAL OPERATOR SHIFT ( a,m : INTEGER ) bit_shift : INTEGER;
   EXTERNAL a_bshf;

{---------------------------------------------------------------
 |      XOR : operator performs bit-wise 'xor' operation for   |
 |      ===   integers a,b                                     |
 ---------------------------------------------------------------}
GLOBAL OPERATOR XOR ( a,b : INTEGER ) bit_xor : INTEGER;
   EXTERNAL a_bxor;

{---------------------------------------------------------------
 |      intmin - : unary minus of integer a (two's complement) |
 |      ========                                               |
 ---------------------------------------------------------------}
GLOBAL OPERATOR - ( a : INTEGER ) intmin : INTEGER;
   EXTERNAL a_umin;

{---------------------------------------------------------------
 |      intadd + : addition of integers a,b                    |
 |      ========                                               |
 ---------------------------------------------------------------}
GLOBAL OPERATOR + ( a,b : INTEGER ) intadd : INTEGER;
   EXTERNAL a_add_;

{---------------------------------------------------------------
 |      intsub + : subtraction of integers a,b                 |
 |      ========                                               |
 ---------------------------------------------------------------}
GLOBAL OPERATOR - ( a,b : INTEGER ) intadd : INTEGER;
   EXTERNAL a_sub_;

{---------------------------------------------------------------
 |      intmul + : subtraction of integers a,b                 |
 |      ========                                               |
 ---------------------------------------------------------------}
GLOBAL OPERATOR * ( a,b : INTEGER ) intmul : INTEGER;
   EXTERNAL a_mul_;

{---------------------------------------------------------------
 |      write : procedure writing a representation of bits 'm' |
 |      =====   through 'n' of integer 'i' to the text file 'f'|
 |              using format 'mode'.                           |
 |              mode = 'b','B' : binary representation         |
 ---------------------------------------------------------------}
GLOBAL PROCEDURE write ( VAR f : TEXT; i : INTEGER; mode : CHAR;
				       m : INTEGER; n : INTEGER );
  VAR k,v,b : INTEGER;
BEGIN
   IF (mode='b') OR (mode='B') THEN
   BEGIN
      IF m>n THEN
      BEGIN
	 FOR k:=m DOWNTO n DO
	    IF i BTEST k THEN WRITE(f,'1') ELSE WRITE(f,'0');
      END
      ELSE
      BEGIN
	 FOR k:=m TO n DO
	    IF i BTEST k THEN WRITE(f,'1') ELSE WRITE(f,'0');
      END
   END
   ELSE
   IF (mode='x') OR (mode='X') THEN
   BEGIN
      v := 0;
      IF m>n THEN
      BEGIN
         b := 3-((m-n) mod 4);
	 FOR k:=m DOWNTO n DO
         BEGIN
            b := b+1;
            v := v SHIFT 1;
	    IF i BTEST k THEN v := v+1;
            IF b=4 THEN 
            BEGIN
               IF v>9 THEN WRITE(f,chr(v+ord(mode)-33))
                      ELSE WRITE(f,chr(v+ord('0')));
               b := 0;  v := 0;
            END;
         END;
      END
      ELSE
      BEGIN
         b := 3-((n-m) mod 4);
	 FOR k:=m TO n DO
         BEGIN
            b := b+1;
            v := v SHIFT 1;
	    IF i BTEST k THEN v := v+1;
            IF b=4 THEN 
            BEGIN
               IF v>9 THEN WRITE(f,chr(v+ord(mode)-33))
                      ELSE WRITE(f,chr(v+ord('0')));
               b := 0;  v := 0;
            END;
         END;
      END;
   END;
END;

END.
