MODULE eingabe;

USE GLOBAL fknoten;


{ E I N G A B E . P A K  }
{}
{ Dieses Unterprogramm-Paket fuehrt die Syntax-Analyse der eingegebenen For-  
{ melausdruecke durch und baut die zugehoerigen Baeume auf ( mit Hilfe der im 
{ Paket FKNOTEN.PAK definierten Funktionen und Operatoren ).                  
{ Die Syntax der zugelassenen Formelausdruecke ist im folgenden in BNF-Nota-  
{ tion angegeben. Hierin bedeutet | die Trennung von Alternativen und <<...>> 
{ bedeutet, dass die Syntax-Einheit in den geschweiften Klammern beliebig oft 
{ hintereinander auftreten darf ( kann auch leer sein, ausser bei <DGL> ).    
{                                                                             
{ <Eingabe>   ::=   <DGL> ; <Ausgabeliste> ;                                  
{                                                                             
{ <Ausgabeliste>::=  <<  <Variable> | <Kommentar>  >>                         
{                                                                             
{ <DGL> ::= <<  <Zuweisung> | <Druckbefehl> | <Var.liste> | <Kommentar>   >>  
{                                                                             
{ <Druckbefehl>::= <Ausdruck> ?  ( Ausgabe des Wertes von <Ausdruck>   )      
{                                                                             
{ <Var-liste> ::=  ?             ( Ausgabe einer Liste aller Variablen )      
{                                                                             
{ <Kommentar> ::= $  Text bis zum Zeilenende                                  
{      ( ; innerhalb eines Kommentars wird nicht als Eingabeende erkannt )    
{                                                                             
{ <Zuweisung> ::= <Variable> = <Ausdruck>                                     
{      ( bei der Eingabe wird auch <Ausdruck> = <Ausdruck> ohne Fehlermeldung 
{        akzeptiert, wenn der linke <Ausdruck> mit einer Variablen beginnt.   
{        In diesem Fall wird dieser ersten Variablen der rechte <Ausdruck>    
{        zugewiesen, der Rest des linken <Ausdruck> wird ignoriert.           
{                                                                             
{ <Ausdruck>  ::= <Term>  |  <<  + <Term> | - <Term>  >>  |                   
{                    <Term>  <<  + <Term> | - <Term>  >>                      
{                                                                             
{   <Term>    ::= <Faktor> | <Faktor>  <<  * <Faktor> | / <Faktor>  >>        
{                                                                             
{  <Faktor>   ::= <Wert> | <Wert>  <<  ^ <Wert>  >>                           
{                                                                             
{   <Wert>    ::= <Variable> | <Konstante> | ( <Ausdruck> ) | <Funktion>      
{      ( eine <Variable> darf erst dann als <Wert> auftreten wenn sie mindes- 
{        tens einmal auf der linken Seite einer Zuweisung gestanden war. Aus- 
{        nahmen sind die U1,...,UN, siehe <reservierte Variable> )            
{                                                                             
{  <Funktion> ::= SQR( <Ausdruck> ) |   SQRT( <Ausdruck> ) |    EXP( <Ausdruck> )|
{                 LOG( <Ausdruck> ) |    SIN( <Ausdruck> ) |    COS( <Ausdruck> )|
{                SINH( <Ausdruck> ) |   COSH( <Ausdruck> ) |   SICO( <SC-Ausdr> )|
{                 TAN( <Ausdruck> ) |    COT( <Ausdruck> ) |   TANH( <SC-Ausdr> )|
{                COTH( <Ausdruck> ) | ARCSIN( <Ausdruck> ) | ARCCOS( <SC-Ausdr> )|
{              ARCTAN( <Ausdruck> ) | ARCCOT( <Ausdruck> ) | ARSINH( <SC-Ausdr> )|
{              ARCOSH( <Ausdruck> ) | ARTANH( <Ausdruck> ) | ARCOTH( <SC-Ausdr> )|
{        Alle Funktionsnamen duerfen auch klein geschrieben werden, Gross-    
{        und Klein-Buchstaben duerfen sogar gemischt werden.                  
{        SICO liefert den "Partner-Knoten" des Arguments (s.u.); also z.B.:   
{                S = sin(2*U1)     C = sico(S)   => es ist:  C = cos(2*U1)    
{        oder:  ch = cosh(U1*U2)  sh = sico(ch)  => es ist: sh = sinh(U1*U2)  
{                                                                             
{  <SC-Ausdr> ::= Funktionsaufruf von SIN, COS, SINH, COSH oder eine Variable,
{                 die durch einen solchen Funktionsaufruf definiert wurde.    
{                                                                             
{  <Variable> ::= <reservierte Variable> | max. SMAX (=15) alphanumerische    
{                                          Zeichen ohne Blank in einer Zeile. 
{                                          Erstes Zeichen keine Ziffer.       
{                                                                             
{  <reservierte Variable> ::= U1 | U2 | ... | UN   |   F1 | F2 | ... | FN     
{                                                                             
{                             Die Ui, 1 <= i <= N, stehen fuer die N gesuchten
{                             Funktionen und duerfen nicht auf der linken Sei-
{                             te einer Zuweisung stehen.                      
{                             Die Fi, 1 <= i <= N, stehen fuer die N Funktio- 
{                             nen der rechten Seite der DGL und duerfen wie   
{                             normale Variablen verwendet werden.  Zu Beginn  
{                             der Eingabe sind alle Fi mit 0 initialisiert.   
{                             ( Auch nach jedem Teilabschnitt, wenn waehrend  
{                             der Integration die DGL geaendert wird. )       
{                                                                             
{ <Konstante> ::= <REAL-Konstante> | <Intervall-Konstante>                    
{                                                                             
{ <Intervall-Konstante> ::= [ <REAL-Konstante> , <REAL-Konstante> ]           
{                                                                             
{   <REAL-Konstante>    ::= REAL-Konstante wie in PASCAL-SC                   
}

GLOBAL
FUNCTION LASTITEM( VAR F : TEXT ) : BOOLEAN;
{ Ueberliest Blanks, F^ bleibt auf dem naechsten Zeichen <> Blank stehen
{ ( falls vorhanden ). LASTITEM:= TRUE falls EOF(F) erreicht, sonst FALSE.
}
BEGIN
  WHILE NOT EOF(F) AND (F^=' ') DO GET(F);
  LASTITEM:= EOF(F);
END   {  LASTITEM  };

GLOBAL
FUNCTION F_AUFBAUEN ( N : INTEGER; VAR ANZKNOT : INTEGER; 
                      VAR INP,INFILE,OUTP,OUTFILE : TEXT;
                      VAR F_START : H_VARIABLEN;
                      VAR U : FVECTOR; VAR T : FORMEL;
                      VAR PROTOKOLL : BOOLEAN )  : FVECTOR[1..N];
{  N        : Anzahl Differentialgleichungen (Eingabe)             }
{  ANZKNOT  : Anzahl Knoten aller eingelesenen Ausdruecke (Ausgabe) }
{  INFILE   : Console oder Diskfile fuer Dateneingabe               }   
{  OUTP  : Ausgabefile fuer Benutzerdialog ( = OUTPUT )          }
{  F_START  : Start der Ausgabeliste (Ausgabe)                     }
{  U        : Formelknoten fuer die unbekannten Funtionen U1,...,UN }
{             (Ausgabe)                                            }
{  T        : Formelknoten fuer die unabhaengige Variable t (Ausgabe)}
VAR

   W              :  INTERVAL;      { Wert, falls Real- oder Inter-}
                                    { vall-Konstante eingelesen    }
   VARI,VAR_LISTE :  H_VARIABLEN;   { Zeiger zum Aufbau einer Liste}
                                    { aller Eingabe-Variablen      }
   ATOM           :  OPERATION;
   LEFTSIDE       :  BOOLEAN;

FUNCTION KEIN_TRENNZ : BOOLEAN;
{ Stellt fest ob in INFILE^ ein Trennzeichen steht ( = alles was <> Buchstabe
{ und <> Ziffer ist ), und liefert als Nebeneffekt in ATOM das in INFILE^ ge-
{ fundene Zeichen ( in den Typ OPERATION konvertiert ).                       
{ Globale Variablen : ATOM, INFILE
}
BEGIN
  KEIN_TRENNZ:= FALSE;
  CASE INFILE^ OF
  '+' : ATOM:= PLUS;
  '-' : ATOM:= MINUS;
  '*' : ATOM:= MAL;
  '/' : ATOM:= DURCH;
  '^' : ATOM:= HOCH;
  '(' : ATOM:= KLAUF;
  ')' : ATOM:= KLZU;
  '[' : ATOM:= ECKAUF;
  '=' : ATOM:= GLEICH;
  ';' : ATOM:= STRICHP;
  '?' : ATOM:= PRT;
  '$' : ATOM:= DOLLAR;
  ' ' : IF  LASTITEM(INFILE)  THEN  ATOM:= FEHLER  ELSE  ATOM:= BLANK;
  '0'..'9'              : BEGIN
                            ATOM:= ZIFFER;  KEIN_TRENNZ:= TRUE;
                          END;
  'A'..'Z','a'..'z','_' : BEGIN
                            ATOM:= BUCHST;  KEIN_TRENNZ:= TRUE;
                          END;
  ELSE : ATOM:= FEHLER;
  END;
END  {  KEIN_TRENNZ  };
 
FUNCTION SUCH( S : STRING ) : H_VARIABLEN;
{ Sucht in der Variablenliste ( die bei VAR_LISTE beginnt ) ob eine Variable  
{ mit dem Name S eingetragen ist. Wenn ja, wird als Ergebnis ein Verweis auf  
{ diese Variable geliefert, wenn nein, wird eine neue Variable mit dem Namen  
{ S erzeugt, eingetragen und ein Verweis darauf als Ergebnis geliefert.       
{ Globale Variable : VAR_LISTE                                                
}
VAR H : H_VARIABLEN;
    B : BOOLEAN;
BEGIN
  H:= VAR_LISTE;  B:= TRUE;
  WHILE (H<>NIL) AND B DO
     IF  H^.NAME<>S  THEN  H:= H^.NEXT  ELSE  B:= FALSE;
  IF H=NIL THEN
  BEGIN
    NEW(H);
    H^.NAME:= S;
    H^.AUSDRUCK:= NIL;
    H^.NEXT:= VAR_LISTE;
    VAR_LISTE:= H;
  END;
  SUCH:= H;
END  {  SUCH  };

PROCEDURE NEXTATOM;
{ Zerlegt die Eingabe von INFILE in syntaktische "Atome" (vom Typ OPERATION),
{ die in der globalen Variablen ATOM an die aufrufenden Prozeduren weiterge-  
{ geben werden. Es muessen nur diejenigen Atome untersucht werden, die in     
{ KEIN_TRENNZ noch nicht vollstaendig erkannt wurden. Dies sind :             
{    BUCHST : KEIN_TRENNZ hat nur einen Buchstaben erkannt; dieser kann ge-   
{             hoeren zu einer Variablen (ATOM=VARIABLE) oder zu einer Stan-   
{             dardfunktion (ATOM=QUAD,WURZ,EFKT,LOGA,SINUS,COSIN,SINUSH oder  
{             COSINH).                                                        
{             Wurde eine Variable gefunden, so wird in der globalen Variablen 
{             VARI ein Verweis auf deren Eintrag in der Variablen-Liste ge-   
{             liefert.                                                        
{    ZIFFER : KEIN_TRENNZ hat nur eine Ziffer erkannt; diese ist der Anfang   
{             einer REAL-Zahl, die eingelesen werden muss. Der Wert dieser    
{             Konstanten wird in der globalen Variablen W weitergegeben.      
{    BLANK  : KEIN_TRENNZ hat ein Blank gelesen; dieses muss ignoriert werden 
{             (rekursiver Aufruf von NEXTATOM).                               
{    DOLLAR : KEIN_TRENNZ hat ein Dollar-Zeichen gelesen; da dieses der Beginn
{             eines Kommentars ist, muss der Rest der Zeile ueberlesen werden.
{             Anschliessend NEXTATOM (rekursiv) aufrufen um naechstes Atom zu 
{             erhalten.                                                       
{    FEHLER : In KEIN_TRENNZ wurde ein falsches Zeichen oder EOF gefunden.    
{ GLobale Variablen : ATOM, W, VARI, INFILE
}
VAR S : STRING[SMAX];
    I : INTEGER;
BEGIN
  IF NOT LASTITEM(INFILE) THEN
    IF KEIN_TRENNZ THEN
      IF ATOM=ZIFFER THEN
      BEGIN      { REAL - Konstante liegt vor }
        READ(INFILE,W);
        ATOM:= KONST;
      END ELSE
      BEGIN      {  ATOM = BUCHST, Variable oder Standardfunktion liegt vor }
        S:= '';
        WHILE KEIN_TRENNZ DO BEGIN               { Namen in S einlesen. }
                                S:= S + INFILE^;
                                GET(INFILE);
                              END;
        IF      (S = 'SQR'   ) OR ( S = 'sqr'   ) THEN ATOM:= QUAD
        ELSE IF (S = 'SQRT'  ) OR ( S = 'sqrt'  ) THEN ATOM:= WURZ
        ELSE IF (S = 'EXP'   ) OR ( S = 'exp'   ) THEN ATOM:= EFKT
        ELSE IF (S = 'LOG'   ) OR ( S = 'log'   ) THEN ATOM:= LOGA
        ELSE IF (S = 'SIN'   ) OR ( S = 'sin'   ) THEN ATOM:= SINUS
        ELSE IF (S = 'COS'   ) OR ( S = 'cos'   ) THEN ATOM:= COSIN
        ELSE IF (S = 'TAN'   ) OR ( S = 'tan'   ) THEN ATOM:= TANG
        ELSE IF (S = 'COT'   ) OR ( S = 'cot'   ) THEN ATOM:= COTANG
        ELSE IF (S = 'SINH'  ) OR ( S = 'sinh'  ) THEN ATOM:= SINUSH
        ELSE IF (S = 'COSH'  ) OR ( S = 'cosh'  ) THEN ATOM:= COSINH
        ELSE IF (S = 'TANH'  ) OR ( S = 'tanh'  ) THEN ATOM:= TANGH
        ELSE IF (S = 'COTH'  ) OR ( S = 'coth'  ) THEN ATOM:= COTANGH
        ELSE IF (S = 'SICO'  ) OR ( S = 'sico'  ) THEN ATOM:= SINCOS
        ELSE IF (S = 'ARCSIN') OR ( S = 'arcsin') THEN ATOM:= ARCSI
        ELSE IF (S = 'ARCCOS') OR ( S = 'arccos') THEN ATOM:= ARCCO
        ELSE IF (S = 'ARCTAN') OR ( S = 'arctan') THEN ATOM:= ARCTG
        ELSE IF (S = 'ARCCOT') OR ( S = 'arccot') THEN ATOM:= ARCCTG
        ELSE IF (S = 'ARSINH') OR ( S = 'arsinh') THEN ATOM:= ARSIH
        ELSE IF (S = 'ARCOSH') OR ( S = 'arcosh') THEN ATOM:= ARCOH
        ELSE IF (S = 'ARTANH') OR ( S = 'artanh') THEN ATOM:= ARTGH
        ELSE IF (S = 'ARCOTH') OR ( S = 'arcoth') THEN ATOM:= ARCTGH
        ELSE BEGIN
                IF (S = string('T'))  THEN ATOM:= UVARIABLE ELSE ATOM:= AVARIABLE;
                VARI:= SUCH( S );
             END;
      END  { ATOM = BUCHST }
    ELSE
    IF ATOM=ECKAUF THEN
    BEGIN  { Intervall-Konstante liegt vor }
      READ( INFILE, W );
      ATOM:= KONST;
    END ELSE
    IF  ATOM=BLANK  THEN NEXTATOM ELSE
    IF  ATOM=DOLLAR THEN BEGIN
                           WHILE NOT EOLN(INFILE) DO GET(INFILE);
                           NEXTATOM;
                         END  ELSE
                         BEGIN
                            IF ATOM=FEHLER THEN 
	                       IF ORD(INFILE^) IN [32..126]
				  THEN WRITELN(OUTP,NEXTA_1,'''',INFILE^,'''')
	                          ELSE WRITELN(OUTP,NEXTA_1,'chr(',ORD(INFILE^),')');
                            GET(INFILE);
                         END
  ELSE WRITELN(OUTP,NEXTA_2);
END  {  NEXTATOM  };
 
PROCEDURE LISTEN_BAUEN;
{ Legt die Variablenliste VAR_LISTE an und traegt bereits die 2*N reservier-  
{ ten Variablen U1,...,UN und F1,...,FN ein ( in der Reihenfolge F1,U1,F2,U2, 
{ ...,FN,UN ). Am Ende Zeigen VAR_LISTE und F_START beide auf die zuletzt ein-
{ getragene Variable UN. Beim weiteren Einlesen in der Prozedur F_AUBAUEN wird
{ VAR_LISTE weiter veraendert, wenn neue Variablen eingetragen werden, aber   
{ F_START bleibt unveraendert, um ein schnelleres Auffinden der Fi zu ermoeg- 
{ lichen ( zur Zuweisung an F_VON_U ).                                        
{ Globale Variablen : VAR_LISTE, N, VARI, F_START                             
}
VAR  I : INTEGER;
     S : STRING[SMAX];
BEGIN
  VAR_LISTE:= NIL;
  FOR I:= 1 TO N DO
  BEGIN
    IF I<10 THEN  S:= 'F' + CHR( I + ORD('0') )
            ELSE  S:= 'F' + CHR( I DIV 10 + ORD('0') )
                          + CHR( I MOD 10 + ORD('0') );
    NEW(VARI);
    VARI^.NAME    := S;
    VARI^.AUSDRUCK:= CON(0);
    VARI^.NEXT    := VAR_LISTE;
    VAR_LISTE     := VARI;
    S[1]:= 'U';
    U[I]:= AVARIA(I,0);
    NEW(VARI);
    VARI^.NAME    := S;
    VARI^.AUSDRUCK:= U[I];
    VARI^.NEXT    := VAR_LISTE;
    VAR_LISTE     := VARI;
  END;
  F_START:= VAR_LISTE;
  S:= 'T';
  T:= UVARIA;
  NEW(VARI);
  VARI^.NAME    := S;
  VARI^.AUSDRUCK:= T;
  VARI^.NEXT    := VAR_LISTE;
  VAR_LISTE     := VARI;
END {  LISTEN_BAUEN  };
 
FUNCTION AUSDRUCK : FORMEL;  FORWARD;
{ FORWARD-Deklaration wegen rekursiver Aufrufe bereits in WERT noetig }
 
FUNCTION WERT : FORMEL;
{ Liest die Syntax-Einheit <Wert> nach obiger Grammatik ein und liefert als   
{ Ergebnis einen entsprechenden Knoten zum Aufbau des Formel-Baumes. Diese    
{ Knoten werden mit Hilfe der in FKNOTEN.PAK enthaltenen Funktionen und Opera-
{ toren erzeugt.                                                              
{ Globale Variablen : ATOM, VARI                                              
}
VAR P  : FORMEL;
    OP : OPERATION;
    NA : STRING;
BEGIN
  CASE ATOM OF
  AVARIABLE,
  UVARIABLE: IF VARI^.AUSDRUCK = NIL THEN
             BEGIN
	       NA:= VARI^.NAME;
	       NEXTATOM;
	       IF NOT LEFTSIDE OR ( LEFTSIDE AND (ATOM<>GLEICH) ) THEN
		      WRITELN(OUTP,WERT_0,NA);
	       P:= CON(0);
             END ELSE
	     BEGIN
	        P:= VARI^.AUSDRUCK;
	        NEXTATOM;
		IF ATOM = KLAUF THEN WRITELN(OUTP,WERT_01,VARI^.NAME);
             END;
  KONST :    BEGIN
	       P:= CON(W);
	       NEXTATOM;
             END;
  KLAUF :    BEGIN
               NEXTATOM;
               P:= AUSDRUCK;
               IF ATOM<>KLZU THEN  WRITELN(OUTP,WERT_1);
               NEXTATOM;
             END;
  QUAD,WURZ,EFKT,LOGA,SINUS,COSIN,TANG,COTANG,SINUSH,COSINH,TANGH,COTANGH,
  ARCSI,ARCCO,ARCTG,ARCCTG,ARSIH,ARCOH,ARTGH,ARCTGH,SINCOS:
             BEGIN
               OP:= ATOM; NEXTATOM;
               IF ATOM<>KLAUF THEN WRITELN(OUTP,WERT_2)
                              ELSE NEXTATOM;
               CASE OP OF
               QUAD  : P:= SQR   ( AUSDRUCK );
               WURZ  : P:= SQRT  ( AUSDRUCK );
               EFKT  : P:= EXP   ( AUSDRUCK );
               LOGA  : P:= LN    ( AUSDRUCK );
               SINUS : P:= SIN   ( AUSDRUCK );
               COSIN : P:= COS   ( AUSDRUCK );
               TANG  : P:= TAN   ( AUSDRUCK );
               COTANG: P:= COT   ( AUSDRUCK );
               SINUSH: P:= SINH  ( AUSDRUCK );
               COSINH: P:= COSH  ( AUSDRUCK );
               TANGH : P:= TANH  ( AUSDRUCK );
              COTANGH: P:= COTH  ( AUSDRUCK );
               ARCSI : P:= ARCSIN( AUSDRUCK );
               ARCCO : P:= ARCCOS( AUSDRUCK );
               ARCTG : P:= ARCTAN( AUSDRUCK );
               ARCCTG: P:= ARCCOT( AUSDRUCK );
               ARSIH : P:= ARSINH( AUSDRUCK );
               ARCOH : P:= ARCOSH( AUSDRUCK );
               ARTGH : P:= ARTANH( AUSDRUCK );
               ARCTGH: P:= ARCOTH( AUSDRUCK );
               SINCOS: BEGIN
                         P:= AUSDRUCK; 
                         IF NOT (P^.OPER IN [SINUS..COSINH])
                              THEN WRITELN(OUTP,WERT_3)
                              ELSE P:= P^.RE;
                       END;
               END;
               IF ATOM<>KLZU THEN  WRITELN(OUTP,WERT_4);
	       NEXTATOM;
             END;
    ELSE:    BEGIN
               { P undefiniert }
	       NEXTATOM;
             END;
  END;
  WERT:= P;
END  {  WERT  };
 
FUNCTION FAKTOR : FORMEL;
{ Liest die Syntax-Einheit <Faktor> nach obiger Grammatik ein und liefert als 
{ Ergebnis einen entsprechenden Knoten zum Aufbau des Formel-Baumes. Diese    
{ Knoten werden mit Hilfe des in FKNOTEN.PAK enthaltenen Operators ** erzeugt.
{ Globale Variable : ATOM                                                     
}
VAR P : FORMEL;
BEGIN
   P:= WERT;
   WHILE ATOM=HOCH DO
   BEGIN
      NEXTATOM;
      P:= P ** WERT;
   END;
   FAKTOR:= P;
END  {  FAKTOR  };
 
FUNCTION TERM : FORMEL;
{ Liest die Syntax-Einheit <Term> nach obiger Grammatik ein und liefert als   
{ Ergebnis einen entsprechenden Knoten zum Aufbau des Formel-Baumes. Diese    
{ Knoten werden mit Hilfe der in FKNOTEN.PAK enthaltenen Operatoren * und /   
{ erzeugt.                                                                    
{ Globale Variable : ATOM                                                     
}
VAR P  : FORMEL;
    OP : OPERATION;
BEGIN
  P:= FAKTOR;
  WHILE (ATOM=MAL) OR (ATOM=DURCH) DO
  BEGIN
    OP:= ATOM; NEXTATOM;
    IF OP=MAL THEN P:= P * FAKTOR
              ELSE P:= P / FAKTOR;
  END;
  TERM:= P;
END  {  TERM  };
 
FUNCTION AUSDRUCK : FORMEL;
{ Liest die Syntax-Einheit <Ausdruck> nach obiger Grammatik ein und liefert   
{ als Ergebnis einen entsprechenden Knoten zum Aufbau des Formel-Baumes. Diese
{ Knoten werden mit Hilfe der in FKNOTEN.PAK enthaltenen Operatoren + und -   
{ erzeugt.                                                                    
{ Globale Variable : ATOM                                                     
}
VAR P  : FORMEL;
    OP : OPERATION;
BEGIN
  IF ATOM=PLUS THEN
  BEGIN
    NEXTATOM;  P:= TERM;
  END ELSE
  IF ATOM=MINUS THEN
  BEGIN
    NEXTATOM;  P:= - TERM;
  END ELSE  P:= TERM;
  WHILE (ATOM=PLUS) OR (ATOM=MINUS) DO
  BEGIN
    OP:= ATOM;  NEXTATOM;
    IF OP=PLUS THEN P:= P + TERM
               ELSE P:= P - TERM;
  END;
  AUSDRUCK:= P;
END  {  AUSDRUCK  };
 
PROCEDURE ZUWEIS;
{ Liest die Syntax-Einheit <Zuweisung> nach obiger Grammatik ein und traegt   
{ einen Verweis auf den rechts von := gefundenen Ausdruck in die Komponente   
{ AUSDRUCK der links von := stehenden Variablen ein. Der Wert dieser Variablen
{ ist also ab sofort der gelesene Ausdruck.                                   
{ Globale Variablen : ATOM, VARI                                              
}
VAR H : H_VARIABLEN;
    A : FORMEL;
BEGIN
   LEFTSIDE:= TRUE;
   IF ATOM=PRT THEN
   BEGIN
      H:= VAR_LISTE;
      WHILE H<>NIL DO
      BEGIN
	 IF    H^.AUSDRUCK = NIL    THEN {WRITELN( OUTP, WERT_0,H^.NAME)} ELSE
         IF H^.AUSDRUCK^.OPER=KONST THEN  WRITELN( OUTP, H^.NAME:SMAX,' = ',
							 H^.AUSDRUCK^.WERT );
         H:= H^.NEXT;
      END;
      NEXTATOM;
   END ELSE 
   BEGIN
      IF ATOM=AVARIABLE THEN  H:= VARI;
      A:= AUSDRUCK;
      IF ATOM=GLEICH THEN
      BEGIN
	LEFTSIDE:= FALSE;
        NEXTATOM;
        H^.AUSDRUCK:= AUSDRUCK;
      END ELSE
      IF ATOM=PRT THEN
      BEGIN
	 IF    A = NIL    THEN  WRITELN( OUTP, WERT_0,H^.NAME) ELSE
         IF A^.OPER=KONST THEN  WRITELN( OUTP, A^.WERT );
         NEXTATOM;
      END;
   END;
END  {  ZUWEIS  };
 
{ Diese Prozedur ist das einzige Verbindungs-Stueck zwischen dem Hauptprogramm
{ und den Prozeduren und Funktionen hier in EINGABE.PAK ( mit Ausnahme von    
{ LASTITEM ). F_AUFBAUEN veranlasst den Aufbau der Variablenliste VAR_LISTE,  
{ sowie das Einlesen aller Zuweisungen bis der beendende Strichpunkt erkannt  
{ wird. Danach werden die Ausdruecke, die in der Variablenliste fuer die re-  
{ servierten Variablen F1,...,FN eingetragen sind, in den Array F_VON_U ueber-
{ tragen, der im Hauptprogramm stets die rechte Seite der DGL darstellt.      
{ Nachdem der erste Strichpunkt erkannt wurde ist die Eingabe der DGL beendet 
{ und es kann nun eine Liste von Variablen eingegeben werden, die waehrend der
{ Eingabe der DGL definiert wurden und deren Werte in jedem Zeitschritt ausge-
{ geben werden sollen ( Diese Ausgabe erscheint in umgekehrter Reihenfolge wie
{ in dieser Liste. ). Die Liste ist ( auch wenn sie leer ist ) mit einem wei- 
{ teren Strichpunkt zu beenden. Ist die Ausgabeliste leer, so wird die Ausgabe
{ durch die Variable AUSGABE gesteuert ( Siehe Hauptprogramm, insbesondere die
{ Prozeduren EINLESEN1 und SCHREIBE.                                          
{ Globale Variablen : N, ATOM, VARI, F_START, F_VON_U                         
}
VAR I                : INTEGER;
    C                : CHAR;
    H,H1             : H_VARIABLEN;
    FIRST,INFI       : BOOLEAN;
    INFNAME,OUTFNAME : STRING[127];
BEGIN
  WRITELN(OUTP,FAUF_6);
  WRITE  (OUTP,FAUF_7,chr(8));
  READLN(INP); READ(INP,INFNAME);

  INFI:= LENGTH(INFNAME) > 0;
  IF INFI THEN reset(INFILE,INFNAME)
          ELSE reset(INFILE,'' );

  IF NOT INFI THEN WRITELN(OUTP,FAUF_8);

  LISTEN_BAUEN;
  NEXTATOM;
  WHILE ATOM<>STRICHP DO  ZUWEIS;
  

  ANZKNOT:= N;
  {  F_AUFBAUEN durch F1,...,FN definieren :   }
  VARI:= F_START;  I:= N;
  WHILE VARI<>NIL DO
  BEGIN
    VARI:= VARI^.NEXT;
    F_AUFBAUEN[I]:= VARI^.AUSDRUCK;
    COUNT_KNOTEN( VARI^.AUSDRUCK, ANZKNOT );
    I:= I-1;
    VARI:= VARI^.NEXT;
  END;

  {  Ausgabeliste einlesen und zusammenbauen : }
  WRITELN(OUTP);
  WRITELN(OUTP,FAUF_9);
  WRITELN(OUTP);
  F_START:= NIL;
  FIRST:= TRUE;
  NEXTATOM;
  WHILE ATOM<>STRICHP DO
  BEGIN
     IF (ATOM<>AVARIABLE) AND (ATOM<>UVARIABLE) THEN WRITELN(OUTP,FAUF_10)           ELSE
     IF       VARI^.AUSDRUCK = NIL              THEN WRITELN(OUTP,WERT_0,VARI^.NAME) ELSE
     BEGIN
        NEW(H);  
        H^.NAME:= VARI^.NAME;
        H^.AUSDRUCK:= VARI^.AUSDRUCK;
        H^.NEXT:= NIL;
        IF FIRST THEN 
        BEGIN
           F_START:= H;
           FIRST:= FALSE;
        END 
        ELSE  H1^.NEXT:= H;
        H1:= H;
     END;
     NEXTATOM;
  END;

  IF INFI THEN
     IF LASTITEM(INFILE) THEN reset(INFILE,'');


  WRITE(OUTP,FAUF_11,CHR(8));
  READLN(INP); READ(INP,OUTFNAME);

  PROTOKOLL:=  LENGTH(OUTFNAME) > 0;

  IF PROTOKOLL THEN
  BEGIN
     rewrite(OUTFILE,OUTFNAME);
     WRITELN(OUTFILE);
     IF INFI THEN WRITE(OUTFILE,FAUF_12,INFNAME)
             ELSE WRITE(OUTFILE,FAUF_13);
     WRITELN(OUTFILE,FAUF_14);
     WRITELN(OUTFILE);
  END;

END  {  F_AUFBAUEN  };

{}
{ -----------  ENDE   E I N G A B E . P A K  --------------------------- }

END.
