MODULE fknoten;

USE GLOBAL awatypes;

GLOBAL
PROCEDURE COUNT_KNOTEN( F : FORMEL; VAR ANZ : INTEGER );
BEGIN
   WITH F^ DO
   IF I=0 THEN
   BEGIN
      IF OPER<>KONST THEN  ANZ:= SUCC(ANZ);
      I:= ANZ;
      CASE OPER OF
         PLUS,MINUS,MAL,DURCH,
         SINUS,COSIN,TANG,COTANG,
	 SINUSH,COSINH,TANGH,COTANGH,
	 ARCSI,ARCCO,ARCTG,ARCCTG,
	 ARSIH,ARCOH,ARTGH,ARCTGH :   BEGIN
                                         COUNT_KNOTEN( LI, ANZ );
                                         COUNT_KNOTEN( RE, ANZ );
                                      END;
         UMINUS,QUAD,WURZ,EFKT,LOGA,HOCH: COUNT_KNOTEN( R, ANZ );
         AVARIABLE,UVARIABLE,KONST : ; 
      END;
   END;
END;


{ F K N O T E N . P A K   }
{ }
{ Dieses Modul enthaelt alle Funktionen und Operatoren, die noetig
{ sind, um die Knoten zu erzeugen, die in der Baumstruktur zur Formeldar- 
{ stellung verwendet werden. Zum Teil werden sehr einfache algebraische Um-   
{ formungen vorgenommen ( Addition von 0, Multiplikation mit 0 oder 1 etc. ). 
{                                                                             
{ Die Funktion KNOTEN uebernimmt dabei zentral die eigentliche Erzeugung der 
{ Knoten ( mit Ausnahme der Funktion VARIA ). Die Parameter bedeuten dabei :  
 
{  OP : die Operation, die der Knoten darstellen soll                         
{ A,B : Zeiger auf den linken und den rechten Operanden ( falls jeweils vor-  
{       handen, sonst = NIL )                                                 
{   I : Intervall-Datum, bei OP=KONST der Wert der Konstanten,                
{                        bei OP=HOCH  der Wert des Exponenten.                
}

FUNCTION KNOTEN( OP : OPERATION; A,B : FORMEL; W : INTERVAL ) : FORMEL;
{ Globale Variablen : FERTIG, N }
VAR F : FORMEL;
    K : INTEGER;
BEGIN
   NEW( F );
   F^.OPER  := OP;
   F^.STATUS:= 0; { FERTIG }
   F^.I     := 0;
   CASE OP OF
      PLUS,MINUS,MAL,DURCH,
      TANG,COTANG,TANGH,COTANGH,
      ARCSI,ARCCO,ARCTG,ARCCTG,
      ARSIH,ARCOH,ARTGH,ARCTGH:   BEGIN
                                     F^.LI:= A;
                                     F^.RE:= B;
                                  END;
      SINUS,COSIN,SINUSH,COSINH:  BEGIN  F^.LI:= A;  NEW(F^.RE);
                                         CASE OP OF
      { Diese Knoten werden paarwei-}      SINUS  : F^.RE^.OPER:= COSIN;
      { se angelegt. Der Zeiger auf }      COSIN  : F^.RE^.OPER:= SINUS;
      { den Operanden ist jeweils LI}      SINUSH : F^.RE^.OPER:= COSINH;
      { der Zeiger auf den Partner- }      COSINH : F^.RE^.OPER:= SINUSH;
      { Knoten ist jeweils RE.      }    END;
                                         F^.RE^.STATUS:= 0; { FERTIG }
                                         F^.RE^.I := 0;
                                         F^.RE^.LI:= A;
                                         F^.RE^.RE:= F;
                                  END;
      UMINUS,QUAD,WURZ,EFKT,LOGA: F^.R:= A;
      HOCH:                       BEGIN
                                     F^.R   := A;
                                     F^.EXPO:= W;
                                  END;
      KONST:                      F^.WERT:= W;
   END;
   KNOTEN:= F;
END   {  KNOTEN  };

GLOBAL FUNCTION AVARIA( I,J : INTEGER ) : FORMEL;
{ Fuer abhaengige Variablen : NR ist der Index der Variablen }
VAR F : FORMEL;
BEGIN
   NEW( F );
   F^.OPER  := AVARIABLE;
   F^.STATUS:= 0 { FERTIG };
   F^.I     := I;
   F^.J     := J;
   AVARIA:= F;
END   {  VARIA  };

GLOBAL FUNCTION UVARIA : FORMEL;
{ Fuer unabhaengige Variablen }
VAR F : FORMEL;
BEGIN
   NEW( F );
   F^.OPER  := UVARIABLE;
   F^.STATUS:= 0 { FERTIG };
   UVARIA:= F;
END   {  VARIA  };

GLOBAL FUNCTION CON ( A : INTEGER  ) : FORMEL;
{ Fuer Konstanten : A ist deren Wert }
BEGIN
   CON:= KNOTEN( KONST, NIL, NIL, INTVAL(A) );
END;

GLOBAL FUNCTION CON ( A : INTERVAL ) : FORMEL;
{ Fuer Konstanten : A ist deren Wert }
BEGIN
   CON:= KNOTEN( KONST, NIL, NIL, A );
END;

GLOBAL OPERATOR - ( A : FORMEL ) UMIN : FORMEL;
{ unaeres Minus : A ist Zeiger auf den Operanden }
BEGIN
   IF  A^.OPER=KONST  THEN  UMIN:= CON( -A^.WERT )
                      ELSE  UMIN:= KNOTEN( UMINUS, A, NIL, INTVAL(0.0) );
END   {  UMIN  };

GLOBAL OPERATOR + ( A,B : FORMEL ) ADD : FORMEL;
{ Addition : A,B sind Zeiger auf die Operanden }
BEGIN
   IF  (A^.OPER=KONST)  AND  (A^.WERT=INTVAL(0.0))  THEN  ADD:= B  ELSE
   IF  (B^.OPER=KONST)  AND  (B^.WERT=INTVAL(0.0))  THEN  ADD:= A  ELSE
   IF  (A^.OPER=KONST)  AND  (B^.OPER=KONST)
                THEN    ADD:= CON( A^.WERT + B^.WERT )
                ELSE    ADD:= KNOTEN( PLUS, A, B, INTVAL(0.0) );
END   {  ADD  };

GLOBAL OPERATOR - ( A,B : FORMEL ) SUB : FORMEL;
{ Subtraktion : A,B sind Zeiger auf die Operanden }
BEGIN
   IF  (A^.OPER=KONST)  AND  (A^.WERT=INTVAL(0.0))  THEN  SUB:= -B  ELSE
   IF  (B^.OPER=KONST)  AND  (B^.WERT=INTVAL(0.0))  THEN  SUB:=  A  ELSE
   IF  (A^.OPER=KONST)  AND  (B^.OPER=KONST)
                THEN    SUB:= CON( A^.WERT - B^.WERT )
                ELSE    SUB:= KNOTEN( MINUS, A, B, INTVAL(0.0) );
END   {  FSUB  };

GLOBAL OPERATOR * ( A,B : FORMEL ) MUL : FORMEL;
{ Multiplikation : A,B sind Zeiger auf die Operanden }
BEGIN
   IF  (A^.OPER=KONST)  AND  (A^.WERT=INTVAL(0.0))  THEN  MUL:= CON(0) ELSE
   IF  (A^.OPER=KONST)  AND  (A^.WERT=INTVAL(1.0))  THEN  MUL:=   B    ELSE
   IF  (B^.OPER=KONST)  AND  (B^.WERT=INTVAL(0.0))  THEN  MUL:= CON(0) ELSE
   IF  (B^.OPER=KONST)  AND  (B^.WERT=INTVAL(1.0))  THEN  MUL:=   A    ELSE
   IF  (A^.OPER=KONST)  AND  (B^.OPER=KONST)
                THEN    MUL:= CON( A^.WERT * B^.WERT )
                ELSE    MUL:= KNOTEN( MAL, A, B, INTVAL(0.0) );
END   {  MUL  };

GLOBAL OPERATOR / ( A,B : FORMEL ) DIVI : FORMEL;
{ Division : A,B sind Zeiger auf die Operanden }
BEGIN
   IF  (A^.OPER=KONST)  AND  (A^.WERT=INTVAL(0.0))  THEN  DIVI:= CON(0) ELSE
   IF  (B^.OPER=KONST)  AND  (B^.WERT=INTVAL(1.0))  THEN  DIVI:=   A    ELSE
   IF  (A^.OPER=KONST)  AND  (B^.OPER=KONST)
                THEN    DIVI:= CON( A^.WERT / B^.WERT )
                ELSE    DIVI:= KNOTEN( DURCH, A, B, INTVAL(0.0) );
END   {  DIVI  };

GLOBAL FUNCTION SQR ( A : FORMEL ) : FORMEL;
{ Quadrat-Funktion : A ist Zeiger auf das Argument }
BEGIN
   IF  A^.OPER=KONST  THEN  SQR:= CON( SQR(A^.WERT) )
                      ELSE  SQR:= KNOTEN( QUAD, A, NIL, INTVAL(0.0) );
END   {  SQR  };

GLOBAL FUNCTION SQRT( A : FORMEL ) : FORMEL;
{ Quadratwurzel : A ist Zeiger auf den Radikanden }
BEGIN
   IF  A^.OPER=KONST  THEN  SQRT:= CON( SQRT(A^.WERT) )
                      ELSE  SQRT:= KNOTEN( WURZ, A, NIL, INTVAL(0.0) );
END   {  SQRT  };

GLOBAL OPERATOR ** ( A : FORMEL; B : FORMEL ) POT : FORMEL;
{ Potenz-Funktion : A ist der Zeiger auf die Basis, B der Zeiger auf den kon- 
{ stanten Exponenten.
}
BEGIN
   IF  B^.OPER<>KONST THEN  A^.WERT:= INTVAL(1/0) ELSE
   IF  A^.OPER=KONST  THEN  POT:= CON( EXP(B^.WERT*LN(A^.WERT)) )
                      ELSE  POT:= KNOTEN( HOCH, A, NIL, B^.WERT );
END   {  POT  };

GLOBAL FUNCTION EXP( A : FORMEL ) : FORMEL;
{ e-Funktion : A ist der Zeiger auf das Argument }
BEGIN
   IF  A^.OPER=KONST  THEN  EXP:= CON( EXP(A^.WERT) )
                      ELSE  EXP:= KNOTEN( EFKT, A, NIL, INTVAL(0.0) );
END   {  EXP  };

GLOBAL FUNCTION LN( A : FORMEL ) : FORMEL;
{ nat. Logarithmus : A ist der Zeiger auf das Argument }
BEGIN
   IF  A^.OPER=KONST  THEN  LN:= CON( LN(A^.WERT) )
                      ELSE  LN:= KNOTEN( LOGA, A, NIL, INTVAL(0.0) );
END   {  LN  };

GLOBAL FUNCTION SIN( A : FORMEL ) : FORMEL;
{ Sinus-Funktion : A ist der Zeiger auf das Argument }
BEGIN
   IF  A^.OPER=KONST  THEN  SIN:= CON( SIN(A^.WERT) )
                      ELSE  SIN:= KNOTEN( SINUS, A, NIL, INTVAL(0.0) );
END   {  SIN  };

GLOBAL FUNCTION COS( A : FORMEL ) : FORMEL;
{ Cosinus-Funktion : A ist der Zeiger auf das Argument }
BEGIN
   IF  A^.OPER=KONST  THEN  COS:= CON( COS(A^.WERT) )
                      ELSE  COS:= KNOTEN( COSIN, A, NIL, INTVAL(0.0) );
END   {  COS  };

GLOBAL FUNCTION TAN( A : FORMEL ) : FORMEL;
{ Tangens-Funktion : A ist der Zeiger auf das Argument }
BEGIN
   IF  A^.OPER=KONST  THEN  TAN:= CON( TAN(A^.WERT) )
                      ELSE  TAN:= KNOTEN( TANG, A, SQR(COS(A)), INTVAL(0.0) );
END   {  TAN  };

GLOBAL FUNCTION COT( A : FORMEL ) : FORMEL;
{ Cotangens-Funktion : A ist der Zeiger auf das Argument }
BEGIN
   IF  A^.OPER=KONST  THEN  COT:= CON( COT(A^.WERT) )
                      ELSE  COT:= KNOTEN( COTANG, A, SQR(SIN(A)), INTVAL(0.0) );
END   {  COT  };

GLOBAL FUNCTION SINH( A : FORMEL ) : FORMEL;
{ Hyperbolischer Sinus : A ist der Zeiger auf das Argument }
BEGIN
   IF  A^.OPER=KONST  THEN  SINH:= CON( SINH(A^.WERT) )
                      ELSE  SINH:= KNOTEN( SINUSH, A, NIL, INTVAL(0.0) );
END   {  SINH  };

GLOBAL FUNCTION COSH( A : FORMEL ) : FORMEL;
{ Hyperbolischer Cosinus : A ist der Zeiger auf das Argument }
BEGIN
   IF  A^.OPER=KONST  THEN  COSH:= CON( COSH(A^.WERT) )
                      ELSE  COSH:= KNOTEN( COSINH, A, NIL, INTVAL(0.0) );
END   {  COSH  };

GLOBAL FUNCTION TANH( A : FORMEL ) : FORMEL;
{ Hyperbolischer Tangens : A ist der Zeiger auf das Argument }
BEGIN
   IF  A^.OPER=KONST  THEN  TANH:= CON( TANH(A^.WERT) )
                      ELSE  TANH:= KNOTEN( TANGH, A, SQR(COSH(A)), INTVAL(0.0) );
END   {  TANH  };

GLOBAL FUNCTION COTH( A : FORMEL ) : FORMEL;
{ Hyperbolischer Cotangens : A ist der Zeiger auf das Argument }
BEGIN
   IF  A^.OPER=KONST  THEN  COTH:= CON( COTH(A^.WERT) )
                      ELSE  COTH:= KNOTEN( COTANGH, A, SQR(SINH(A)), INTVAL(0.0) );
END   {  COTH  };

GLOBAL FUNCTION ARCSIN( A : FORMEL ) : FORMEL;
{ Arcussinus : A ist der Zeiger auf das Argument }
BEGIN
   IF A^.OPER=KONST THEN ARCSIN:= CON( ARCSIN(A^.WERT) )
                    ELSE ARCSIN:= KNOTEN( ARCSI,A,SQRT(CON(1)-SQR(A)),INTVAL(0.0) );
END   {  ARCSIN  };

GLOBAL FUNCTION ARCCOS( A : FORMEL ) : FORMEL;
{ Arcustangens : A ist der Zeiger auf das Argument }
BEGIN
   IF A^.OPER=KONST THEN ARCCOS:= CON( ARCCOS(A^.WERT) )
                    ELSE ARCCOS:= KNOTEN( ARCCO,A,SQRT(CON(1)-SQR(A)),INTVAL(0.0) );
END   {  ARCCOS  };

GLOBAL FUNCTION ARCTAN( A : FORMEL ) : FORMEL;
{ Arcustangens : A ist der Zeiger auf das Argument }
BEGIN
   IF A^.OPER=KONST THEN ARCTAN:= CON( ARCTAN(A^.WERT) )
                    ELSE ARCTAN:= KNOTEN( ARCTG,A,CON(1)+SQR(A),INTVAL(0.0) );
END   {  ARCTAN  };

GLOBAL FUNCTION ARCCOT( A : FORMEL ) : FORMEL;
{ Arcuscotangens : A ist der Zeiger auf das Argument }
BEGIN
   IF A^.OPER=KONST THEN ARCCOT:= CON( ARCCOT(A^.WERT) )
                    ELSE ARCCOT:= KNOTEN( ARCCTG,A,CON(1)+SQR(A),INTVAL(0.0) );
END   {  ARCCOT  };

GLOBAL FUNCTION ARSINH( A : FORMEL ) : FORMEL;
{ Arcussinus : A ist der Zeiger auf das Argument }
BEGIN
   IF A^.OPER=KONST THEN ARSINH:= CON( ARSINH(A^.WERT) )
                    ELSE ARSINH:= KNOTEN( ARSIH,A,SQRT(SQR(A)+CON(1)),INTVAL(0.0) );
END   {  ARSINH  };

GLOBAL FUNCTION ARCOSH( A : FORMEL ) : FORMEL;
{ Arcustangens : A ist der Zeiger auf das Argument }
BEGIN
   IF A^.OPER=KONST THEN ARCOSH:= CON( ARCOSH(A^.WERT) )
                    ELSE ARCOSH:= KNOTEN( ARCOH,A,SQRT(SQR(A)-CON(1)),INTVAL(0.0) );
END   {  ARCOSH  };

GLOBAL FUNCTION ARTANH( A : FORMEL ) : FORMEL;
{ Arcustangens : A ist der Zeiger auf das Argument }
BEGIN
   IF A^.OPER=KONST THEN ARTANH:= CON( ARTANH(A^.WERT) )
                    ELSE ARTANH:= KNOTEN( ARTGH,A,CON(1)-SQR(A),INTVAL(0.0) );
END   {  ARTANH  };

GLOBAL FUNCTION ARCOTH( A : FORMEL ) : FORMEL;
{ Arcuscotangens : A ist der Zeiger auf das Argument }
BEGIN
   IF A^.OPER=KONST THEN ARCOTH:= CON( ARCOTH(A^.WERT) )
                    ELSE ARCOTH:= KNOTEN( ARCTGH,A,CON(1)-SQR(A),INTVAL(0.0) );
END   {  ARCOTH  };

GLOBAL FUNCTION FUNKT_MATRIX ( VAR F : FVECTOR; VAR ANZKNOTEN : INTEGER )
                      : FMATRIX[ lb(F)..ub(F),
                                 lb(F)..ub(F) ];
{ Berechnet die Funktionalmatrix von f formal  }
VAR I,J,ANZ                 : INTEGER;
    DF,FNULL,FEINS,FZWEI    : FORMEL; 
    ABLEITUNG               : DYNAMIC ARRAY [1..ANZKNOTEN] OF FORMEL; 

FUNCTION DIFF ( VAR F : FORMEL; K : INTEGER ) : FORMEL;
{ Fuehrt formale Differentiation der Formel,auf die F zeigt, nach u[i] durch.}
{ Diese partiellen Ableitungen werden jeweils in ABLEITUNG[I] eingetragen, um}
{ evtl. mehrfache Erzeugung identischer Teilbaeume zu umgehen.               }
VAR D : FORMEL;
BEGIN
   WITH  F^  DO
   CASE  OPER  OF
      KONST,UVARIABLE:    D:= FNULL;
      AVARIABLE: IF  I=K  THEN  D:= FEINS  ELSE  D:= FNULL;
      ELSE:     IF  ABLEITUNG[I]<>NIL  THEN  D:= ABLEITUNG[I]  ELSE
                BEGIN
                  CASE  OPER  OF
                   UMINUS: D:= - DIFF(R,K);
                   PLUS:   D:=  DIFF(LI,K) + DIFF(RE,K);
                   MINUS:  D:=  DIFF(LI,K) - DIFF(RE,K);
                   MAL:    D:=  DIFF(LI,K)*RE + LI*DIFF(RE,K);
                   DURCH:  D:=  ( DIFF(LI,K) - F*DIFF(RE,K) ) / RE;
                   QUAD:   D:=  FZWEI * DIFF(R,K) * R;
                   WURZ:   D:=  DIFF(R,K) / (FZWEI*F);
                   HOCH:   D:=  CON(EXPO)*DIFF(R,K)*(R**CON(EXPO-1));
                   EFKT:   D:=  DIFF(R,K)*F;
                   LOGA:   D:=  DIFF(R,K)/R;
                   COSIN:  D:= -DIFF(LI,K)*RE;
                   SINUS,
		   SINUSH,
		   COSINH: D:=  DIFF(LI,K)*RE;
                   TANG,
		   TANGH,
		   ARCSI,
		   ARCTG,
		   ARSIH,
		   ARCOH,
		   ARTGH,
		   ARCTGH: D:=  DIFF(LI,K)/RE;
		   COTANG,
		   COTANGH,
		   ARCCO,
		   ARCCTG: D:= -DIFF(LI,K)/RE;
                  END;
                  ABLEITUNG[I]:= D;
                END;
   END;
   DIFF:= D;
END  {  DIFF  };
 
BEGIN
   FNULL:= CON(0);
   FEINS:= CON(1);
   FZWEI:= CON(2);
   ANZ  := ANZKNOTEN;
   FOR J:= lb(F) TO ub(F) DO
   BEGIN
      FOR I:= 1 TO ANZ DO ABLEITUNG[I]:= NIL;
      FOR I:= lb(F) TO ub(F) DO
      BEGIN
         DF:= DIFF( F[I], J );
         COUNT_KNOTEN( DF, ANZKNOTEN );
         FUNKT_MATRIX[I,J]:= DF;
      END;
   END;
END  {  FUNKT_MATRIX  };
 

{}
{ --------------  ENDE   F K N O T E N . P A K    ------------------------ }

END.
