/**************************************************************************/
/*                                                                        */
/*                                 Sources                                */
/*                                from  the                               */
/*                           XSC SOFTWARE PACKAGE                         */
/*                                                                        */
/*                Copyright (c) 1997 University of Karlsruhe              */
/*                     Institute for Appl. Mathematics                    */
/*                                                                        */
/* This program is free software for non-commercial use. For details, see */
/* the file LICENSE.XSC, distributed with this software package.          */
/*                                                                        */
/* This program is distributed WITHOUT ANY WARRANTY. For details, see the */
/* file LICENSE.XSC, distributed with this software package.              */
/*                                                                        */
/**************************************************************************/

/* RCS $Id: a_div_.c,v 1.2 1997/09/25 16:51:08 ae59 Exp $ */
/*
$Log: a_div_.c,v $
 * Revision 1.2  1997/09/25  16:51:08  ae59
 * Copyright-message
 *
 * Revision 1.1  1994/01/27  13:20:43  ae59
 * Put under RCS control
 *
 */
/****************************************************************/
/*                                                              */
/*      Filename        : a_div_.c                              */
/*                                                              */
/*      Entries         : a_intg a_div_(i,j)                    */
/*                        a_intg i,j;                           */
/*                                                              */
/*      Arguments       : i = integer numerator                 */
/*                        j = integer denominator               */
/*                                                              */
/*      Description     : Integer division according to         */
/*                        PASCAL Standard:                      */
/*                           error if j is zero                 */
/*                 abs(i)-abs(j) < abs((i div j)*j) <= abs(j)   */
/*                 sign(i div j) = sign(i)*sign(j)              */
/*                                                              */
/*      Note            : 0 returned in case of error           */
/*                                                              */
/*      Date            : 1991-05-20                            */
/*      1991-05-20 : OVERFLOW detected if MININT/-1             */
/*      1990-11-15 : E_TPUSH+E_TPOPP                            */
/****************************************************************/

#ifndef ALL_IN_ONE
#ifdef AIX
#include "/u/p88c/runtime/o_defs.h"
#else
#include "o_defs.h"
#endif
#define local
#endif

#ifdef LINT_ARGS
local a_intg a_div_(a_intg i,a_intg j)
#else
local a_intg a_div_(i,j)

a_intg i;
a_intg j;
#endif
        {
        a_intg res;

        E_TPUSH("a_div_")

        if (j==0)
           {
           e_trap(DIV_BY_ZERO,4,E_TINT+E_TEXT(1),&i,
                                E_TINT+E_TEXT(2),&j);
           res = 0;
           }
        else
           {
           if (i>=0)
              {
              if (j>0)
                 res = i/j;
              else if (j==MININT)
                 res = 0;
              else
                 res = -(i/(-j));
              }
           else if (i==MININT)
              {
              if (j>0)
                 {
                 i += j;
                 res = -((-i)/j)-1;
                 }
              else if (j==MININT)
                 res = 1;
              else if (j==-1)
                 {
                 e_trap(OVERFLOW,4,E_TINT+E_TEXT(1),&i,
                                   E_TINT+E_TEXT(2),&j);
                 res = 0;
                 }
              else
                 {
                 i -= j;
                 res = (-i)/(-j)+1;
                 }
              }
           else
              {
              if (j>0)
                 res = -((-i)/j);
              else
                 res = (j==MININT) ? 0 : (-i)/(-j);
              }
           }

        E_TPOPP("a_div_")
        return(res);
        }

