/**************************************************************************/
/*                                                                        */
/*                                 Sources                                */
/*                                from  the                               */
/*                           XSC SOFTWARE PACKAGE                         */
/*                                                                        */
/*                Copyright (c) 1997 University of Karlsruhe              */
/*                     Institute for Appl. Mathematics                    */
/*                                                                        */
/* This program is free software for non-commercial use. For details, see */
/* the file LICENSE.XSC, distributed with this software package.          */
/*                                                                        */
/* This program is distributed WITHOUT ANY WARRANTY. For details, see the */
/* file LICENSE.XSC, distributed with this software package.              */
/*                                                                        */
/**************************************************************************/

/* RCS $Id: b_bcdi.c,v 1.2 1997/09/25 16:57:18 ae59 Exp $ */
/*
$Log: b_bcdi.c,v $
 * Revision 1.2  1997/09/25  16:57:18  ae59
 * Copyright-message
 *
 * Revision 1.1  1994/01/27  13:21:03  ae59
 * Put under RCS control
 *
 */
/************************************************************************/
/*                                                                      */
/* Descriptive Name : b_bcdi.c          Processor : C                   */
/*                                                                      */
/* Conversion of an IEEE double variable to intern representation.      */
/* The conversion is done without rounding.                             */
/*                                                                      */
/* Function value : int     0 - normalized number converted             */
/*                      ALLOC - allocation error                        */
/*                      DENOR - denormalized number converted           */
/*                      PINFI - +infinity     (no conversion!!!)        */
/*                      MINFI - -infinity     (no conversion!!!)        */
/*                      NANDE - NAN           (no conversion!!!)        */
/*                                                                      */
/*      Date            : 1991-04-23                            */
/*      1991-04-23 : MANT_INFINITY() used                       */
/*      1991-01-29 : B_ASHR() used                              */
/*      1991-01-08 : new argument 'rnd' / argument i is pointer */
/*      1990-10-30 : uh removed                                 */
/*      1990-10-26 : completely new                             */
/*      1990-07-16 : LINT_ARGS prototype                        */
/************************************************************************/

#ifndef ALL_IN_ONE
#ifdef AIX
#include "/u/p88c/runtime/o_defs.h"
#else
#include "o_defs.h"
#endif
#define local
#endif

#ifdef LINT_ARGS
local int b_bcdi(a_real d,multiprecision *i,a_intg rnd)
#else
local int b_bcdi(d,i,rnd)

a_real d;                 /* value of double variable             */
multiprecision *i;       /* pointer to intern variable           */
a_intg rnd;
#endif
        {
        a_btyp mant[D_U_RATIO+1];
        a_bool vz;
        a_intg expo;
        a_intg c;
        a_intg k,l;
/*----------------------------------------------------------------------*/
        (*i)->r = 0;    /* operation is exact                           */
/*----------------------------------------------------------------------*/
        if (b_deko(d,&expo,mant,&vz))
           {
           (*i)->z = 1;
           if ((*i)->l)
              {
              (*i)->l = 0;
#ifdef HEAP_CHECK
b_freh((a_char *)&(*i)->m,(a_char *)(*i)->m,(a_char *)"b_bcdi");
#endif
              B_FREE((*i)->m)
              }
           return(0);
           }            /* double value is zero                         */
/*----------------------------------------------------------------------*/
        (*i)->z = 0;    /* number value is non-zero                     */
/*----------------------------------------------------------------------*/
        (*i)->s = vz;   /* determine sign of double value               */
/*----------------------------------------------------------------------*/
        if (expo>EXPO_MAX)
           {
           if (MANT_INFINITY(mant)) return( (vz) ? MINFI : PINFI );
           return(NANDE);
           }            /* double variable holds a NAN or infinity      */
/*----------------------------------------------------------------------*/
        (*i)->e = B_ASHR(expo,LOG_B_LENGTH);
                           /* exponent of (*i)->m[0]                   */
        c = ((*i)->e<<LOG_B_LENGTH)-expo+(B_LENGTH-ZERO_BITS);
                           /* shift value                               */
/*----------------------------------------------------------------------*/
        if (c<0)
           b_shlu(mant,(l = D_U_RATIO),-c);
        else if (c>0)
           {
           mant[D_U_RATIO] = ZERO;
           b_shru(mant,(l = D_U_RATIO+1),c);
           }
        else
           l = D_U_RATIO;
                                /* shift mantissa */
/*----------------------------------------------------------------------*/
        for (k=0;mant[k]==ZERO;k++) (*i)->e--;
/*----------------------------------------------------------------------*/
        while (mant[l-1]==ZERO) l--;
/*----------------------------------------------------------------------*/
        b_badj(l-k,*i);   /* adjust intern variable                      */
/*----------------------------------------------------------------------*/
        for (l=0;l<(*i)->l;l++) (*i)->m[l] = mant[l+k];
                         /* copy mantissa digits                        */
/*----------------------------------------------------------------------*/
        return(0);
        }
