/**************************************************************************/
/*                                                                        */
/*                                 Sources                                */
/*                                from  the                               */
/*                           XSC SOFTWARE PACKAGE                         */
/*                                                                        */
/*                Copyright (c) 1997 University of Karlsruhe              */
/*                     Institute for Appl. Mathematics                    */
/*                                                                        */
/* This program is free software for non-commercial use. For details, see */
/* the file LICENSE.XSC, distributed with this software package.          */
/*                                                                        */
/* This program is distributed WITHOUT ANY WARRANTY. For details, see the */
/* file LICENSE.XSC, distributed with this software package.              */
/*                                                                        */
/**************************************************************************/

/* RCS $Id: b_bivp.c,v 1.2 1997/09/25 16:58:33 ae59 Exp $ */
/*
$Log: b_bivp.c,v $
 * Revision 1.2  1997/09/25  16:58:33  ae59
 * Copyright-message
 *
 * Revision 1.1  1994/01/27  13:21:08  ae59
 * Put under RCS control
 *
 */
/****************************************************************/
/*                                                              */
/*      Filename        : b_bivp.c                              */
/*                                                              */
/*      Entries         : a_btyp b_bivp(func,arg,rlb,rub)       */
/*                         int (*func)();                       */
/*                         a_real arg, *rlb, *rub;              */
/*                                                              */
/*      Arguments       : func = function to be evaluated for   */
/*                               a given real argument          */
/*                        arg = point argument of function      */
/*                        rlb = lower bound of result           */
/*                        rub = upper bound of result           */
/*                                                              */
/*      Function value  : function dependent return code.       */
/*                                                              */
/*      Description     : evaluate bounds of the function value */
/*                        for a given point argument            */
/*                                                              */
/*      Note            : Rounding is required if a_real is     */
/*                        not in a base 2 representation.       */
/*                                                              */
/*      Date            : 1991-07-31                            */
/*      1991-07-31 : restore b_maxl after conversion to real    */
/*      1991-01-08 : b_rtol() with rounding argument            */
/****************************************************************/

#ifndef ALL_IN_ONE
#ifdef AIX
#include "/u/p88c/runtime/o_defs.h"
#else
#include "o_defs.h"
#endif
#define local
extern a_btyp b_maxl;
#endif

#ifdef LINT_ARGS
local a_btyp b_bivp(int (*func)(multiprecision,multiprecision),
                      a_real arg, a_real *rlb, a_real *rub)
#else
local a_btyp b_bivp(func, arg, rlb, rub)

int (*func)();
a_real arg;
a_real *rlb;
a_real *rub;
#endif
        {
        a_intg code;               /* return code of long function */
        multiprecision l_arg, l_res;
        a_btyp rc;
        a_btyp Oldb_maxl;
        a_intg down = -1, up = 1;  /* rounding mode */

        l_init(&l_arg);
        l_init(&l_res);

        /* it is assumed that there is no conversion error      */
        /* from the current real format to multiprecision       */
        if (b_rtol(arg,&l_arg,(a_intg)0)) return(ALLOCATION);

        Oldb_maxl = b_maxl;
        b_maxl = 3;

        code = (*func)(l_arg,l_res);

        rc = b_ltor(l_res,rlb,down); /* lower bound */
        rc += b_ltor(l_res,rub,up);  /* upper bound */

        b_maxl = Oldb_maxl;

        l_free(&l_arg);
        l_free(&l_res);

        return((code)?(a_btyp)code:rc);
        }

