/**************************************************************************/
/*                                                                        */
/*                                 Sources                                */
/*                                from  the                               */
/*                           XSC SOFTWARE PACKAGE                         */
/*                                                                        */
/*                Copyright (c) 1997 University of Karlsruhe              */
/*                     Institute for Appl. Mathematics                    */
/*                                                                        */
/* This program is free software for non-commercial use. For details, see */
/* the file LICENSE.XSC, distributed with this software package.          */
/*                                                                        */
/* This program is distributed WITHOUT ANY WARRANTY. For details, see the */
/* file LICENSE.XSC, distributed with this software package.              */
/*                                                                        */
/**************************************************************************/

/* RCS $Id: b_cot_.c,v 1.2 1997/09/25 17:01:31 ae59 Exp $ */
/*
$Log: b_cot_.c,v $
 * Revision 1.2  1997/09/25  17:01:31  ae59
 * Copyright-message
 *
 * Revision 1.1  1994/01/27  13:21:19  ae59
 * Put under RCS control
 *
 */
/*************************************************************************
 *                                                                       *
 * Descriptive Name : Lcot               Processor : C                   *
 *                                                                       *
 * Cotangent Function for Multiple Precision Arithmetic                  *
 * ====================================================                  *
 *                                                                       *
 * Include files :  b_lari.h  - definitions for INTERN standard functions*
 *                                                                       *
 * Function value : int       - 0                                        *
 *                                                                       *
 * Argument Range :    All nonnegative numbers                           *
 * Result Range :      Nonnegative numbers                               *
 *                                                                       *
 * Used Functions :                                                      *
 *    sicovea         Computation of function value and error bound      *
 *    Lassign         Assignment of function value and number of ulp's   *
 *                                                                       *
 * Used Global INTERN Variables and Constants :                          *
 *    LhF, LhD        Used for debugging only                            *
 *                                                                       *
 * Used UNSIGNED Global Variables :                                      *
 *    Maxl            Length of INTERN Variables                         *
 *    Lcurrprec       Initial Length of INTERN Variables                 *
 *    Ldebug          Flag for printing additional information           *
 *    Lversion        Flag for printing additional information           *
 *                                                                       */
/*      Date            : 1991-09-09                            */
/*      1991-09-09 : reduce maximum line length                 */
/*      1990-09-04 : Larith.h replaced by b_lari.h              *
 *      1990-07-16 : LINT_ARGS prototype                        *
 *************************************************************************/

#define  Name        "Lcot"
#define  Main

#ifdef AIX
#include "/u/p88c/runtime/base/b_lari.h"
#else
#include "b_lari.h"
#endif

#define Lguard  2

static char  *function = Name;

/****************************************************************
 * Algorithm                                                    *
 ****************************************************************/

#ifdef LINT_ARGS
int b_cot_(dynamic *xi,dynamic *ri)
#else
int b_cot_(xi,ri)

dynamic *xi;
dynamic *ri;
#endif
#define  LRoutine    "Lcot"
{
   extern dynamic   LhF, LhD, LhE, Leps;
   extern a_btyp  Maxl;
   extern a_btyp  Lcurrprec;
   extern char      *Lroutine;

   int    rc;
#define poolvars  0

#ifdef Debug
   extern int       Ldebug;
#endif

#ifdef Debug
   Ldebug -= 1;                         /* Diminish Debug Level */
   if (Ldebug >= 0) printf("\n Entering Routine %s",LRoutine);
#endif



  /*----------------*
   | Initialization |
   *----------------*/

   Lroutine = function;
   Lcurrprec = Maxl;            /* Save precision setting       */
   rc = 0;


  /*----------------------------*
   | Check and Handle Case xi=0 |
   *----------------------------*/

   if (xi->z == 1) {
      ERREXIT(RANGE,RANGE,0);   /* Error message and handling   */
      }


  /*-------------------------------------*
   | Check and Handle Undefined Mantissa |
   *-------------------------------------*/

   if (xi->m[0] == 0) {
      ERREXIT(NANDE,NANDE,0);   /* Error message and handling   */
      }

   SICO_(xi);           /* Compute function value and error bound */

#ifdef Debug
   if (Ldebug >= 0) {
      printf("\n  (Lcot 130) LhF = ");  Lprinti(&LhF);
      printf("\n  (Lcot 131) LhD = ");  Lprinti(&LhD);
      }
#endif

   Maxl = Lcurrprec + Lguard;
   DIV_(&LhD,&LhF,&LhF);

   Leps.e = 1 - Maxl;
   Maxl = Minlerr;
   SHIFT_(1,&LhE,&LhE);
   NEXT_(&LhE,&LhE);
   ADD_(&LhE,&Leps,&LhE);
   NEXT_(&LhE,&LhE);

   if (rc != 0) {
      ERREXIT(RESUL,RESUL,0);   /* Error message and handling   */
      }

   ASSIGN_(ri);         /* Assign cotangent and set number of ulp's */

#ifdef Debug
   if (Ldebug >= 0) {
      printf("\n  (Lcot 153) cot = ");  Lprinti(ri);
      printf("   (+ %d ulp)",ri->r);
      }
#endif

   RETURN(0);
}
