/**************************************************************************/
/*                                                                        */
/*                                 Sources                                */
/*                                from  the                               */
/*                           XSC SOFTWARE PACKAGE                         */
/*                                                                        */
/*                Copyright (c) 1997 University of Karlsruhe              */
/*                     Institute for Appl. Mathematics                    */
/*                                                                        */
/* This program is free software for non-commercial use. For details, see */
/* the file LICENSE.XSC, distributed with this software package.          */
/*                                                                        */
/* This program is distributed WITHOUT ANY WARRANTY. For details, see the */
/* file LICENSE.XSC, distributed with this software package.              */
/*                                                                        */
/**************************************************************************/

/* RCS $Id: b_rndn.c,v 1.2 1997/09/25 17:06:12 ae59 Exp $ */
/*
$Log: b_rndn.c,v $
 * Revision 1.2  1997/09/25  17:06:12  ae59
 * Copyright-message
 *
 * Revision 1.1  1994/01/27  13:21:38  ae59
 * Put under RCS control
 *
 */
/****************************************************************/
/*                                                              */
/*      Filename        : b_rndn.c                              */
/*                                                              */
/*      Entries         : a_btyp b_rndn(lang,expo)              */
/*                        a_btyp *lang;                         */
/*                        a_intg *expo;                         */
/*                                                              */
/*      Arguments       : lang = result mantissa of length BSIZE*/
/*                        expo = exponent of result             */
/*                        vz = sign of result                   */
/*                                                              */
/*      Description     : Round to nearest IEEE number by       */
/*                        changing value of lang+D_U_RATIO-1    */
/*                                                              */
/*      Function value  : error code                            */
/*                        OVERFLOW - overflow occurred          */
/*                        INEXACT - inexact result              */
/*                                                              */
/*      Date            : 1991-11-11                            */
/*      1991-11-11 : set overflow flag                          */
/*      1991-05-28 : rounding made correct                      */
/*      1991-05-20 : no code changed but comments               */
/*      1991-05-02 : use status flags                           */
/*      1990-07-06 : LINT_ARGS prototype                        */
/*                   unnecessary e_sofo() removed               */
/****************************************************************/

#ifndef ALL_IN_ONE
#ifdef AIX
#include "/u/p88c/runtime/o_defs.h"
#else
#include "o_defs.h"
#endif
#define local
extern a_bool e_efie;
extern a_bool e_efof;
extern a_bool e_ofie;
extern a_bool e_ofof;
#endif

#ifdef LINT_ARGS
local a_btyp b_rndn(a_btyp *lang,a_intg *expo)
#else
local a_btyp b_rndn(lang,expo)

a_btyp *lang;
a_intg *expo;
#endif
        {
        a_intg i;

        /* msb of lang+D_U_RATIO is set */
        if (lang[D_U_RATIO] & MSB)
           {

           /* test trailing mantissa to be zero */
           if ((lang[D_U_RATIO] & NOT_MSB)==ZERO)
              {
              for (i=D_U_RATIO+1;i<BSIZE;i++)
                 if (lang[i]) break;

              /* return if lsb of lang+D_U_RATIO-1 is zero */
              if (i>=BSIZE)
                 {
                 if (!(lang[D_U_RATIO-1] & LSB))
                    {
                    if (e_efie) return(INEXACT);
                    e_ofie = TRUE;
                    return(NO_ERROR);
                    }
                 }
              }

           /* propagate carry */
           b_addc(lang+(D_U_RATIO-1));

           /* normalization (only msb is set in mantissa) */
           if (SHFT_MASK & *lang)
              {
              *lang = HIDDEN_BIT;
              (*expo)++;

              /* overflow occurred */
              if (*expo>EXPO_MAX)
                 {
                 if (e_efof)
                    {
                    *expo -= EXPO_ADJUST;
                    return(OVERFLOW);
                    }
                 e_ofof = TRUE;
                 }
              }
           if (e_efie) return(INEXACT);
           e_ofie = TRUE;
           return(NO_ERROR);
           }

        /* test for inexact data */
        for (i=D_U_RATIO;i<BSIZE;i++)
           {
           if (lang[i])
              {
              if (e_efie) return(INEXACT);
              e_ofie = TRUE;
              return(NO_ERROR);
              }
           }

        return(NO_ERROR);
        }
