/**************************************************************************/
/*                                                                        */
/*                                 Sources                                */
/*                                from  the                               */
/*                           XSC SOFTWARE PACKAGE                         */
/*                                                                        */
/*                Copyright (c) 1997 University of Karlsruhe              */
/*                     Institute for Appl. Mathematics                    */
/*                                                                        */
/* This program is free software for non-commercial use. For details, see */
/* the file LICENSE.XSC, distributed with this software package.          */
/*                                                                        */
/* This program is distributed WITHOUT ANY WARRANTY. For details, see the */
/* file LICENSE.XSC, distributed with this software package.              */
/*                                                                        */
/**************************************************************************/

/* RCS $Id: b_sin_.c,v 1.2 1997/09/25 17:07:17 ae59 Exp $ */
/*
$Log: b_sin_.c,v $
 * Revision 1.2  1997/09/25  17:07:17  ae59
 * Copyright-message
 *
 * Revision 1.1  1994/01/27  13:21:42  ae59
 * Put under RCS control
 *
 */
/*************************************************************************
 *                                                                       *
 * Descriptive Name : Lsin               Processor : C                   *
 *                                                                       *
 * Sine Function for Multiple Precision Arithmetic                       *
 * ===============================================                       *
 *                                                                       *
 * Include files :  b_lari.h  - definitions for INTERN standard functions*
 *                                                                       *
 * Function value : int       - 0                                        *
 *                                                                       *
 * Argument Range :    All nonnegative numbers                           *
 * Result Range :      Nonnegative numbers                               *
 *                                                                       *
 * Used Functions :                                                      *
 *    sicovea         Computation of function value and error bound      *
 *    Lassign         Assignment of function value and number of ulp's   *
 *                                                                       *
 * Used Global INTERN Variables and Constants :                          *
 *    LhF, LhE        Used for debugging only                            *
 *                                                                       *
 * Used UNSIGNED Global Variables :                                      *
 *    Maxl            Length of INTERN Variables                         *
 *    Lcurrprec       Initial Length of INTERN Variables                 *
 *    Ldebug          Flag for printing additional information           *
 *    Lversion        Flag for printing additional information           *
 *                                                                       */
/*      Date            : 1991-09-09                            */
/*      1991-09-09 : reduce max. line length                    */
/*      1991-08-01 : INT_HPREC introduced                       *
 *      1990-09-04 : Larith.h replaced by b_lari.h              *
 *      1990-07-30 : LhI inserted                               *
 *      1990-07-16 : LINT_ARGS prototype                        *
 *************************************************************************/

#define  Name        "Lsin"
#define  Main

#ifdef AIX
#include "/u/p88c/runtime/base/b_lari.h"
#else
#include "b_lari.h"
#endif

static char  *function = Name;

/****************************************************************
 * Algorithm                                                    *
 ****************************************************************/

#ifdef LINT_ARGS
int b_sin_(dynamic *xi,dynamic *ri)
#else
int b_sin_(xi,ri)

dynamic *xi;
dynamic *ri;
#endif
#define  LRoutine    "Lsin"
{
   extern a_btyp  Maxl;
   extern a_btyp  Lcurrprec;
   extern char      *Lroutine;
   extern a_btyp LhI;  /* entier(x/(pi/4)) mod 16 when calling sin(x) */

   int    rc;
#define poolvars  0

#ifdef Debug
   extern int       Ldebug;
   extern dynamic   LhF, LhE;
#endif

#ifdef Debug
   Ldebug -= 1;                            /* Diminish Debug Level      */
   if (Ldebug >= 0) printf("\n Entering Routine %s",LRoutine);
#endif

  /*----------------*
   | Initialization |
   *----------------*/

   Lroutine = function;
   Lcurrprec = Maxl;            /* Save precision setting               */
   rc = 0;
   LhI = 0;

  /*----------------------------*
   | Check and Handle Case xi=0 |
   *----------------------------*/

   if (xi->z == 1) {
      ri->z = 1;                /* ri = sin(0) = 0                      */
      ri->r = 0;
      EXIT(0);
      }


  /*-------------------------------------*
   | Check and Handle Undefined Mantissa |
   *-------------------------------------*/

   if (xi->m[0] == 0) {
      ERREXIT(NANDE,NANDE,0);   /* Error message and handling           */
      }

   SICO_(xi);           /* Compute function value and error bound */

#ifdef Debug
   if (Ldebug >= 0) {
      printf("\n  (Lsin 130) LhF = ");  Lprinti(&LhF);
      printf("\n  (Lsin 131) LhE = ");  Lprinti(&LhE);
      }
#endif

#if INT_HPREC
   b_farg = xi;
   b_case = LESS_ABS_ONE | LESS_ABS_ARG;
#endif

   ASSIGN_(ri);         /* Assign sine and set number of ulp's      */

#ifdef Debug
   if (Ldebug >= 0) {
      printf("\n  (Lsin 139) sin = ");  Lprinti(ri);
      printf("   (+ %d ulp)",ri->r);
      }
#endif

   RETURN(0);
}
