/**************************************************************************/
/*                                                                        */
/*                                 Sources                                */
/*                                from  the                               */
/*                           XSC SOFTWARE PACKAGE                         */
/*                                                                        */
/*                Copyright (c) 1997 University of Karlsruhe              */
/*                     Institute for Appl. Mathematics                    */
/*                                                                        */
/* This program is free software for non-commercial use. For details, see */
/* the file LICENSE.XSC, distributed with this software package.          */
/*                                                                        */
/* This program is distributed WITHOUT ANY WARRANTY. For details, see the */
/* file LICENSE.XSC, distributed with this software package.              */
/*                                                                        */
/**************************************************************************/

/* RCS $Id: b_tan_.c,v 1.2 1997/09/25 17:08:37 ae59 Exp $ */
/*
$Log: b_tan_.c,v $
 * Revision 1.2  1997/09/25  17:08:37  ae59
 * Copyright-message
 *
 * Revision 1.1  1994/01/27  13:21:47  ae59
 * Put under RCS control
 *
 */
/*************************************************************************
 *                                                                       *
 * Descriptive Name : Ltan               Processor : C                   *
 *                                                                       *
 * Tangent Function for Multiple Precision Arithmetic                    *
 * ==================================================                    *
 *                                                                       *
 * Include files :  b_lari.h  - definitions for INTERN standard functions*
 *                                                                       *
 * Function value : int       - 0                                        *
 *                                                                       *
 * Argument Range :    All numbers                                       *
 * Result Range :      All numbers                                       *
 *                                                                       *
 * Used Functions :                                                      *
 *    sicovea         Computation of function value and error bound      *
 *    Lassign         Assignment of function value and number of ulp's   *
 *                                                                       *
 * Used Global INTERN Variables and Constants :                          *
 *    LhF, LhD        Used for debugging only                            *
 *                                                                       *
 * Used UNSIGNED Global Variables :                                      *
 *    Maxl            Length of INTERN Variables                         *
 *    Lcurrprec       Initial Length of INTERN Variables                 *
 *    Ldebug          Flag for printing additional information           *
 *    Lversion        Flag for printing additional information           *
 *                                                                       */
/*      Date            : 1991-09-10                            */
/*      1991-09-10 : reduced max. line length                   */
/*      1991-08-01 : INT_HPREC introduced                       *
 *      1991-04-11 : LhI initialized                            *
 *      1990-09-04 : Larith.h replaced by b_lari.h              *
 *      1990-07-16 : LINT_ARGS prototype                        *
 *************************************************************************/

#define  Name        "Ltan"
#define  Main

#ifdef AIX
#include "/u/p88c/runtime/base/b_lari.h"
#else
#include "b_lari.h"
#endif

#define Lguard  2

static char  *function = Name;

/****************************************************************
 * Algorithm                                                    *
 ****************************************************************/

#ifdef LINT_ARGS
int b_tan_(dynamic *xi,dynamic *ri)
#else
int b_tan_(xi,ri)

dynamic *xi;
dynamic *ri;
#endif
#define  LRoutine    "Ltan"
{
#if INT_HPREC
   extern dynamic Lone;
#endif
   extern dynamic   LhF, LhD, LhE, Leps;
   extern a_btyp  Maxl;
   extern a_btyp  Lcurrprec;
   extern char      *Lroutine;
   extern a_btyp LhI;  /* entier(x/(pi/4)) mod 16 when calling sin(x) */

   int    rc;
#define poolvars  0

#ifdef Debug
   extern int       Ldebug;
#endif

#ifdef Debug
   Ldebug -= 1;                 /* Diminish Debug Level          */
   if (Ldebug >= 0) printf("\n Entering Routine %s",LRoutine);
#endif



  /*----------------*
   | Initialization |
   *----------------*/

   Lroutine = function;
   Lcurrprec = Maxl;            /* Save precision setting       */
   rc = 0;
   LhI = 0;

  /*----------------------------*
   | Check and Handle Case xi=0 |
   *----------------------------*/

   if (xi->z == 1) {
      ri->z = 1;                /* ri = tan(0) = 0              */
      ri->r = 0;
      EXIT(0);
      }


  /*-------------------------------------*
   | Check and Handle Undefined Mantissa |
   *-------------------------------------*/

   if (xi->m[0] == 0) {
      ERREXIT(NANDE,NANDE,0);   /* Error message and handling   */
      }

   SICO_(xi);           /* Compute function value and error bound   */

#ifdef Debug
   if (Ldebug >= 0) {
      printf("\n  (Ltan 132) LhF = ");  Lprinti(&LhF);
      printf("\n  (Ltan 133) LhD = ");  Lprinti(&LhD);
      }
#endif

   Maxl = Lcurrprec + Lguard;
   DIV_(&LhF,&LhD,&LhF);

   Leps.e = 1 - Maxl;
   Maxl = Minlerr;
   SHIFT_(1,&LhE,&LhE);
   NEXT_(&LhE,&LhE);
   ADD_(&LhE,&Leps,&LhE);
   NEXT_(&LhE,&LhE);

   if (rc != 0) {
      ERREXIT(RESUL,RESUL,0);   /* Error message and handling   */
      }

#if INT_HPREC
   if (LT_ABS_(xi,&Lone))
      {
      b_farg = xi;
      b_case = GREATER_ABS_ARG;
      }
#endif

   ASSIGN_(ri);         /* Assign tangent and set number of ulp's   */

#ifdef Debug
   if (Ldebug >= 0) {
      printf("\n  (Ltan 155) tan = ");  Lprinti(ri);
      printf("   (+ %d ulp)",ri->r);
      }
#endif

   RETURN(0);
}
