/**************************************************************************/
/*                                                                        */
/*                                 Sources                                */
/*                                from  the                               */
/*                           XSC SOFTWARE PACKAGE                         */
/*                                                                        */
/*                Copyright (c) 1997 University of Karlsruhe              */
/*                     Institute for Appl. Mathematics                    */
/*                                                                        */
/* This program is free software for non-commercial use. For details, see */
/* the file LICENSE.XSC, distributed with this software package.          */
/*                                                                        */
/* This program is distributed WITHOUT ANY WARRANTY. For details, see the */
/* file LICENSE.XSC, distributed with this software package.              */
/*                                                                        */
/**************************************************************************/

/* RCS $Id: b_tdiv.c,v 1.2 1997/09/25 17:09:05 ae59 Exp $ */
/*
$Log: b_tdiv.c,v $
 * Revision 1.2  1997/09/25  17:09:05  ae59
 * Copyright-message
 *
 * Revision 1.1  1994/01/27  13:21:49  ae59
 * Put under RCS control
 *
 */
/****************************************************************/
/*                                                              */
/*      Filename        : b_tdiv.c                              */
/*                                                              */
/*      Entries         : int  b_tdiv(a,b,res)                  */
/*                        tenbyte *a,*b,*res;                   */
/*                                                              */
/*      Arguments       : a = numerator                         */
/*                        b = denominator                       */
/*                        res = quotient                        */
/*                                                              */
/*      Description     : Division of numbers.                  */
/*                        Result is rounded according to        */
/*                        flag "b_rflg".                        */
/*                                                              */
/*      Date            : 1991-06-20                            */
/*      1991-06-20 : new routine                                */
/****************************************************************/

#ifndef ALL_IN_ONE
#ifdef AIX
#include "/u/p88c/runtime/o_defs.h"
#else
#include "o_defs.h"
#endif
#define local
#endif

#ifdef LINT_ARGS
local int b_tdiv(tenbyte *a,tenbyte *b,tenbyte *res)
#else
local int b_tdiv(a,b,res)

tenbyte *a;
tenbyte *b;
tenbyte *res;
#endif
        {
        a_btyp manta[BSIZE];
        a_btyp mantb[BSIZE];
        a_btyp mantc[BSIZE];
        a_intg expoa,expob,expoc,i;
        a_bool  vza,vzb,vzc,zeroa,zerob;

        E_TPUSH("b_tdiv")

        /* initialize unused result positions                   */
        for (i=D_U_RATIO+1;i<BSIZE;i++) mantc[i] = ZERO;

        /* decomposition                                        */
        zeroa = b_tdek(a,&expoa,manta,&vza);
        zerob = b_tdek(b,&expob,mantb,&vzb);

        /* sign of quotient                                     */
        vzc = vza^vzb;

        /* numerator is zero                                    */
        if ( zeroa )
        {
           /* zero/zero                        */
           if (zerob) {
              e_trap(INV_OP+E_IEEE,6,E_TMSG,2,E_TDBL,&a,E_TDBL,&b);

              E_TPOPP("b_tdiv");
              return(1);
              }

           expoc = -tCHARAC;
           b_tcom(res,expoa,manta,vzc);

           E_TPOPP("b_tdiv")
           return(0);
           }

        /* denominator is zero                                  */
        if ( zerob ) {
            e_trap(DIV_BY_ZERO+E_IEEE,4,E_TDBL,&a,E_TDBL,&b);

            E_TPOPP("b_tdiv");
            return(1);
            }

        /* denormalized numbers                                 */
        while (!(tFIRST_BIT & manta[0]))
                {
                b_shl1(manta,BSIZE);
                expoa--;
                }
        while (!(tFIRST_BIT & mantb[0]))
                {
                b_shl1(mantb,BSIZE);
                expob--;
                }

        /* exponent                                             */
        expoc = expoa-expob;

        /* division of mantissas                                */
        b_tmdv(manta,mantb,mantc,&expoc);

        /* adjust denormalized number                           */
        b_tadj(mantc,&expoc);

        /* rounding                                             */
        b_trnd(mantc,&expoc,vzc);

        /* composition                                          */
        b_tcom(res,expoc,mantc,vzc);

        E_TPOPP("b_tdiv")
        return(0);
        }
