/**************************************************************************/
/*                                                                        */
/*                                 Sources                                */
/*                                from  the                               */
/*                           XSC SOFTWARE PACKAGE                         */
/*                                                                        */
/*                Copyright (c) 1997 University of Karlsruhe              */
/*                     Institute for Appl. Mathematics                    */
/*                                                                        */
/* This program is free software for non-commercial use. For details, see */
/* the file LICENSE.XSC, distributed with this software package.          */
/*                                                                        */
/* This program is distributed WITHOUT ANY WARRANTY. For details, see the */
/* file LICENSE.XSC, distributed with this software package.              */
/*                                                                        */
/**************************************************************************/

/* RCS $Id: d_stan.c,v 1.2 1997/09/25 17:14:59 ae59 Exp $ */
/*
$Log: d_stan.c,v $
 * Revision 1.2  1997/09/25  17:14:59  ae59
 * Copyright-message
 *
 * Revision 1.1  1994/01/27  13:22:09  ae59
 * Put under RCS control
 *
 */
/****************************************************************/
/*                                                              */
/*      Filename        : d_stan.c                              */
/*                                                              */
/*      Entries         : a_real d_stan(a)                      */
/*                        d_otpr a;                             */
/*                                                              */
/*      Arguments       : a = dotprecision variable             */
/*                                                              */
/*      Description     : Round dotprecision variable to        */
/*                        nearest IEEE double value according   */
/*                        to IEEE rounding conventions          */
/*                                                              */
/*      Date            : 1991-04-22                            */
/*      1991-04-22 : real assignment replaced by R_ASSIGN()     */
/*      1991-01-07 : SET_QUIET used / union q removed           */
/*      1990-11-12 : E_TPUSH+E_TPOPP                            */
/*      1990-10-08 : consider A_MZERO,A_PZERO                   */
/****************************************************************/

#ifndef ALL_IN_ONE
#ifdef AIX
#include "/u/p88c/runtime/o_defs.h"
#else
#include "o_defs.h"
#endif
#define local
extern a_real *r_sero;
extern a_real *r_zero;
#endif

#ifdef LINT_ARGS
local a_real d_stan(d_otpr a)
#else
local a_real d_stan(a)

d_otpr a;
#endif
        {
        a_btyp rc,lang[BSIZE];
        a_intg expo;
        a_bool vz;
        a_real res;

        E_TPUSH("d_stan")

        if (a[A_STATUS] & A_QUIETNAN)
           {
           ((a_btyp *)&res)[B_HPART] = SET_QUIET(EXPO_MASK);
           ((a_btyp *)&res)[B_LPART] = a[A_LOWNAN];
           }
        else if (a[A_STATUS] & A_PINFINITY)
           {
           ((a_btyp *)&res)[B_HPART] = EXPO_MASK;
           ((a_btyp *)&res)[B_LPART] = ZERO;
           }
        else if (a[A_STATUS] & A_MINFINITY)
           {
           ((a_btyp *)&res)[B_HPART] = EXPO_MASK | MSB;
           ((a_btyp *)&res)[B_LPART] = ZERO;
           }

        /* get accu value in IEEE length                        */
        else if (b_geta( a, lang, &expo, &vz ))
           {
           R_ASSIGN(res,
              ((a[A_STATUS] & A_MZERO)==ZERO) ? *r_sero : *r_zero);
           }
        else
           {

           /* adjust denormalized number                        */
           if ( (rc = b_adj( lang, &expo )) ==NO_ERROR)
              rc = b_rndn( lang, &expo );
           else
              (void)b_rndn( lang, &expo );

           /* composition                                       */
           b_comp( &res, expo, lang, vz );

           if (rc) e_trap(rc+E_IEEE,2,E_TDTP,&a);
           }

        /* free temporary dotprecison variables                 */
        if (a[A_STATUS] & A_TEMPORARY) d_free(&a);

        E_TPOPP("d_stan")
        return(res);
        }
