/**************************************************************************/
/*                                                                        */
/*                                 Sources                                */
/*                                from  the                               */
/*                           XSC SOFTWARE PACKAGE                         */
/*                                                                        */
/*                Copyright (c) 1997 University of Karlsruhe              */
/*                     Institute for Appl. Mathematics                    */
/*                                                                        */
/* This program is free software for non-commercial use. For details, see */
/* the file LICENSE.XSC, distributed with this software package.          */
/*                                                                        */
/* This program is distributed WITHOUT ANY WARRANTY. For details, see the */
/* file LICENSE.XSC, distributed with this software package.              */
/*                                                                        */
/**************************************************************************/

/* RCS $Id: f_rint.c,v 1.2 1997/09/25 17:21:50 ae59 Exp $ */
/*
$Log: f_rint.c,v $
 * Revision 1.2  1997/09/25  17:21:50  ae59
 * Copyright-message
 *
 * Revision 1.1  1994/01/27  13:22:35  ae59
 * Put under RCS control
 *
 */
/****************************************************************/
/*                                                              */
/*      Filename        : f_rint.c                              */
/*                                                              */
/*      Entry           : a_intg f_rint(device,i)               */
/*                        FILE *device;                         */
/*                        a_intg *i;                            */
/*                                                              */
/*      Arguments       : device - text file to be read         */
/*                        i - integer value read                */
/*                            at input: i holds window character*/
/*                                                              */
/*      Return value    : last character read                   */
/*                                                              */
/*      Description     : perform PASCAL integer read.          */
/*                                                              */
/*      Note            : BS PASCAL Standard 6.9.1 c)           */
/*                                                              */
/*      Date            : 1993-02-09                            */
/*      1993-02-09 : overflow checking changed for MAXINt =b=   */
/*      1991-02-25 : overflow checking changed                  */
/*      1990-11-15 : E_TPUSH+E_TPOPP                            */
/*      1990-09-05 : character ch passed to e_trap instead of c */
/*      1990-08-06 : type of c changed to int                   */
/****************************************************************/

#ifndef ALL_IN_ONE
#ifdef AIX
#include "/u/p88c/runtime/o_defs.h"
#else
#include "o_defs.h"
#endif
#define local
#endif

#ifdef LINT_ARGS
local a_intg f_rint(FILE *device,a_intg *i)
#else
local a_intg f_rint(device,i)

FILE *device;
a_intg *i;
#endif
        {
        int c;
        a_bool sign;
        char ch;

        E_TPUSH("f_rint")

        sign = FALSE;
        c = (int)*i;

        /* ignore leading blanks and linefeed characters        */
        while (c==' ' || c==EOLN) c = fgetc(device);

        /* check for sign                                       */
        if (c=='+') c = fgetc(device);
        else if (c=='-')
           {
           c = fgetc(device);
           sign = TRUE;
           }

        /* handle first decimal digit                           */
        if (!isdigit(c))
           {
           if (c==EOF)
              {
              e_trap(I_O_ERROR,2,E_TMSG,20);
              }
           else
              {
              ungetc(c,device);
              ch = c;
              e_trap(I_O_ERROR,4,E_TMSG,21,E_TCHR+E_TEXT(10),&ch);
              }
           E_TPOPP("f_rint")
           return(c);
           }

        *i = c-'0';
        c = fgetc(device);

        /* read decimal digits                                  */
        while (isdigit(c))
           {

           /* check for integer overflow                        */
           if (MAXINT/10-*i<=0)
              {
              if ((MAXINT/10-*i<0) || (c-'0'>MAXINT%10))
                 {
                 if (sign) *i = -*i;
                 e_trap(OVERFLOW,2,E_TMSG,15);
                 do c = fgetc(device);
                 while (isdigit(c));
                 if (c!=EOF) ungetc(c,device);
                 E_TPOPP("f_rint")
                 return(c);
                 }
              }

           /* add digit                                         */
           *i = (*i*10)+c-'0';
           c = fgetc(device);
           }

        if (sign) *i = -*i;

        if (c!=EOF) ungetc(c,device);

        E_TPOPP("f_rint")
        return(c);
        }
