/**************************************************************************/
/*                                                                        */
/*                                 Sources                                */
/*                                from  the                               */
/*                           XSC SOFTWARE PACKAGE                         */
/*                                                                        */
/*                Copyright (c) 1997 University of Karlsruhe              */
/*                     Institute for Appl. Mathematics                    */
/*                                                                        */
/* This program is free software for non-commercial use. For details, see */
/* the file LICENSE.XSC, distributed with this software package.          */
/*                                                                        */
/* This program is distributed WITHOUT ANY WARRANTY. For details, see the */
/* file LICENSE.XSC, distributed with this software package.              */
/*                                                                        */
/**************************************************************************/

/* RCS $Id: r_mant.c,v 1.2 1997/09/25 17:48:12 ae59 Exp $ */
/*
$Log: r_mant.c,v $
 * Revision 1.2  1997/09/25  17:48:12  ae59
 * Copyright-message
 *
 * Revision 1.1  1994/01/27  13:24:05  ae59
 * Put under RCS control
 *
 */
/****************************************************************/
/*                                                              */
/*      Filename        : r_mant.c                              */
/*                                                              */
/*      Entries         : real r_mant(m)                        */
/*                        real m;                               */
/*                                                              */
/*      Arguments       : m = mantissa                          */
/*                                                              */
/*      Description     : Mantissa of IEEE value.               */
/*                        0.5 <= |mantissa| < 1.0               */
/*                                                              */
/*      Note            : m is returned if r==0.0,qNaN, or      */
/*                        +/- infinity.                         */
/*                        sNaN causes an INV_OP exception       */
/*                                                              */
/*      Date            : 1992-11-23                            */
/*      1992-11-23 : add E_ieee for e_trap(INV_op ... =b=       */
/*      1991-05-20 : new normalization                          */
/*      1991-04-23 : MANT_INFINITY() used                       */
/*      1991-02-27 : e_trap() arguments changed / only if sNaN  */
/*      1991-01-31 : Changes made according to Note.            */
/*      1990-11-19 : E_TPUSH+E_TPOPP                            */
/*      1990-09-06 : only one mantissa range                    */
/*      1990-09-03 : mantissa ranges changed                    */
/*      1990-07-16 : 0.0 returned if argument is 0.0            */
/****************************************************************/

#ifndef ALL_IN_ONE
#ifdef AIX
#include "/u/p88c/runtime/o_defs.h"
#else
#include "o_defs.h"
#endif
#define local
#endif

#ifdef LINT_ARGS
local a_real r_mant(a_real m)
#else
local a_real r_mant(m)

a_real m;
#endif
        {
        a_bool vz;
        a_intg expo;
        a_btyp mant[D_U_RATIO];

        E_TPUSH("r_mant")

        if (b_deko(m,&expo,mant,&vz))
           {
           }

        /* Nan or infinity                              */
        else if (expo>EXPO_MAX)
           {
           if (!MANT_INFINITY(mant))
           if (SIGNALING(mant[0]))
              e_trap(INV_OP+E_IEEE,4,E_TMSG,5,E_TDBL+E_TEXT(7),&m);
           }

        else
           {

           /* denormalized IEEE value */
           if ((mant[0] & HIDDEN_BIT)==ZERO)
              {
              do {
                 b_shl1(mant,D_U_RATIO);
                 }
              while((mant[0] & HIDDEN_BIT)==ZERO);
              }

           b_comp(&m,(a_intg)-1,mant,vz);
           }

        E_TPOPP("r_mant")
        return(m);
        }

