/**************************************************************************/
/*                                                                        */
/*                                 Sources                                */
/*                                from  the                               */
/*                           XSC SOFTWARE PACKAGE                         */
/*                                                                        */
/*                Copyright (c) 1997 University of Karlsruhe              */
/*                     Institute for Appl. Mathematics                    */
/*                                                                        */
/* This program is free software for non-commercial use. For details, see */
/* the file LICENSE.XSC, distributed with this software package.          */
/*                                                                        */
/* This program is distributed WITHOUT ANY WARRANTY. For details, see the */
/* file LICENSE.XSC, distributed with this software package.              */
/*                                                                        */
/**************************************************************************/

/* RCS $Id: t_sinh.c,v 1.2 1997/09/25 18:22:26 ae59 Exp $ */
/*
$Log: t_sinh.c,v $
 * Revision 1.2  1997/09/25  18:22:26  ae59
 * Copyright-message
 *
 * Revision 1.1  1994/01/27  13:25:47  ae59
 * Put under RCS control
 *
 */
/****************************************************************/
/*                                                              */
/*      Filename        : t_sinh.c                              */
/*                                                              */
/*      Entries         : a_real t_sinh(arg)                    */
/*                        a_real arg;                           */
/*                                                              */
/*      Arguments       : arg  = argument of hyp. sine          */
/*                                                              */
/*      Description     : Hyperbolic sine function.             */
/*                                                              */
/*      Date            : 1992-07-23                            */
/*      1992-07-23 : add  SUN4_CPP_c =b=                        */
/*      1992-05-11 : changed "10" to "sizeof(ExtReal)"          */
/*      1991-11-14 : save and restore flags                     */
/*      1991-09-24 : macro call StdFctReal replaced             */
/*      1991-09-17 : remove semicolon after makro               */
/*      1991-08-05 : remove unused variable ret                 */
/*      1991-06-21 : new routine                                */
/****************************************************************/

#ifdef AIX
#include "/u/p88c/runtime/tbyte/t_defs.h"
#else
#include "t_defs.h"
#endif

/* #include "sinh.h"       */
/* Teilpunkt               */
/* +1.447715759277344E-001 */
#define Tp                                  \
           EXTREAL(0x3F, 0xFC,              \
                   0x94, 0x3F, 0x00, 0x00,  \
                   0x00, 0x00, 0x00, 0x00)

/* StdFctReal(t_sinh,sinhee) */
#ifdef LINT_ARGS
a_real t_sinh(a_real arg)
#else
a_real t_sinh(arg)

a_real arg;
#endif
        {
        int      rnd, rc;
        a_real   res;
        ExtReal  a, r;

        E_SPUSH("t_sinh")

        rnd = getrndmode();
        longreal_to_extreal((LongReal *)&arg, &a);

        if ((rc = sinhee(&a, &r))!=0
            || (rc = extreal_to_longreal(&r, (LongReal *)&res))!=0)
           ieee_abortr1(rc, &arg);

        setrndmode(rnd);

        E_SPOPP("t_sinh")
        return res;
        }
/* ------------------------------------------------------------ */

/*--------------------------------------------------------------*
 | sinh                                                         |
 *--------------------------------------------------------------*/
#ifdef ANSI_C
#ifdef LINT_ARGS
int sinhee(const ExtReal *arg, ExtReal *res)
#else
int sinhee(arg, res)
const ExtReal   *arg;
      ExtReal   *res;
#endif /* LINT_ARGS */
#else  /* NOT ANSI_C */
#ifdef LINT_ARGS
int sinhee(ExtReal *arg, ExtReal *res)
#else
int sinhee(arg, res)
ExtReal   *arg;
ExtReal   *res;
#endif /* LINT_ARGS */
#endif /* ANSI_C */
{
#ifdef ANSI_C
#if SUN4_CPP_C
   static ExtReal tp;
   static char tpc[sizeof(ExtReal)] = Tp;
#else
   static const ExtReal tp = Tp; /* Teilpunkt   */
#endif
#else
   static ExtReal tp;
   static char tpc[sizeof(ExtReal)] = Tp;
#endif
   ExtReal     a;          /* Absolutes Argument                */
   ExtReal     e;          /* Ergebnis Exp                      */
   ExtReal     r;          /* Ergebnis Summe                    */
   ExtReal     ep;         /* Ergebnis positiver Exponent       */
   ExtReal     en;         /* Ergebnis negativer Exponent       */
   ExtReal     n;          /* ZwischenErgebnis Nenner           */
   ExtReal     q;          /* ZwischenErgebnis Quotient         */
   int         sgn;        /* Vorzeichen Argument               */
   int         rnd;        /* RundungsMode                      */
   int         check;      /* Rueckgabe von Makro ArgCheck      */

#ifndef ANSI_C
   memcpy(&tp, tpc, sizeof(tp));
#else
#if SUN4_CPP_C
   memcpy(&tp, tpc, sizeof(tp));
#endif
#endif

   /* --- pruefe Argument, dann Pruefung aus --- */
   ArgCheck1(Sinh, arg, res);
   arg_check = Off;

   /* --- RundungsMode sichern, NEAR setzen --- */
   rnd  = getrndmode();
   setrndmode(NEAR);

   /* --- a := |arg| --- */
   sgn = SGNE(arg);
   absee(arg, &a);

   /* --- arg > Teilungspunkt --- */
   if(1==cmpee(&a, &tp)) {
      expee(&a, &ep);
      divee(&One, &ep, &en);
      subee(&ep, &en, &r);
   }
   else {
      expm1ee(&a, &e);
      addee(&e, &One, &n);
      divee(&e, &n, &q);
      addee(&e, &q, &r);
   }
   scaliee(&r, -1, res);         /* /2                          */

   /* --- if arg < 0 ==> res := -res --- */
   if(NEG==sgn)
      chsee(res, res);

   /* --- RundungsMode zuruecksetzen --- */
   setrndmode(rnd);
   arg_check = On;

   return NoErr;
} /* sinhee() */
