/**********************************************************************/
/*                                                                    */
/* "DP_PRINT.C"                                                       */
/*        Version from 09-02-1991                                     */
/*        A.Just (Nemesis Software)                                   */
/*                                                                    */
/* This file contains routines for output of work trees.              */
/*                                                                    */
/* 09-02-91 (Just): Comments added,                                   */
/*                  'd_vblkdump()': 'varinfo' -> 'varblk'             */
/* 31.03.95  PJ     NEW_PROTO replaced by LINT_ARGS                   */
/*                                                                    */
/**********************************************************************/



/* #define TRACE_ON */



#include <stdio.h>

#include "impor.t"
#include        "error.h"
#include        "scan.h"
#include        "symtab.h"
#include        "idtab.h"
#include        "blktab.h"
#include        "optab.h"
#include        "bintab.h"
#include        "arrx.h"

#include        "dp_nodes.h"
#include        "dp_error.h"
#include "expor.t"
#include        "dp_print.h"
#include "implemen.t"



/***************************************************************************/



#define SPACES "    "    /* spaces to be inserted with depth          */



/***************************************************************************/



#ifdef LINT_ARGS
     static void d_wnodeprint ( tNODE*, int );
     static void d_wprintname ( tNODE* );
     static void d_printvar   ( binptr );
     static void d_vblkdump   ( tVARBLK* );
#else
     static void d_wnodeprint ();
     static void d_wprintname ();
     static void d_printvar   ();
     static void d_vblkdump   ();
#endif



/**********************************************************************/



static int vblk_dump;



/**********************************************************************/



/* PJ 31.03.95  NEW_PROTO replaced by LINT_ARGS */

#ifdef LINT_ARGS
void d_workprint(tNODE *ptr,int depth)
#else
void d_workprint(ptr,depth)
tNODE     *ptr;
int       depth;
#endif

/*--------------------------------------------------------------------*/
/* print work tree to standart output                                 */
/*--------------------------------------------------------------------*/
/* ptr   = pointer to the work tree                                   */
/* depth = depth of subtree in whole tree                             */
/*--------------------------------------------------------------------*/

{
     /*---------------------------------------------------------------*/
#ifdef TRACE_ON
          d_trace("'d_workprint()' :");
#endif
     /*---------------------------------------------------------------*/



     /*---------------------------------------------------------------*/
     /* Set flag, to decide wether making a VarBlock-dump             */
     /*---------------------------------------------------------------*/

     if( depth == 0 )
          {
          print("Varblock-Dump on (j|n) ? ");
          printline();
          switch(getchar())
               {
               case 'j' : case 'J' : vblk_dump = -1; break;
               default             : vblk_dump =  0; break;
               }
          getchar();
          printline();
          }



     /*---------------------------------------------------------------*/
     /* a NULL-pointer is end of binary search                        */
     /*---------------------------------------------------------------*/

     if( ptr == NULL)return;



     if(ptr->right != NULL)d_workprint(ptr->right,depth+1);
     d_wnodeprint(ptr,depth);
     if(ptr->left   != NULL)d_workprint(ptr->left ,depth+1);

} /* d_workprint() */



/***************************************************************************/



/* PJ 31.03.95  NEW_PROTO replaced by LINT_ARGS */

#ifdef LINT_ARGS
static void d_wnodeprint(tNODE *ptr,int depth)
#else
static void d_wnodeprint(ptr,depth)
tNODE     *ptr;
int       depth;
#endif

/*--------------------------------------------------------------------*/
/* print information on worknode to standart output                   */
/*--------------------------------------------------------------------*/
/* ptr   = pointer to the work node                                   */
/* depth = depth of node in work tree                                 */
/*--------------------------------------------------------------------*/

{
     int  cnt;



     /*---------------------------------------------------------------*/
#ifdef TRACE_ON
          d_trace("'d_wnodeprint()' :");
#endif
     /*---------------------------------------------------------------*/



     /*---------------------------------------------------------------*/
     /* set position on screen                                        */
     /*---------------------------------------------------------------*/

     for(cnt=0; cnt<depth; cnt++)print(SPACES);



     /*---------------------------------------------------------------*/
#ifdef CHECK
     if( ptr == NULL ){d_checkerr(cISNULL,pos,"ptr");return;}
#endif
     /*---------------------------------------------------------------*/



     /*---------------------------------------------------------------*/
     /* print name of the node                                        */
     /*---------------------------------------------------------------*/

     d_wprintname(ptr);



     /*---------------------------------------------------------------*/
     /* print basetype of node                                        */
     /*---------------------------------------------------------------*/

     switch(ptr->basetype)
          {
          case no_arithtype   : print("NOBASE | ");         break;
          case d_anyarith     : print("any arith | ");      break;
          case d_integer      : print("integer | ");        break;
          case d_real         : print("real | ");           break;
          case d_complex      : print("complex | ");        break;
          case d_interval     : print("interval | ");       break;
          case d_cinterval    : print("cinterval | ");      break;
          case d_rdot         : print("r_dot | ");          break;
          case d_cdot         : print("c_dot | ");          break;
          case d_idot         : print("i_dot | ");          break;
          case d_cidot        : print("ci_dot | ");         break;
          default             : print("UNKNOWN ARITH | ");
          }



     /*---------------------------------------------------------------*/
     /* print structure of node                                       */
     /*---------------------------------------------------------------*/

     switch(ptr->structure)
          {
          case no_structype : print("NOSTRUC | "); break;
          case d_scalar     : print("scalar | ");  break;
          case d_vector     : print("vector | ");  break;
          case d_matrix     : print("matrix | ");  break;
          default           : print("UNKNOWN STRUC | ");
          }



     /*---------------------------------------------------------------*/
     /* print node kind                                               */
     /*---------------------------------------------------------------*/

     switch(ptr->kind)
          {
          case no_nodetype    : print("NOKIND | ");         break;
          case d_constant     : print("const | ");          break;
          case d_nconst       : print("named const | ");    break;
          case d_variable     : print("var | ");            break;
          case d_function     : print("func | ");           break;
          case d_Operator     : print("op | ");             break;
          case d_expression   : print("expr | ");           break;
          default             : print("UNKNOWN KIND | ");   break;
          }



     /*---------------------------------------------------------------*/
     /* print name of father node                                     */
     /*---------------------------------------------------------------*/

     print("f = ");
     if( ptr->father != NULL )
          d_wprintname(ptr->father);
     else
          print("no father | ");



     printline();



     /*---------------------------------------------------------------*/
     /* dump VarInfo if exists                                        */
     /*---------------------------------------------------------------*/

     if( (ptr->namekind==d_varblk) && vblk_dump )
          d_vblkdump(ptr->name.varinfo);

} /* d_wnodeprint() */



/***************************************************************************/


/* PJ 31.03.95  NEW_PROTO replaced by LINT_ARGS */

#ifdef LINT_ARGS
static void d_wprintname(tNODE *ptr)
#else
static void d_wprintname(ptr)
tNODE     *ptr;
#endif

/*--------------------------------------------------------------------*/
/* The function prints the name of a work node.                       */
/*--------------------------------------------------------------------*/
/* ptr = pointer to work node                                         */
/*--------------------------------------------------------------------*/

{
     char buff[32];




     /*---------------------------------------------------------------*/
#ifdef TRACE_ON
          d_trace("'d_wprintname()' :");
#endif
     /*---------------------------------------------------------------*/



     /*---------------------------------------------------------------*/
     /* print information of indirect node names                      */
     /*---------------------------------------------------------------*/

     switch( ptr->namekind )
          {
          /*----------------------------------------------------------*/
          case no_namekind :
               print("NO NAMEKIND | ");
               return;
          /*----------------------------------------------------------*/
          case d_binptr :
               switch( ptr->kind )
                    {
                    case d_variable     :
                         d_printvar(ptr->name.bin);
                         print(" | ");
                         break;
                    case d_expression   :
                         print("nameless | ");
                         break;
                    default :
                         print("ILLEGAL KIND | ");
                         break;
                    } /* switch( ptr->kind ) */
               return;
          /*----------------------------------------------------------*/
          case d_varblk :
               d_printvar(ptr->name.varinfo->element);
               return;
          /*----------------------------------------------------------*/
          case d_value :
               (void) sprintf(buff,"%ld | ",ptr->name.value);
               print(buff);
               return;
          /*----------------------------------------------------------*/
          case d_charptr :
               print(ptr->name.str);
               print(" | ");
               return;
          /*----------------------------------------------------------*/
          } /* switch( ptr->namekind ) */



     /*---------------------------------------------------------------*/
     /* print direct node names                                       */
     /*---------------------------------------------------------------*/

     if( ptr->namekind == d_direct )
          switch(ptr->name.direct)
               {
               case no_nodename    : print("NONAME | ");         return;
               case d_intval       : print("intval | ");         return;
               case d_inf          : print("inf | ");            return;
               case d_sup          : print("sup | ");            return;
               case d_re           : print("re | ");             return;
               case d_im           : print("im | ");             return;
               case d_conj         : print("conj | ");           return;
               case d_compl        : print("compl | ");          return;
               case d_transp       : print("transp | ");         return;
               case d_id           : print("id | ");             return;
               case d_herm         : print("herm | ");           return;
               case d_rvector      : print("rvector | ");        return;
               case d_ivector      : print("ivector | ");        return;
               case d_cvector      : print("cvector | ");        return;
               case d_civector     : print("civector | ");       return;
               case d_rmatrix      : print("rmatrix | ");        return;
               case d_imatrix      : print("imatrix | ");        return;
               case d_cmatrix      : print("cmatrix | ");        return;
               case d_cimatrix     : print("cimatrix | ");       return;
               case d_sharp        : print("# | ");              return;
               case d_sharpnext    : print("#* | ");             return;
               case d_sharpup      : print("#> | ");             return;
               case d_sharpdown    : print("#< | ");             return;
               case d_sharpintval  : print("## | ");             return;
               case d_plus         : print("+ | ");              return;
               case d_minus        : print("- | ");              return;
               case d_mult         : print("* | ");              return;
               case d_mplus        : print("+(m) | ");           return;
               case d_mminus       : print("-(m) | ");           return;
               case d_sum          : print("sum | ");            return;
               case d_assign       : print(":= | ");             return;
               case d_to           : print("to | ");             return;
               case d_downto       : print("downto | ");         return;
               case d_comma        : print(", | ");              return;
               case d_lpard        : print("(<...) | ");         return;
               case d_lparu        : print("(>...) | ");         return;
               default             : print("UNKNOWN NAME | ");   return;
               } /* switch(...) */

     print("ILLEGALE NAMEKIND | ");

} /* d_wprintname() */



/**************************************************************************/


/* PJ 31.03.95  NEW_PROTO replaced by LINT_ARGS */

#ifdef LINT_ARGS
static void d_printvar(binptr expr)
#else
static void d_printvar(expr)
binptr    expr;
#endif

/*--------------------------------------------------------------------*/
/* The function prints the name of a variable, if accessable.         */
/*--------------------------------------------------------------------*/
/* expr = pointer to variable-expression                              */
/*--------------------------------------------------------------------*/

{
     while( !binisvariable(expr) )
          {
          switch( binsymc(expr) )
               {
               case cCall  : expr = binright(expr);                   break;
               case cComma : expr = binleft(expr);                    break;
               case cArry  : expr = binleft(expr);                    break;
               default     : print("no access on name | ");return;
               } /* switch() */
          if( expr == NIL )
               {print("no access on name | ");return;}
          }

     print(symtab(idsym(expr->pt.id)));
     print(" | ");

} /* d_printvar() */



/**************************************************************************/


/* PJ 31.03.95  NEW_PROTO replaced by LINT_ARGS */

#ifdef LINT_ARGS
static void d_vblkdump(tVARBLK *varblk)
#else
static void d_vblkdump(varblk)
tVARBLK   *varblk;
#endif

/*--------------------------------------------------------------------*/
/* The function makes a dump of all information in the VarBlock       */
/* structure.                                                         */
/*--------------------------------------------------------------------*/
/* varblk = pointer to VarBlock-structure                             */
/*--------------------------------------------------------------------*/

{
     print("VarBlock-Dump:");printline();
     if(varblk==NULL)
          {
          print("\tPointer to VarBlock is NULL !");
          printline();
          return;
          }



     if(varblk->original)
          print("\toriginal ");
     else
          print("\tnon original ");

     /*---------------------------------------------------------------*/
     /* print element type                                            */
     /*---------------------------------------------------------------*/
     if( varblk->eltype == idINTEGER    ){print("integer ");     goto ntpr;}
     if( varblk->eltype == idREAL       ){print("real ");        goto ntpr;}
     if( varblk->eltype == idINTERVAL   ){print("interval ");    goto ntpr;}
     if( varblk->eltype == idCOMPLEX    ){print("complex ");     goto ntpr;}
     if( varblk->eltype == idCINTERVAL ){print("cinterval ");    goto ntpr;}
     if( varblk->eltype == idBAD_ID     ){print("bad_id ");      goto ntpr;}
     print("UNKNOWN ");
     ntpr : ;

     /*---------------------------------------------------------------*/
     /* print element structure                                       */
     /*---------------------------------------------------------------*/
     switch(varblk->numdim)
          {
          case 0  : print("scalar type");         printline();break;
          case 1  : print("vector type");         printline();break;
          case 2  : print("matrix type");         printline();break;
          default : print("(ILLEGAL NUMDIM)");    printline();break;
          }

     /*---------------------------------------------------------------*/
     /* dump info about column and row informations                   */
     /*---------------------------------------------------------------*/
     switch(varblk->numdim)
          {
          case 2    : if(varblk->colptr!=NIL)
                           {print("\tcolptr exists !");printline();}
                      else
                           {print("\tcolptr is missing !");printline();}
                      print("\tDump of collb:");  printline();
                      dump(varblk->collb);        printline();
                      print("\tDump of colub:");  printline();
                      dump(varblk->colub);        printline();
          case 1    : if(varblk->rowptr!=NIL)
                           {print("\trowptr exists !");printline();}
                      else
                           {print("\trowptr is missing !");printline();}
                      print("\tDump of rowlb:");  printline();
                      dump(varblk->rowlb);        printline();
                      print("\tDump of rowub:");  printline();
                      dump(varblk->rowub);        printline();
          case 0    : break;
          default   : print("\tILLEGAL NUMDIM !");printline();
          } /* switch(varblk->numdim) */

     /*---------------------------------------------------------------*/
     /* dump variable if exist                                        */
     /*---------------------------------------------------------------*/
     if(varblk->element!=NIL)
          {
          print("\tDump of Variable:"); printline();
          dump(varblk->element);        printline();
          }
     else
          {print("\tElement is missing");printline();}

} /* d_vblkdump() */



/***************************************************************************/
