/**********************************************************************/
/*                                                                    */
/*   "DP_TREE.C"                                                      */
/*                                                                    */
/*   Last change    == Apr-26-92 ==                                   */
/*        A.Just / R.Hammer                                           */
/*                                                                    */
/* Routines handling the creation of the working tree for dotpreci-   */
/* sion expressions.                                                  */
/*                                                                    */
/* Feb-01-92 Just : Macro WPRINT_ON and dependend include of          */
/*                  "DP_UTIL.INC" added.                              */
/* Feb-03-92 Just : Prototype deklarations of static functions        */
/*                  removed.                                          */
/* Feb-09-92 Just : New instructions for types d_cdot, d_idot &       */
/*                  d_cidot inserted, all marked with "New_dot"       */
/* Mar-07-92 Just : macro 'NEW_PROTO' --> 'LINT_ARGS'                 */
/* Apr-23-92 Just : Creation of named constants changed.              */
/* Apr-26-92 Just : Error removed in function 'd_namedconst()'.       */
/*                                                                    */
/**********************************************************************/



/* #define TRACE_ON   */
/* #define WPRINT_ON  */



     /*-----------------------------------------------*/
     /* constants for use of function 'd_checknode()' */
     /*-----------------------------------------------*/
#define chkNAME 0
#define chkKIND 1


#include <stdio.h>

#include "impor.t"
#ifdef IBMva
#include <varargs.h>
#else
#include <stdarg.h>
#endif

#include    "error.h"
#include    "scan.h"
#include    "symtab.h"
#include    "idtab.h"
#include    "blktab.h"
#include    "optab.h"
#include    "bintab.h"
#include    "std.h"
#include    "seman.h"
#include    "arrx.h"

#include    "dp_nodes.h"
#include    "dp_alloc.h"
#include    "dp_error.h"
#include "expor.t"
#include    "dp_tree.h"
#include "implemen.t"


#if defined(TRACE_ON) || defined(WPRINT_ON)
#include "dp_util.inc"
#endif



/**************************************************************************/



#ifdef LINT_ARGS
static void d_namedconst(register tNODE *New,register binptr source)
#else
static void d_namedconst(New,source)
register tNODE      *New;
register binptr     source;
#endif

/*--------------------------------------------------------------------*/
/* This function compute all information about the spezified named    */
/* constant.                                                          */
/*--------------------------------------------------------------------*/
/* New = pointer to the work node                                     */
/* source = pointer to the source node                                */
/*--------------------------------------------------------------------*/

{
     register binptr     help,id_val;



     /*---------------------------------------------------------------*/
#ifdef TRACE_ON
     d_trace("'d_namedconst()'");
#endif
     /*---------------------------------------------------------------*/



     /*---------------------------------------------------------------*/
#ifdef CHECK
     if( New == NULL )
          {d_checkerr(cISNULL,pos,"New");return;}
     if( New->father == NULL )
          {d_checkerr(cISNULL,New->pos,"New->father");return;}
#endif
     /*---------------------------------------------------------------*/



     New->kind = d_nconst;

     /*----------------------------------------------------------*/
     /* now compute the valid kinds of constants                 */
     /*----------------------------------------------------------*/

     id_val=idvalue(binid(source));
     help = iddatyp(binid(source));

     /*----------------------------------------------------------*/
     if( help == integertype ){
          New->basetype   = d_integer;
          New->structure  = d_scalar;
          New->namekind   = d_value;
          New->name.value = binvalue(id_val);
          return;}
     /*----------------------------------------------------------*/
     if( help == realtype ){
          New->basetype  = d_real;
          New->structure = d_scalar;
          New->namekind  = d_binptr;
          New->name.bin  = bincopy(source);
          return;}
     /*----------------------------------------------------------*/



     d_noncritical(dNOFLAG,dWRGNCNST,New);

} /* d_namedconst() */



/**************************************************************************/



#ifdef IBMva

#ifdef LINT_ARGS
#else
#endif
static boolean d_checknode(ptr,num,va_alist)
tNODE     *ptr;
int       num;
va_dcl

#else

#ifdef LINT_ARGS
#else
#endif
static boolean d_checknode(tNODE *ptr,int num, ... )

#endif

/*--------------------------------------------------------------------*/
/* This function checks the worknode 'ptr', if it fits to one of the  */
/* demanded node kinds or direct names.                               */
/* Description:                                                       */
/*        d_checknode(ptr,nof_demands                                 */
/*                       chkKIND,d_function,                          */
/*                       chkNAME,d_plus      );                       */
/*                                                                    */
/*        nof_demands: number of checks to perform                    */
/*                                                                    */
/* This function call for example will test 'ptr' to be a function    */
/* node OR a plus-Operator node (with direct name 'd_plus')           */
/* But make sure that the value of 'nof_demands' fits with the argu-  */
/* ments passed in the function call or there will be suspicious re-  */
/* sults.                                                             */
/*--------------------------------------------------------------------*/
/* ptr = pointer to work node                                         */
/* num = number of test to be performed on the work node              */
/*--------------------------------------------------------------------*/

{
     va_list   arg_ptr;
     tNODETYPE kind;
     tNODENAME name;
     boolean   result;
     int       cnt,kind_of_check;



#ifdef IBMva
     va_start(arg_ptr);
#else
     va_start(arg_ptr,num);
#endif



     if( ptr == NULL )return(false);



     result = false;
     for(cnt=0;cnt<num;cnt++){
          kind_of_check = va_arg(arg_ptr,int);
          switch(kind_of_check){
               case chkKIND : kind = va_arg(arg_ptr,tNODETYPE);
                              result |= (kind==ptr->kind);
                              break;
               case chkNAME : name = va_arg(arg_ptr,tNODENAME);
                              if(ptr->namekind!=d_direct)break;
                              result |= (name==ptr->name.direct);
                              break;
               default      : result = false;
#ifdef CHECK
                              d_checkerr(cILVALUE,ptr->pos,"kind_of_check");
#endif
               }/*switch(...)*/ }/*for(...*/

     return(result);

} /* d_checknode() */



/**************************************************************************/



#ifdef LINT_ARGS
static void d_namekind(register tNODE *New,register binptr source)
#else
static void d_namekind(New,source)
register tNODE      *New;
register binptr     source;
#endif

/*--------------------------------------------------------------------*/
/* This function computes the kind and name of the worknode.          */
/*--------------------------------------------------------------------*/
/* New = pointer to the work node                                     */
/* source = pointer to the source node                                */
/*--------------------------------------------------------------------*/

{
     register tNODE *ptr;



     /*---------------------------------------------------------------*/
#ifdef TRACE_ON
     d_trace("'d_namekind()'");
#endif
     /*---------------------------------------------------------------*/



     /*---------------------------------------------------------------*/
#ifdef CHECK
     if( New == NULL ){d_checkerr(cISNULL,pos,"New");return;}
#endif
     /*---------------------------------------------------------------*/



     /*---------------------------------------------------------------*/
     /* check if 'New' is a right child of a 'd_assign' node of a     */
     /* left child of a 'd_to' or 'd_downto' node. In one of these    */
     /* cases the New node will be a 'd_expression'.                  */
     /*---------------------------------------------------------------*/

     if( (ptr=New->father) != NULL ){
          if( ptr->namekind == d_direct )
               switch( ptr->name.direct ){
                    /*-----------------------------------------------------*/
                    case d_assign  :
                         if( source == binright(binfather(source)) ){
                              New->kind      = d_expression;
                              New->namekind  = d_binptr;
                              New->name.bin  = source;
                              return;}
                         break;
                    /*-----------------------------------------------------*/
                    case d_to      :
                    case d_downto  :
                         if( source == binleft(binfather(source)) ){
                              New->kind      = d_expression;
                              New->namekind  = d_binptr;
                              New->name.bin  = source;
                              return;}
                         break;
                    /*-----------------------------------------------------*/
                    } /* switch() */} /* if( (ptr=New->father) != NULL ) */



     /*---------------------------------------------------------------*/
     /* compute name and kind of named constants                      */
     /*---------------------------------------------------------------*/

     if( idgetkind(binid(source)) == IDCONST ){
          d_namedconst(New,source);
          return;}



     /*---------------------------------------------------------------*/
     /* compute name and kind of a variable node                      */
     /*---------------------------------------------------------------*/

     if ( binisvariable(source) ){
          New->kind      = d_variable;
          New->name.bin  = source;
          New->namekind  = d_binptr;
          return;}



     /*---------------------------------------------------------------*/
     /* compute name and kind of other node types                     */
     /*---------------------------------------------------------------*/

     switch( binsymc(source) ){
          /*----------------------------------------------------------*/
          case cDot       : New->name.direct = d_sharp;
                            New->kind        = d_Operator;
                            New->namekind    = d_direct;    break;
          /*----------------------------------------------------------*/
          case cDotN      : New->name.direct = d_sharpnext;
                            New->kind        = d_Operator;
                            New->namekind    = d_direct;    break;
          /*----------------------------------------------------------*/
          case cDotD      : New->name.direct = d_sharpdown;
                            New->kind        = d_Operator;
                            New->namekind    = d_direct;    break;
          /*----------------------------------------------------------*/
          case cDotU      : New->name.direct = d_sharpup;
                            New->kind        = d_Operator;
                            New->namekind    = d_direct;    break;
          /*----------------------------------------------------------*/
          case cDotS      : New->name.direct = d_sharpintval;
                            New->kind        = d_Operator;
                            New->namekind    = d_direct;    break;
          /*----------------------------------------------------------*/
          case cPlus      : New->name.direct = d_plus;
                            New->kind        = d_Operator;
                            New->namekind    = d_direct;    break;
          /*----------------------------------------------------------*/
          case cMinus     : New->name.direct = d_minus;
                            New->kind        = d_Operator;
                            New->namekind    = d_direct;    break;
          /*----------------------------------------------------------*/
          case cMult      : New->name.direct = d_mult;
                            New->kind        = d_Operator;
                            New->namekind    = d_direct;    break;
          /*----------------------------------------------------------*/
          case cMPlus     : New->name.direct = d_mplus;
                            New->kind        = d_Operator;
                            New->namekind    = d_direct;    break;
          /*----------------------------------------------------------*/
          case cMMinus    : New->name.direct = d_mminus;
                            New->kind        = d_Operator;
                            New->namekind    = d_direct;    break;
          /*----------------------------------------------------------*/
          case cFOR       : New->name.direct = d_sum;
                            New->kind        = d_Operator;
                            New->namekind    = d_direct;    break;
          /*----------------------------------------------------------*/
          case cAssign    : New->name.direct = d_assign;
                            New->kind        = d_Operator;
                            New->namekind    = d_direct;    break;
          /*----------------------------------------------------------*/
          case cTO        : New->name.direct = d_to;
                            New->kind        = d_Operator;
                            New->namekind    = d_direct;    break;
          /*----------------------------------------------------------*/
          case cDOWNTO    : New->name.direct = d_downto;
                            New->kind        = d_Operator;
                            New->namekind    = d_direct;    break;
          /*----------------------------------------------------------*/
          case cComma     : New->name.direct = d_comma;
                            New->kind        = d_Operator;
                            New->namekind    = d_direct;    break;
          /*----------------------------------------------------------*/
          case cINTVAL    : New->name.direct = d_intval;
                            New->kind        = d_function;
                            New->namekind    = d_direct;    break;
          /*----------------------------------------------------------*/
          case cINF       : New->name.direct = d_inf;
                            New->kind        = d_function;
                            New->namekind    = d_direct;    break;
          /*----------------------------------------------------------*/
          case cSUP       : New->name.direct = d_sup;
                            New->kind        = d_function;
                            New->namekind    = d_direct;    break;
          /*----------------------------------------------------------*/
          case cRE        : New->name.direct = d_re;
                            New->kind        = d_function;
                            New->namekind    = d_direct;    break;
          /*----------------------------------------------------------*/
          case cIM        : New->name.direct = d_im;
                            New->kind        = d_function;
                            New->namekind    = d_direct;    break;
          /*----------------------------------------------------------*/
          case cCONJ      : New->name.direct = d_conj;
                            New->kind        = d_function;
                            New->namekind    = d_direct;    break;
          /*----------------------------------------------------------*/
          case cCOMPL     : New->name.direct = d_compl;
                            New->kind        = d_function;
                            New->namekind    = d_direct;    break;
          /*----------------------------------------------------------*/
          case cTRANSP    : New->name.direct = d_transp;
                            New->kind        = d_function;
                            New->namekind    = d_direct;    break;
          /*----------------------------------------------------------*/
          case cID        : New->name.direct = d_id;
                            New->kind        = d_function;
                            New->namekind    = d_direct;    break;
          /*----------------------------------------------------------*/
          case cHERM      : New->name.direct = d_herm;
                            New->kind        = d_function;
                            New->namekind    = d_direct;    break;
          /*----------------------------------------------------------*/
          case cRVECTOR   : New->name.direct = d_rvector;
                            New->kind        = d_function;
                            New->namekind    = d_direct;    break;
          /*----------------------------------------------------------*/
          case cIVECTOR   : New->name.direct = d_ivector;
                            New->kind        = d_function;
                            New->namekind    = d_direct;    break;
          /*----------------------------------------------------------*/
          case cCVECTOR   : New->name.direct = d_cvector;
                            New->kind        = d_function;
                            New->namekind    = d_direct;    break;
          /*----------------------------------------------------------*/
          case cCIVECTOR  : New->name.direct = d_civector;
                            New->kind        = d_function;
                            New->namekind    = d_direct;    break;
          /*----------------------------------------------------------*/
          case cRMATRIX   : New->name.direct = d_rmatrix;
                            New->kind        = d_function;
                            New->namekind    = d_direct;    break;
          /*----------------------------------------------------------*/
          case cIMATRIX   : New->name.direct = d_imatrix;
                            New->kind        = d_function;
                            New->namekind    = d_direct;    break;
          /*----------------------------------------------------------*/
          case cCMATRIX   : New->name.direct = d_cmatrix;
                            New->kind        = d_function;
                            New->namekind    = d_direct;    break;
          /*----------------------------------------------------------*/
          case cCIMATRIX  : New->name.direct = d_cimatrix;
                            New->kind        = d_function;
                            New->namekind    = d_direct;    break;
          /*----------------------------------------------------------*/
          case cRcon      : if( !d_checknode(New->father,2,chkNAME,d_lpard,
                                                           chkNAME,d_lparu) )
                                 d_warn(wNOCPAR,New->pos);
                            New->name.str    = d_getstring(bincharp(source));
                            New->kind        = d_constant;
                            New->namekind    = d_charptr;
                            New->structure   = d_scalar;
                            New->basetype    = d_real;
                            break;
          /*----------------------------------------------------------*/
          case cIcon      : New->name.value  = binvalue(source);
                            New->kind        = d_constant;
                            New->namekind    = d_value;
                            New->structure   = d_scalar;
                            New->basetype    = d_integer;   break;
          /*----------------------------------------------------------*/
          case cLParD     : New->name.direct = d_lpard;
                            New->kind        = d_Operator;
                            New->namekind    = d_direct;    break;
          /*----------------------------------------------------------*/
          case cLParU     : New->name.direct = d_lparu;
                            New->kind        = d_Operator;
                            New->namekind    = d_direct;    break;
          /*----------------------------------------------------------*/
          default : d_noncritical(dNOFLAG,dILLSMBL,New);
          /*----------------------------------------------------------*/
          } /* switch( binsymc(source) ) */

} /* d_namekind() */



/**************************************************************************/



#ifdef LINT_ARGS
static tARITHTYPE d_oparith(register tARITHTYPE arith1,
                            register tARITHTYPE arith2)
#else
static tARITHTYPE d_oparith(arith1,arith2)
register tARITHTYPE arith1,arith2;
#endif

/*--------------------------------------------------------------------*/
/* This funtcion computes the resulting structype and arithtype of a  */
/* simple operation (+, -, *).                                        */
/*--------------------------------------------------------------------*/
/* arith1, arith2 = arithmetical basetype of two operands             */
/*--------------------------------------------------------------------*/

{
     /*---------------------------------------------------------------*/
#ifdef TRACE_ON
     d_trace("'d_oparith()'");
#endif
     /*---------------------------------------------------------------*/



     /*---------------------------------------------------------------*/
     /* if one is not defined, return without action                  */
     /*---------------------------------------------------------------*/

     if( (arith1==no_arithtype) || (arith2==no_arithtype) )
          return(no_arithtype);



     /*---------------------------------------------------------------*/
     /* if both are same type, the result is given                    */
     /*---------------------------------------------------------------*/

     if( arith1 == arith2 ) return(arith1);



     /*------------------------------------------------------*/
     /* if the are of different types,the result is computed */
     /*------------------------------------------------------*/

     switch(arith1){
          /*---------------------------------------------------------------*/
          case d_integer   :
          case d_real      :
               return(arith2);
          /*---------------------------------------------------------------*/
          case d_rdot      :
               switch(arith2){
                    case d_integer   :
                    case d_real      :
                    case d_rdot      : return(d_rdot);
                    case d_complex   :
                    case d_cdot      : return(d_cdot);
                    case d_interval  :
                    case d_idot      : return(d_idot);
                    case d_cinterval :
                    case d_cidot     : return(d_cidot);
                    default : 
#ifdef CHECK
                              d_checkerr(cILBASE,pos,NULL);
#endif
                              return(no_arithtype);
                    } /* switch(arith2) */
          /*---------------------------------------------------------------*/
          case d_complex   :
               switch(arith2){
                    case d_integer   :
                    case d_real      :
                    case d_complex   : return(d_complex);
                    case d_rdot      :
                    case d_cdot      : return(d_cdot);
                    case d_interval  :
                    case d_cinterval : return(d_cinterval);
                    case d_idot      :
                    case d_cidot     : return(d_cidot);
                    default :
#ifdef CHECK
                              d_checkerr(cILBASE,pos,NULL);
#endif
                              return(no_arithtype);
                    } /* switch(arith2) */
          /*---------------------------------------------------------------*/
          case d_cdot      :
               switch(arith2){
                    case d_integer   :
                    case d_real      :
                    case d_rdot      :
                    case d_complex   :
                    case d_cdot      : return(d_cdot);
                    case d_interval  :
                    case d_idot      :
                    case d_cinterval :
                    case d_cidot     : return(d_cidot);
                    default :
#ifdef CHECK
                              d_checkerr(cILBASE,pos,NULL);
#endif
                              return(no_arithtype);
                    } /* switch(arith2) */
          /*---------------------------------------------------------------*/
          case d_interval  :
               switch(arith2){
                    case d_integer   :
                    case d_real      :
                    case d_interval  : return(d_interval);
                    case d_rdot      :
                    case d_idot      : return(d_idot);
                    case d_complex   :
                    case d_cinterval : return(d_cinterval);
                    case d_cdot      :
                    case d_cidot     : return(d_cidot);
                    default :
#ifdef CHECK
                              d_checkerr(cILBASE,pos,NULL);
#endif
                              return(no_arithtype);
                    } /* switch(arith2) */
          /*---------------------------------------------------------------*/
          case d_idot      :
               switch(arith2){
                    case d_integer   :
                    case d_real      :
                    case d_rdot      :
                    case d_interval  :
                    case d_idot      : return(d_idot);
                    case d_complex   :
                    case d_cdot      :
                    case d_cinterval :
                    case d_cidot     : return(d_cidot);
                    default :
#ifdef CHECK
                              d_checkerr(cILBASE,pos,NULL);
#endif
                              return(no_arithtype);
                    } /* switch(arith2) */
          /*---------------------------------------------------------------*/
          case d_cinterval :
               switch(arith2){
                    case d_integer   :
                    case d_real      :
                    case d_complex   :
                    case d_interval  :
                    case d_cinterval : return(d_cinterval);
                    case d_rdot      :
                    case d_cdot      :
                    case d_idot      :
                    case d_cidot     : return(d_cidot);
                    default :
#ifdef CHECK
                              d_checkerr(cILBASE,pos,NULL);
#endif
                              return(no_arithtype);
                    } /* switch(arith2) */
          /*---------------------------------------------------------------*/
          case d_cidot     :
               return(d_cidot);
          /*---------------------------------------------------------------*/
          default :
#ifdef CHECK
                    d_checkerr(cILBASE,pos,NULL);
#endif
                    return(no_arithtype);
          /*---------------------------------------------------------------*/
          } /* switch(arith1) */


} /* d_oparith() */



/**************************************************************************/



#ifdef LINT_ARGS
static void d_calcfunc(register tNODE *New)
#else
static void d_calcfunc(New)
register tNODE *New;
#endif

/*--------------------------------------------------------------------*/
/* This function computes the structype and arithtype of a function   */
/* work node.                                                         */
/*--------------------------------------------------------------------*/
/* New = pointer to work node                                         */
/*--------------------------------------------------------------------*/

{
     register tNODE *arg,*help;
     register int   nof_args;



     /*---------------------------------------------------------------*/
#ifdef TRACE_ON
     d_trace("'d_calcfunc()'");
#endif
     /*---------------------------------------------------------------*/



     /*---------------------------------------------------------------*/
#ifdef CHECK
     if( New == NULL ){d_checkerr(cISNULL,pos,"New");return;}
#endif
     /*---------------------------------------------------------------*/



     /*---------------------------------------------------------------*/
     /* test structure of function node                               */
     /*---------------------------------------------------------------*/
     if( New->left != NULL )
          {d_noncritical(dSELF,dBADSYNTAX,New);return;}
     if( (arg=New->right) == NULL )
          {d_noncritical(dSELF,dMISSARG,New);return;}



     /*---------------------------------------------------------------*/
     /* test number of function arguments                             */
     /*---------------------------------------------------------------*/
     nof_args = 1;
     help = arg;
     while( (help->namekind==d_direct) && (help->name.direct==d_comma) ){
          if( !d_checknode(help->left,6,chkKIND,d_function,
                                        chkKIND,d_variable,
                                        chkKIND,d_constant,
                                        chkKIND,d_nconst,
                                        chkNAME,d_lpard,
                                        chkNAME,d_lparu)       )
               {d_noncritical(dSELF,dILLOPND,help);return;}
          nof_args++;
          if( (help=help->right) == NULL )
               {d_noncritical(dNOFLAG,dBADSYNTAX,New);return;}} /* while() */

     if( !d_checknode(help,6,chkKIND,d_function,
                             chkKIND,d_variable,
                             chkKIND,d_constant,
                             chkKIND,d_nconst,
                             chkNAME,d_lpard,
                             chkNAME,d_lparu)        )
          {d_noncritical(dSELF,dILLOPND,help);return;}



     /*---------------------------------------------------------------*/
     /* compute result of the various functions                       */
     /*---------------------------------------------------------------*/
     switch(New->name.direct){
          /*----------------------------------------------------------*/
          case d_inf       :
          case d_sup       :
               if( nof_args > 1 ){d_noncritical(dNOFLAG,dSUPLARG,New);break;}
               New->structure = arg->structure;
               switch(arg->basetype){
                    case d_interval  : New->basetype = d_real;        break;
                    case d_idot      : New->basetype = d_rdot;        break;
                    case d_cinterval : New->basetype = d_complex;     break;
                    case d_cidot     : New->basetype = d_cdot;        break;
                    default : d_noncritical(dBASE+dSELF,dWRGARGTP,arg);}
               break;
          /*----------------------------------------------------------*/
          case d_re        :
          case d_im        :
               if( nof_args > 1 ){d_noncritical(dNOFLAG,dSUPLARG,New);break;}
               New->structure = arg->structure;
               switch(arg->basetype){
                    case d_complex   : New->basetype = d_real;        break;
                    case d_cdot      : New->basetype = d_rdot;        break;
                    case d_cinterval : New->basetype = d_interval;    break;
                    case d_cidot     : New->basetype = d_idot;        break;
                    default : d_noncritical(dBASE+dSELF,dWRGARGTP,arg);}
               break;
          /*----------------------------------------------------------*/
          case d_conj      :
               if( nof_args > 1 ){d_noncritical(dNOFLAG,dSUPLARG,New);break;}
               New->structure = arg->structure;
               if(    ( arg->basetype == d_complex   )
                   || ( arg->basetype == d_cdot      )
                   || ( arg->basetype == d_cinterval )
                   || ( arg->basetype == d_cidot     )      )
                    New->basetype = arg->basetype;
               else
                    d_noncritical(dBASE+dSELF,dWRGARGTP,arg);
               break;
          /*----------------------------------------------------------*/
          case d_transp    :
               if( nof_args > 1 ){d_noncritical(dNOFLAG,dSUPLARG,New);break;}
               New->basetype = arg->basetype;
               if(arg->structure==d_matrix)
                    New->structure = d_matrix;
               else
                    d_noncritical(dSTRUC+dSELF,dMATREXPCT,arg);
               break;
          /*----------------------------------------------------------*/
          case d_herm      :
               if( nof_args > 1 ){d_noncritical(dNOFLAG,dSUPLARG,New);break;}
               if(arg->structure==d_matrix)
                    New->structure = d_matrix;
               else
                    d_noncritical(dSTRUC+dSELF,dMATREXPCT,arg);
               switch(arg->basetype){
                    case d_complex   :
                    case d_cinterval : New->basetype = arg->basetype; break;
                    default : d_noncritical(dBASE+dSELF,dWRGARGTP,arg);}
               break;
          /*----------------------------------------------------------*/
          case d_id        :
               if( nof_args > 2 ){d_noncritical(dNOFLAG,dSUPLARG,New);break;}
               if((arg->namekind==d_direct)&&(arg->name.direct==d_comma)){
                    if(    (arg->right->kind != d_variable)
                        && (arg->right->kind != d_function)    ){
                         d_noncritical(dNMKND+dSELF,dIDERROR,arg->right);
                         break;}
                    if(arg->right->structure != d_matrix){
                         d_noncritical(dSTRUC+dSELF,dMATREXPCT,arg->right);
                         break;}
                    if(    (arg->left->kind != d_variable)
                        && (arg->left->kind != d_function)    ){
                         d_noncritical(dNMKND+dSELF,dIDERROR,arg->left);
                         break;}
                    if(arg->left->structure != d_matrix){
                         d_noncritical(dSTRUC+dSELF,dMATREXPCT,arg->left);
                         break;}}
               else{
                    if(    (arg->kind != d_variable)
                        && (arg->kind != d_function)    )
                         {d_noncritical(dNMKND+dSELF,dIDERROR,arg);break;}
                    if(arg->structure != d_matrix)
                         {d_noncritical(dSTRUC+dSELF,dMATREXPCT,arg);break;}}
               New->basetype  = d_real;
               New->structure = d_matrix;
               break;         
          /*----------------------------------------------------------*/
          case d_compl     :
               if( nof_args > 2 ){d_noncritical(dNOFLAG,dSUPLARG,New);break;}
               if(    (arg->namekind   ==d_direct)
                   && (arg->name.direct==d_comma )    ){
                    if(    (arg->left->basetype != d_integer  )
                        && (arg->left->basetype != d_real     )
                        && (arg->left->basetype != d_rdot     )
                        && (arg->left->basetype != d_interval )
                        && (arg->left->basetype != d_idot     )     ){
                         d_noncritical(dBASE+dSELF,dWRGARGTP,arg->left);
                         break;}
                    if(    (arg->right->basetype != d_integer  )
                        && (arg->right->basetype != d_real     )
                        && (arg->right->basetype != d_rdot     )
                        && (arg->right->basetype != d_interval )
                        && (arg->right->basetype != d_idot     )     ){
                         d_noncritical(dBASE+dSELF,dWRGARGTP,arg->right);
                         break;}

                    New->basetype = d_oparith(d_complex,
                                              d_oparith(arg->left->basetype,
                                                        arg->right->basetype)
                                              );

                    if( arg->left->structure != arg->right->structure )
                         d_noncritical(dNOFLAG,dARGSSTRUC,New);
                    else
                         New->structure = arg->left->structure;}
               else{
                    if(    (arg->basetype != d_integer  )
                        && (arg->basetype != d_real     )
                        && (arg->basetype != d_rdot     )
                        && (arg->basetype != d_interval )
                        && (arg->basetype != d_idot     )     ){
                         d_noncritical(dBASE+dSELF,dWRGARGTP,arg);
                         break;}
                    New->structure = arg->structure;
                    New->basetype = d_oparith(d_complex,arg->basetype);}
               break;
          /*----------------------------------------------------------*/
          case d_intval    :
               if( nof_args > 2 ){d_noncritical(dNOFLAG,dSUPLARG,New);break;}
               if(    (arg->namekind   ==d_direct)
                   && (arg->name.direct==d_comma )    ){
                    if(    (arg->left->basetype != d_integer )
                        && (arg->left->basetype != d_real    )
                        && (arg->left->basetype != d_rdot    )
                        && (arg->left->basetype != d_complex )
                        && (arg->left->basetype != d_cdot    )     ){
                         d_noncritical(dBASE+dSELF,dWRGARGTP,arg->left);
                         break;}
                    if(    (arg->right->basetype != d_integer )
                        && (arg->right->basetype != d_real    )
                        && (arg->right->basetype != d_rdot    )
                        && (arg->right->basetype != d_complex )
                        && (arg->right->basetype != d_cdot    )     ){
                         d_noncritical(dBASE+dSELF,dWRGARGTP,arg->right);
                         break;}

                    New->basetype = d_oparith(d_interval,
                                              d_oparith(arg->left->basetype,
                                                        arg->right->basetype)
                                              );

                    if( arg->left->structure != arg->right->structure )
                         d_noncritical(dNOFLAG,dARGSSTRUC,New);
                    else
                         New->structure = arg->left->structure;}
               else{
                    if(    (arg->basetype != d_integer )
                        && (arg->basetype != d_real    )
                        && (arg->basetype != d_rdot    )
                        && (arg->basetype != d_complex )
                        && (arg->basetype != d_cdot    )     ){
                         d_noncritical(dBASE+dSELF,dWRGARGTP,arg);
                         break;}
                    New->structure = arg->structure;
                    New->basetype = d_oparith(d_interval,arg->basetype);}
               break;
          /*----------------------------------------------------------*/
          case d_rvector   :  if( nof_args > 1 )
                                   {d_noncritical(dNOFLAG,dSUPLARG,New);break;}
                              if(arg->basetype!=d_real){
                                   d_noncritical(dBASE+dSELF,dWRGARGTP,arg);
                                   break;}
                              goto vec_test;
          case d_ivector   :  if( nof_args > 1 )
                                   {d_noncritical(dNOFLAG,dSUPLARG,New);break;}
                              if(arg->basetype!=d_interval){
                                   d_noncritical(dBASE+dSELF,dWRGARGTP,arg);
                                   break;}
                              goto vec_test;
          case d_cvector   :  if( nof_args > 1 )
                                   {d_noncritical(dNOFLAG,dSUPLARG,New);break;}
                              if(arg->basetype!=d_complex){
                                   d_noncritical(dBASE+dSELF,dWRGARGTP,arg);
                                   break;}
                              goto vec_test;
          case d_civector  :  if( nof_args > 1 )
                                   {d_noncritical(dNOFLAG,dSUPLARG,New);break;}
                              if(arg->basetype!=d_cinterval){
                                   d_noncritical(dBASE+dSELF,dWRGARGTP,arg);
                                   break;}
               vec_test :
                    if(arg->kind!=d_variable)
                         {d_noncritical(dKIND+dSELF,dVAREXPCT,arg);break;}
                    New->basetype  = arg->basetype;
                    if(arg->structure!=d_vector)
                         d_noncritical(dSTRUC+dSELF,dVECTEXPCT,arg);
                    else
                         New->structure = d_vector;

               break;
          /*----------------------------------------------------------*/
          case d_rmatrix   :  if( nof_args > 1 )
                                   {d_noncritical(dNOFLAG,dSUPLARG,New);break;}
                              if(arg->basetype!=d_real){
                                   d_noncritical(dBASE+dSELF,dWRGARGTP,arg);
                                   break;}
                              goto mat_test;
          case d_imatrix   :  if( nof_args > 1 )
                                   {d_noncritical(dNOFLAG,dSUPLARG,New);break;}
                              if(arg->basetype!=d_interval){
                                   d_noncritical(dBASE+dSELF,dWRGARGTP,arg);
                                   break;}
                              goto mat_test;
          case d_cmatrix   :  if( nof_args > 1 )
                                   {d_noncritical(dNOFLAG,dSUPLARG,New);break;}
                              if(arg->basetype!=d_complex){
                                   d_noncritical(dBASE+dSELF,dWRGARGTP,arg);
                                   break;}
                              goto mat_test;
          case d_cimatrix  :  if( nof_args > 1 )
                                   {d_noncritical(dNOFLAG,dSUPLARG,New);break;}
                              if(arg->basetype!=d_cinterval){
                                   d_noncritical(dBASE+dSELF,dWRGARGTP,arg);
                                   break;}
               mat_test :
                    if(arg->kind!=d_variable)
                         {d_noncritical(dKIND+dSELF,dVAREXPCT,arg);break;}
                    New->basetype  = arg->basetype;
                    if(arg->structure!=d_matrix)
                         d_noncritical(dSTRUC+dSELF,dMATREXPCT,arg);
                    else
                         New->structure = d_matrix;

               break;
          /*----------------------------------------------------------*/
          } /* switch(New->name.direct) */

} /* d_calcfunc() */



/**************************************************************************/



#ifdef LINT_ARGS
static void d_csharp(register tNODE *New)
#else
static void d_csharp(New)
register tNODE *New;
#endif

/*--------------------------------------------------------------------*/
/* compute on sharp symbols ( #, #*, #<, #>, ## )                     */
/* the function computes structure and basetype of the Operators      */
/*--------------------------------------------------------------------*/
/* New = pointer to work node                                         */
/*--------------------------------------------------------------------*/

{
     /*---------------------------------------------------------------*/
#ifdef TRACE_ON
     d_trace("'d_dcsharp()'");
#endif
     /*---------------------------------------------------------------*/



     /*---------------------------------------------------------------*/
#ifdef CHECK
     if( New == NULL ){d_checkerr(cISNULL,pos,"New");return;}
#endif
     /*---------------------------------------------------------------*/



     /*---------------------------------------------------------------*/
     /* test structure of the node                                    */
     /*---------------------------------------------------------------*/

     if( New->father != NULL )
          {d_noncritical(dSELF,dBADSYNTAX,New);return;}
     if( (New->left!=NULL) || (New->right==NULL) )
          {d_noncritical(dSELF,dBADSYNTAX,New);return;}



     /*---------------------------------------------------------------*/
     /* compute the result type                                       */
     /*---------------------------------------------------------------*/

     switch( New->name.direct ){
          /*----------------------------------------------------------*/
          case d_sharp :
               switch(New->right->basetype){
                    case no_arithtype : New->basetype = no_arithtype; break;
                    case d_integer    :
                    case d_real       :
                    case d_rdot       : New->basetype = d_rdot;       break;
                    case d_complex    :
                    case d_cdot       : New->basetype = d_cdot;       break;
                    case d_interval   :
                    case d_idot       : New->basetype = d_idot;       break;
                    case d_cinterval  :
                    case d_cidot      : New->basetype = d_cidot;      break;
                    default           : goto base_err;
                    } /* switch(New->right->basetype) */
               
               switch( New->right->structure ){
                    case no_structype :
                    case d_scalar     :
                         New->structure = New->right->structure;
                         break;
                    default : goto struc_err;
                    } /* switch(New->right->structure) */

               break;
          /*----------------------------------------------------------*/
          case d_sharpnext :
          case d_sharpup   :
          case d_sharpdown :

               switch( New->right->basetype ){
                    case no_arithtype : New->basetype = no_arithtype; break;
                    case d_integer    :
                    case d_real       :
                    case d_rdot       : New->basetype = d_real;       break;
                    case d_complex    :
                    case d_cdot       : New->basetype = d_complex;    break;
                    default : goto base_err;
                    } /* switch(New->right->basetype) */

               New->structure = New->right->structure;

               break;
          /*----------------------------------------------------------*/
          case d_sharpintval :

               switch(New->right->basetype){
                    case no_arithtype : New->basetype = no_arithtype; break;
                    case d_integer    :
                    case d_real       :
                    case d_rdot       :
                    case d_interval   :
                    case d_idot       : New->basetype = d_interval;   break;
                    case d_complex    :
                    case d_cdot       :
                    case d_cinterval  :
                    case d_cidot      : New->basetype = d_cinterval;  break;
                    default           : goto base_err;
                    } /* switch(New->right->basetype) */

               New->structure = New->right->structure;

               break;
          /*----------------------------------------------------------*/
          } /* switch(New->name.direct) */

     return;



     /*---------------------------------------------------------------*/
     /* local jump marks for handling errors                          */
     /*---------------------------------------------------------------*/

     base_err :
          d_noncritical(dBASE,dILLRBASE,New);
          return;
     struc_err :
          d_noncritical(dSTRUC,dILLRSTRUC,New);

} /* d_csharp() */



/**************************************************************************/



#ifdef LINT_ARGS
static void d_cmonadic(register tNODE *New)
#else
static void d_cmonadic(New)
register tNODE *New;
#endif

/*--------------------------------------------------------------------*/
/* compute on monadic Operators                                       */
/* the function computes structure and basetype of the Operators      */
/*--------------------------------------------------------------------*/
/* New = pointer to work node                                         */
/*--------------------------------------------------------------------*/

{
     /*---------------------------------------------------------------*/
#ifdef TRACE_ON
     d_trace("'d_cmonadic()'");
#endif
     /*---------------------------------------------------------------*/



     /*---------------------------------------------------------------*/
#ifdef CHECK
     if( New == NULL ){d_checkerr(cISNULL,pos,"New");return;}
#endif
     /*---------------------------------------------------------------*/



     /*---------------------------------------------------------------*/
     /* test structure of the node                                    */
     /*---------------------------------------------------------------*/

     if( (New->right==NULL) || (New->left!=NULL) )
          {d_noncritical(dSELF,dBADSYNTAX,New);return;}



     /*---------------------------------------------------------------*/
     /* structype and arithtype are same as child                     */
     /*---------------------------------------------------------------*/

     New->structure = New->right->structure;
     New->basetype  = New->right->basetype;

} /* d_cmonadic() */



/**************************************************************************/



#ifdef LINT_ARGS
static tNODE *d_skipmonadic(register tNODE *ptr)
#else
static tNODE *d_skipmonadic(ptr)
register tNODE *ptr;
#endif

/*--------------------------------------------------------------------*/
/* skips all nodes representing monadic Operators, starting with      */
/* 'ptr'.                                                             */
/* The return value is a pointer to the first non-monadic node.       */
/*--------------------------------------------------------------------*/
/* ptr = pointer to work tree (or subtree)                            */
/*--------------------------------------------------------------------*/

{
     if(ptr==NULL)return(NULL);

#define IS_MONADIC ( (ptr->namekind==d_direct) \
	&&((ptr->name.direct==d_mplus)||(ptr->name.direct==d_mminus)) )

     while( IS_MONADIC )ptr = ptr->right;
     return(ptr);

#undef IS_MONADIC

} /* d_skipmonadic() */



/**************************************************************************/



#ifdef LINT_ARGS
static void d_cadditive(register tNODE *New)
#else
static void d_cadditive(New)
register tNODE *New;
#endif

/*--------------------------------------------------------------------*/
/* compute on additive Operators                                      */
/* the function computes structure and basetype of the Operators      */
/*--------------------------------------------------------------------*/
/* New = pointer to work node                                         */
/*--------------------------------------------------------------------*/

{
     register tNODE *hr,*hl;



     /*---------------------------------------------------------------*/
#ifdef TRACE_ON
     d_trace("'d_cadditive()'");
#endif
     /*---------------------------------------------------------------*/



     /*---------------------------------------------------------------*/
#ifdef CHECK
     if( New == NULL ){d_checkerr(cISNULL,pos,"New");return;}
#endif
     /*---------------------------------------------------------------*/



     /*---------------------------------------------------------------*/
     /* test structure of the node                                    */
     /*---------------------------------------------------------------*/
     hr = d_skipmonadic(New->right);
     hl = d_skipmonadic(New->left);

     if( !d_checknode(hr,10,chkKIND,d_variable, chkKIND,d_function,
                            chkKIND,d_constant, chkKIND,d_nconst,
                            chkNAME,d_lpard,    chkNAME,d_lparu,
                            chkNAME,d_mult,     chkNAME,d_sum,
                            chkNAME,d_plus,     chkNAME,d_minus)        )
          {d_noncritical(dSELF,dILLOPND,hr);return;}
     if( !d_checknode(hl,10,chkKIND,d_variable, chkKIND,d_function,
                            chkKIND,d_constant, chkKIND,d_nconst,
                            chkNAME,d_lpard,    chkNAME,d_lparu,
                            chkNAME,d_mult,     chkNAME,d_sum,
                            chkNAME,d_plus,     chkNAME,d_minus)        )
          {d_noncritical(dSELF,dILLOPND,hl);return;}



     /*---------------------------------------------------------------*/
     /* and both MUST have the same structure                         */
     /*---------------------------------------------------------------*/

     if( New->left->structure != New->right->structure )
          {d_noncritical(dSTRUC,dOPSSTRUC,New);return;}



     /*---------------------------------------------------------------*/
     /* compute arithtype of the node                                 */
     /*---------------------------------------------------------------*/

          /* got structure from child   */
     New->structure = New->left->structure;
          /* compute basetype of the Operator    */
     New->basetype = d_oparith(New->left->basetype,New->right->basetype);

} /* d_cadditive() */



/**************************************************************************/



#ifdef LINT_ARGS
static void d_cmult(register tNODE *New)
#else
static void d_cmult(New)
register tNODE *New;
#endif

/*--------------------------------------------------------------------*/
/* compute on mult Operator                                           */
/* the function computes structure and basetype of the Operator       */
/*--------------------------------------------------------------------*/
/* New = pointer to work node                                         */
/*--------------------------------------------------------------------*/

{
     register tNODE *hr,*hl;



     /*---------------------------------------------------------------*/
#ifdef TRACE_ON
     d_trace("'d_cmult()'");
#endif
     /*---------------------------------------------------------------*/



     /*---------------------------------------------------------------*/
#ifdef CHECK
     if( New == NULL ){d_checkerr(cISNULL,pos,"New");return;}
#endif
     /*---------------------------------------------------------------*/



     /*---------------------------------------------------------------*/
     /* test structure of the node                                    */
     /*---------------------------------------------------------------*/

     hr = d_skipmonadic(New->right);
     hl = d_skipmonadic(New->left);

     if( !d_checknode(hr,6,chkKIND,d_variable, chkKIND,d_constant,
                           chkKIND,d_nconst,   chkNAME,d_lpard,
                           chkNAME,d_lparu,    chkKIND,d_function)        )
          {d_noncritical(dSELF,dILLOPND,hr);return;}
     if( !d_checknode(hl,6,chkKIND,d_variable, chkKIND,d_constant,
                           chkKIND,d_nconst,   chkNAME,d_lpard,
                           chkNAME,d_lparu,    chkKIND,d_function)        )
          {d_noncritical(dSELF,dILLOPND,hl);return;}



     /*---------------------------------------------------------------*/
     /* compute resulting basetype                                    */
     /*---------------------------------------------------------------*/

     if(  (hl->basetype==d_rdot ) || (hr->basetype==d_rdot ) ||
          (hl->basetype==d_cdot ) || (hr->basetype==d_cdot ) ||
          (hl->basetype==d_idot ) || (hr->basetype==d_idot ) ||
          (hl->basetype==d_cidot) || (hr->basetype==d_cidot)    )
          d_noncritical(dBASE,dDOTMUL,New);
     else
          New->basetype = d_oparith(New->left->basetype,New->right->basetype);



     /*---------------------------------------------------------------*/
     /* compute resulting structure                                   */
     /*---------------------------------------------------------------*/

     switch(New->left->structure){
          /*----------------------------------------------------------*/
          case no_structype :
               New->structure = no_structype;
               break;
          /*----------------------------------------------------------*/
          case d_scalar :
               New->structure = New->right->structure;
               break;
          /*----------------------------------------------------------*/
          case d_vector :
               switch(New->right->structure){
                    case no_structype : New->structure = no_structype;break;
                    case d_scalar     : New->structure = d_vector;    break;
                    case d_vector     : New->structure = d_scalar;    break;
                    default : d_noncritical(dSTRUC,dWRGMSTRUC,New);
                    } /* switch(New->right->structure) */
               break;
          /*----------------------------------------------------------*/
          case d_matrix : 
               switch(New->right->structure){
                    case no_structype : New->structure = no_structype;break;
                    case d_scalar     : New->structure = d_matrix;    break;
                    case d_vector     : New->structure = d_vector;    break;
                    case d_matrix     : New->structure = d_matrix;    break;
                    default : d_noncritical(dSTRUC,dWRGMSTRUC,New);
                    } /* switch(New->right...) */
               break;
          /*----------------------------------------------------------*/
          default : d_noncritical(dSTRUC,dWRGMSTRUC,New);
          /*----------------------------------------------------------*/
          } /* switch(New->left...) */

} /* d_cmult() */



/**************************************************************************/



#ifdef LINT_ARGS
static void d_csum(register tNODE *New)
#else
static void d_csum(New)
register tNODE *New;
#endif

/*--------------------------------------------------------------------*/
/* compute on sum Operator                                            */
/* the function computes structure and basetype of the Operator       */
/*--------------------------------------------------------------------*/
/* New = pointer to work node                                         */
/*--------------------------------------------------------------------*/

{
     register tNODE      *ptr;



     /*---------------------------------------------------------------*/
#ifdef TRACE_ON
     d_trace("'d_csum()'");
#endif
     /*---------------------------------------------------------------*/



     /*---------------------------------------------------------------*/
#ifdef CHECK
     if( New == NULL ){d_checkerr(cISNULL,pos,"New");return;}
#endif
     /*---------------------------------------------------------------*/



     /*---------------------------------------------------------------*/
     /* Check syntax of the sum expression                            */
     /*---------------------------------------------------------------*/

     if( (ptr=New->left) == NULL )
          {d_noncritical(dSELF,dINTASSMNT,New);return;}
     if( ptr->left == NULL )
          {d_noncritical(dSELF,dVAREXPCT,ptr);return;}
     if( ptr->right == NULL )
          {d_noncritical(dSELF,dINTEXPR,ptr);return;}

     if(  ( ptr->namekind    != d_direct  ) ||
          ( ptr->name.direct != d_assign  ) ||
          ( ptr->basetype    != d_integer ) ||
          ( ptr->structure   != d_scalar  )    ){
          d_noncritical(dBASE+dSTRUC+dNMKND+dNAME+dSELF,dINTASSMNT,ptr);
          return;}

     if( ptr->right->kind != d_expression )
          {d_noncritical(dNMKND+dSELF,dINTEXPR,ptr->right);return;}


     if( (ptr=New->right) == NULL )
          {d_noncritical(dSELF,dTOEXPCT,New);return;}

     if(  (           ptr->namekind != d_direct                    ) ||
          ( (ptr->name.direct!=d_to)&&(ptr->name.direct!=d_downto) )    )
          {d_noncritical(dNMKND+dNAME+dSELF,dTOEXPCT,ptr);return;}



     /*---------------------------------------------------------------*/
     /* structype and basetype same as right child                    */
     /*---------------------------------------------------------------*/

     New->structure = New->right->structure;
     New->basetype  = New->right->basetype;

} /* d_csum() */



/**************************************************************************/



#ifdef LINT_ARGS
static void d_cassign(register tNODE *New)
#else
static void d_cassign(New)
register tNODE *New;
#endif

/*--------------------------------------------------------------------*/
/* compute on assign Operator                                         */
/* the function computes structure and basetype of the Operator       */
/*--------------------------------------------------------------------*/
/* New = pointer to work node                                         */
/*--------------------------------------------------------------------*/

{
     register tNODE *hr,*hl;



     /*---------------------------------------------------------------*/
#ifdef TRACE_ON
     d_trace("'d_cassign()'");
#endif
     /*---------------------------------------------------------------*/



     /*---------------------------------------------------------------*/
#ifdef CHECK
     if( New == NULL ){d_checkerr(cISNULL,pos,"New");return;}
#endif
     /*---------------------------------------------------------------*/



     /*---------------------------------------------------------------*/
     /* test structure of the node                                    */
     /*---------------------------------------------------------------*/
     hr = New->right;
     hl = New->left;
     if( (hl==NULL) || (hr==NULL) )
          {d_noncritical(dSELF,dBADSYNTAX,New);return;}
     if( !d_checknode(hl,1,chkKIND,d_variable) )
          {d_noncritical(dSELF+dKIND,dVAREXPCT,hl);return;}



     /*---------------------------------------------------------------*/
     /* check left child to be same type as right child               */
     /*---------------------------------------------------------------*/

     if(  (hl->basetype  != hr->basetype ) ||
          (hl->structure != hr->structure)    )
          {d_noncritical(dBASE+dSTRUC,dASSTYPE,New);return;}



     /*---------------------------------------------------------------*/
     /* if no error occured arithtype is same as children             */
     /*---------------------------------------------------------------*/

     New->basetype  = hr->basetype;
     New->structure = hr->structure;

} /* d_cassign() */



/**************************************************************************/



#ifdef LINT_ARGS
static void d_cto(register tNODE *New)
#else
static void d_cto(New)
register tNODE *New;
#endif

/*--------------------------------------------------------------------*/
/* compute on Operators to and downto                                 */
/* the function computes structure and basetype of the Operators      */
/*--------------------------------------------------------------------*/
/* New = pointer to work node                                         */
/*--------------------------------------------------------------------*/

{
     register tNODE *hl;



     /*---------------------------------------------------------------*/
#ifdef TRACE_ON
     d_trace("'d_cto()'");
#endif
     /*---------------------------------------------------------------*/



     /*---------------------------------------------------------------*/
#ifdef CHECK
     if( New == NULL ){d_checkerr(cISNULL,pos,"New");return;}
#endif
     /*---------------------------------------------------------------*/



     /*---------------------------------------------------------------*/
     /* test structure of the node                                    */
     /*---------------------------------------------------------------*/
     hl = New->left;
     if( hl == NULL )
          {d_noncritical(dSELF,dINTEXPR,New);return;}
     if( New->right == NULL )
          {d_noncritical(dSELF,dBADSYNTAX,New);return;}



     /*---------------------------------------------------------------*/
     /* check left child to be scalar integer                         */
     /*---------------------------------------------------------------*/

     if(  ( hl->kind      != d_expression ) ||
          ( hl->basetype  != d_integer    ) ||
          ( hl->structure != d_scalar     )    ){
          d_noncritical(dKIND+dBASE+dSTRUC+dSELF,dINTEXPR,hl);
          return;}



     /*---------------------------------------------------------------*/
     /* ARITHTYPE and STRUCTYPE are same as of right node             */
     /*---------------------------------------------------------------*/

     New->basetype  = New->right->basetype;
     New->structure = New->right->structure;

} /* d_cto() */



/**************************************************************************/



#ifdef LINT_ARGS
static void d_cpar( register tNODE *New )
#else
static void d_cpar(New)
register tNODE *New;
#endif

/*--------------------------------------------------------------------*/
/* This function gets the result for constant rounding symbols        */
/*--------------------------------------------------------------------*/
/* New = pointer to work node                                         */
/*--------------------------------------------------------------------*/

{
     /*---------------------------------------------------------------*/
#ifdef TRACE_ON
     d_trace("'d_cpar()'");
#endif
     /*---------------------------------------------------------------*/



     /*---------------------------------------------------------------*/
#ifdef CHECK
     if( New == NULL ){d_checkerr(cISNULL,pos,"New");return;}
#endif
     /*---------------------------------------------------------------*/



     /*---------------------------------------------------------------*/
     /* test structure of the node                                    */
     /*---------------------------------------------------------------*/

     if( (New->left!=NULL) || (New->right==NULL) )
          {d_noncritical(dNOFLAG,dILLCPAR,New);return;}



     /*---------------------------------------------------------------*/
     /* Look for unnamed real constant                                */
     /*---------------------------------------------------------------*/

     if(  ( New->right->kind     != d_constant ) ||
          ( New->right->basetype != d_real     )    )
         {d_noncritical(dBASE+dKIND,dPARNOREAL,New);return;}


     New->structure = d_scalar;
     New->basetype  = d_real;

} /* d_cpar() */



/**************************************************************************/



#ifdef LINT_ARGS
static void d_calcop(register tNODE *New)
#else
static void d_calcop(New)
register tNODE *New;
#endif

/*--------------------------------------------------------------------*/
/* This functin computes the structype and arithtype of an Operator   */
/* work node.                                                         */
/*--------------------------------------------------------------------*/
/* New = pointer to work node                                         */
/*--------------------------------------------------------------------*/

{

     /*---------------------------------------------------------------*/
#ifdef TRACE_ON
     d_trace("'d_calcop()'");
#endif
     /*---------------------------------------------------------------*/



     /*---------------------------------------------------------------*/
#ifdef CHECK
     if( New == NULL ){d_checkerr(cISNULL,pos,"New");return;}
#endif
     /*---------------------------------------------------------------*/



     /*---------------------------------------------------------------*/
     /* compute the result of the Operator                            */
     /*---------------------------------------------------------------*/

     switch( New->name.direct ){
          case d_sharp        :
          case d_sharpnext    :
          case d_sharpup      :
          case d_sharpdown    :
          case d_sharpintval  : d_csharp     (New);         break;
          case d_mplus        :
          case d_mminus       : d_cmonadic   (New);         break;
          case d_plus         :
          case d_minus        : d_cadditive  (New);         break;
          case d_mult         : d_cmult      (New);         break;
          case d_sum          : d_csum       (New);         break;
          case d_assign       : d_cassign    (New);         break;
          case d_to           :
          case d_downto       : d_cto        (New);         break;
          case d_comma        : /* no work to be done ! */  break;
          case d_lpard        :
          case d_lparu        : d_cpar       (New);         break;
          } /* switch(New->name.direct) */

} /* d_calcop() */



/**************************************************************************/



#ifdef LINT_ARGS
static boolean d_testconvf(register tNODE *work,register tSTRUCTYPE convtype)
#else
static boolean d_testconvf(work,convtype)
register tNODE      *work;
register tSTRUCTYPE convtype;
#endif

/*--------------------------------------------------------------------*/
/* This function return true, when the father of 'work' is a conver-  */
/* sion function.                                                     */
/*--------------------------------------------------------------------*/
/* work     = pointer to work node                                    */
/* convtype = arithmetical structure the work node must have          */
/*--------------------------------------------------------------------*/

{
     register tARITHTYPE b;



     /*---------------------------------------------------------------*/
#ifdef TRACE_ON
     d_trace("'d_testconvf()'");
#endif
     /*---------------------------------------------------------------*/



     /*---------------------------------------------------------------*/
#ifdef CHECK
     if( work == NULL )
          {d_checkerr(cISNULL,pos,"work");return(false);}
     if( work->father == NULL )
          {d_checkerr(cISNULL,work->pos,"work->father");return(false);}
#endif
     /*---------------------------------------------------------------*/



     if(work->father->namekind != d_direct)return(false);

     b = work->basetype;

     switch(convtype){
          /*----------------------------------------------------------*/
          case d_vector :
               switch(work->father->name.direct){
                    case d_rvector  : return( (b==d_real     )?true:false );
                    case d_cvector  : return( (b==d_complex  )?true:false );
                    case d_ivector  : return( (b==d_interval )?true:false );
                    case d_civector : return( (b==d_cinterval)?true:false );
                    default         : return(false);} /* switch() */
          /*----------------------------------------------------------*/
          case d_matrix :
               switch(work->father->name.direct){
                    case d_rmatrix  : return( (b==d_real     )?true:false );
                    case d_cmatrix  : return( (b==d_complex  )?true:false );
                    case d_imatrix  : return( (b==d_interval )?true:false );
                    case d_cimatrix : return( (b==d_cinterval)?true:false );
                    default         : return(false);} /* switch() */
          /*----------------------------------------------------------*/
          default :
#ifdef CHECK
                    d_checkerr(cILVALUE,work->pos,"convtype");
#endif
                    return(false);
          /*----------------------------------------------------------*/
          } /* switch(convtype) */

} /* d_testconvf() */



/**************************************************************************/



#ifdef LINT_ARGS
static void d_calcvar(register tNODE *New)
#else
static void d_calcvar(New)
register tNODE *New;
#endif

/*--------------------------------------------------------------------*/
/* This function computes the arithtype and structype of a variable   */
/* work node.                                                         */
/*--------------------------------------------------------------------*/
/* New = pointer to work node                                         */
/*--------------------------------------------------------------------*/

{
     register tVARBLK    *varinfo;
     register idptr      varid;



     /*---------------------------------------------------------------*/
#ifdef TRACE_ON
     d_trace("'d_calcvar()'");
#endif
     /*---------------------------------------------------------------*/



     /*---------------------------------------------------------------*/
#ifdef CHECK
     if( New == NULL ){d_checkerr(cISNULL,pos,"New");return;}
#endif
     /*---------------------------------------------------------------*/



     /*---------------------------------------------------------------*/
     /* computing the structype and arithtype if valid. An error mes- */
     /* sage will only be given, if variable is not function argument */
     /*---------------------------------------------------------------*/

     varinfo             = d_seman(New->name.bin);
     New->name.varinfo   = varinfo;
     New->namekind       = d_varblk;



     /*---------------------------------------------------------------*/
     /* compute the basetype result of the variable                   */
     /*---------------------------------------------------------------*/

     varid = varinfo->eltype;

     if(varid == idINTEGER        ) New->basetype = d_integer;
     if(varid == idREAL           ) New->basetype = d_real;
     if(varid == idINTERVAL       ) New->basetype = d_interval;
     if(varid == idCOMPLEX        ) New->basetype = d_complex;
     if(varid == idCINTERVAL      ) New->basetype = d_cinterval;
     if(varid == idDOTPRECISION   ) New->basetype = d_rdot;
     if(varid == idCDOTPRECISION  ) New->basetype = d_cdot;      /*New_dot*/
     if(varid == idIDOTPRECISION  ) New->basetype = d_idot;      /*New_dot*/
     if(varid == idCIDOTPRECISION ) New->basetype = d_cidot;     /*New_dot*/

     if(varid == idBAD_ID      ){d_noncritical(dBASE,dWVARBASE,New);return;}

#ifdef CHECK
     if(New->basetype == no_arithtype )
          {d_checkerr(cILVARID,New->pos,NULL);return;}
#endif




     switch(varinfo->numdim){
          /*----------------------------------------------------------*/
          case 0  : New->structure = d_scalar;
                    break;
          /*----------------------------------------------------------*/
          case 1  : if(!varinfo->original)
                         if(!d_testconvf(New,d_vector))
                              d_warn(wNOCAST,New->pos);
                    New->structure = d_vector;
                    break;
          /*----------------------------------------------------------*/
          case 2  : if(!varinfo->original)
                         if(!d_testconvf(New,d_matrix))
                              d_warn(wNOCAST,New->pos);
                    New->structure = d_matrix;
                    break;
          /*----------------------------------------------------------*/
          default : d_noncritical(dSTRUC,dWVARSTRUC,New);
          /*----------------------------------------------------------*/
          } /* switch() */

} /* d_calcvar() */



/**************************************************************************/



#ifdef LINT_ARGS
static void d_intexpr(register tNODE *New)
#else
static void d_intexpr(New)
register tNODE      *New;
#endif

/*--------------------------------------------------------------------*/
/* This function computes structure and basetype of an 'd_expression' */
/* work node from the binid of the source node                        */
/*--------------------------------------------------------------------*/
/* New    = pointer to work node                                      */
/*--------------------------------------------------------------------*/

{
     sc_type   sc_dummy;
     binptr    f, source;
     int       is_right;



     /*---------------------------------------------------------------*/
#ifdef TRACE_ON
     d_trace("'d_intexpr()'");
#endif
     /*---------------------------------------------------------------*/



     /*---------------------------------------------------------------*/
#ifdef CHECK
     if( New == NULL ){d_checkerr(cISNULL,pos,"New");return;}
#endif

     source = New->name.bin;
     f = binfather(source);
     is_right = (source == binright(f));

     if( binid(seman(source,FCALL,&sc_dummy)) != idINTEGER )
       {d_noncritical(dSELF,dINTEXPR,New);return;}

     New->name.bin  = is_right ? binright(f) : binleft(f);
     New->basetype  = d_integer;
     New->structure = d_scalar;

} /* d_intexpr() */



/**************************************************************************/



#ifdef LINT_ARGS
static void d_calcresult(register tNODE *New)
#else
static void d_calcresult(New)
register tNODE *New;
#endif

/*--------------------------------------------------------------------*/
/* This function computes the arithtype and structype type of         */
/* the worknode.                                                      */
/*--------------------------------------------------------------------*/
/* New = pointer to work node                                         */
/*--------------------------------------------------------------------*/

{
     /*---------------------------------------------------------------*/
#ifdef TRACE_ON
     d_trace("'d_calcresult()'");
#endif
     /*---------------------------------------------------------------*/



     /*---------------------------------------------------------------*/
#ifdef CHECK
     if( New == NULL ){d_checkerr(cISNULL,pos,"New");return;}
#endif
     /*---------------------------------------------------------------*/



     switch(New->kind){
          case d_function   : d_calcfunc(New);                   break;
          case d_Operator   : d_calcop(New);                     break;
          case d_variable   : d_calcvar(New);                    break;
          case d_expression : d_intexpr(New);                    break;
          case d_constant   :
          case d_nconst     : /* constants are already done */   break;
          } /* switch() */

} /* d_calcresult() */



/**************************************************************************/



#ifdef LINT_ARGS
static boolean d_usedtwice(register tNODE *work)
#else
static boolean d_usedtwice(work)
register tNODE *work;
#endif

/*--------------------------------------------------------------------*/
/* This function tests, if the found sum loop uses a run variable al- */
/* ready used before.                                                 */
/* If an error was found, the function returns 'true'.                */
/*--------------------------------------------------------------------*/
/* work = pointer to work node                                        */
/*--------------------------------------------------------------------*/

{
     register tNODE      *ptr;
     register binptr     runvar;
     register boolean    is_error;



#define RUNVAR(a) ((a)->left->left->name.varinfo->element)



     runvar    = RUNVAR(work);
     is_error  = false;

     for(ptr=work->father;ptr!=NULL;ptr=ptr->father)
          if( (ptr->namekind==d_direct) && (ptr->name.direct==d_sum) )
               if( bincompare(runvar,RUNVAR(ptr)) ){
                    d_noncritical(dNOFLAG+dSELF,dSUMNESTED,ptr->left->left);
                    is_error = true;}

     return(is_error);



#undef RUNVAR

} /* d_usedtwice() */



/**************************************************************************/



#ifdef LINT_ARGS
static boolean d_looksharp(register binptr ptr)
#else
static boolean d_looksharp(ptr)
register binptr ptr;
#endif

/*--------------------------------------------------------------------*/
/* This function makes a check through the bin-subtree 'ptr' if there */
/* exists a sharpsymbol (#, ##, #*, #<, #>). If found one it returns  */
/* the value 'true', else 'false'                                     */
/*--------------------------------------------------------------------*/
/* ptr = pointer to binptr tree (or subtree)                          */
/*--------------------------------------------------------------------*/

{
     register binptr     h_ptr;



     switch( binsymc(ptr) ){
          case cDot  :
          case cDotN :
          case cDotD :
          case cDotU :
          case cDotS : return(true);
          default    : break;} /* switch(... */



     if( (h_ptr=binleft(ptr)) != NIL )
          if( d_looksharp(h_ptr) )return(true);

     if( (h_ptr=binright(ptr)) != NIL )
          if( d_looksharp(h_ptr) )return(true);



     return(false);

} /* d_looksharp() */



/**************************************************************************/



#ifdef LINT_ARGS
tNODE *d_maketree(register binptr source,register tNODE *father)
#else
tNODE *d_maketree(source,father)
register binptr source;
register tNODE  *father;
#endif

/*--------------------------------------------------------------------*/
/* This routine creates the working binary tree from the symbol tree  */
/* and initialises all values.                                        */
/* The return value is a pointer to the created working tree.         */
/* To create a New working tree call                                  */
/*                  'New = d_maketree(head,NULL)' !                   */
/*--------------------------------------------------------------------*/
/* source = pointer to the source subtree                             */
/* father = pointer to the fathernode of the subtree                  */
/*--------------------------------------------------------------------*/

{
     register tNODE *New;
     postype         save_pos;



     /*---------------------------------------------------------------*/
#ifdef TRACE_ON
     d_trace("'d_maketree()', information is about father node:");
#endif
     /*---------------------------------------------------------------*/



     /*---------------------------------------------------------------*/
#ifdef CHECK
     if( source == NIL ){d_checkerr(cISNIL,pos,"source");return(NULL);}
#endif
     /*---------------------------------------------------------------*/


     save_pos = pos;
     pos = binpos(source);



     /*---------------------------------------------------------------*/
     /* create New working node                                       */
     /*---------------------------------------------------------------*/

     New         = d_getnode();    /* create New node                 */
     New->father = father;         /* set pointer to father           */
     New->pos    = binpos(source); /* position in the source text     */
     d_namekind(New,source);       /* set name and kind of New node   */



     /*---------------------------------------------------------------*/
     /* No Sharp-Symbols allowed in dotexpressions                    */
     /*---------------------------------------------------------------*/

     if( father == NULL ){
          if( binleft(source) != NIL )
               {d_noncritical(dNOFLAG,dBADSYNTAX,New);return(New);}
          if( binright(source) == NIL )
               {d_noncritical(dNOFLAG,dBADSYNTAX,New);return(New);}
          else
               if( d_looksharp(binright(source)) )
                    {d_noncritical(dSELF,dNONESTNG,New);return(New);}
          } /* if( father... ) */



     /*---------------------------------------------------------------*/
     /* add left and rigth subtree if original subtrees are not in-   */
     /* cluded into work tree (for example in variable work nodes)    */
     /*---------------------------------------------------------------*/

     if( (New->namekind!=d_binptr) && (New->namekind!=d_varblk) ){
          if( binleft(source) != NIL )
               New->left = d_maketree(binleft(source),New);
          if( binright(source) != NIL )
               New->right = d_maketree(binright(source),New);
          } /* if( New->namekind... ) */



     /*---------------------------------------------------------------*/
     /* compute structype and arithtype type of the New node          */
     /*---------------------------------------------------------------*/

     d_calcresult(New);



     pos = save_pos;
     return(New);  /* return pointer to working subtree    */

} /* d_maketree() */



/**************************************************************************/



#ifdef LINT_ARGS
boolean d_sumnest(register tNODE *work)
#else
boolean d_sumnest(work)
register tNODE *work;
#endif

/*--------------------------------------------------------------------*/
/* This function checks, if nested sum loops are using run variables  */
/* twice.                                                             */
/* If an error was found, the function returns 'true'.                */
/*--------------------------------------------------------------------*/
/* work = pointer to work node                                        */
/*--------------------------------------------------------------------*/

{
     if(work==NULL)return(false);




     if(work->namekind == d_direct)
          switch(work->name.direct){
               case d_plus  :
               case d_minus :
               case d_mult  :
               case d_comma :
                    return((d_sumnest(work->left)||d_sumnest(work->right)));
               case d_sum   :
                    return(d_usedtwice(work)||d_sumnest(work->right));
               default :
                    return(d_sumnest(work->right));} /* switch() */

     return(false);

} /* d_sumnest() */



/**************************************************************************/



#ifdef LINT_ARGS
global void d_remcast(register tNODE *ptr)
#else
global void d_remcast(ptr)
register tNODE *ptr;
#endif

/*--------------------------------------------------------------------*/
/* Removes all typeconversion function from work tree. Should be      */
/* called before the execution of function 'd_generate()'.            */
/*--------------------------------------------------------------------*/
/* ptr = pointer to work tree (or subtree)                            */
/*--------------------------------------------------------------------*/

{
     if( ptr == NULL )return;

     d_remcast(ptr->left);
     d_remcast(ptr->right);

     if( d_checknode(ptr,8,chkNAME,d_rvector, chkNAME,d_cvector,
                           chkNAME,d_ivector, chkNAME,d_civector,
                           chkNAME,d_rmatrix, chkNAME,d_cmatrix,
                           chkNAME,d_imatrix, chkNAME,d_cimatrix )     )
          d_extrnode(ptr);

} /* d_remcast() */



/**************************************************************************/
